/**
   webPages.cpp
   by Matthew Ford,  2021/12/06
   (c)2021 Forward Computing and Control Pty. Ltd.
   NSW, Australia  www.forward.com.au
   This code may be freely used for both private and commerical use.
   Provide this copyright is maintained.

*/
#include "HW_webpages.h"
#include "RTC_NTP_replacement.h"
#include <WiFi.h>
// from https://github.com/ESP32Async
#include <AsyncTCP.h>
#include <ESPAsyncWebServer.h>
#include "HistoryFile.h"
#include <vector>
#include <algorithm> // for std::sort
#include "locks.h"
#include "processData.h"
#include "volatileVars.h"
#include "systemTimeUtils.h"
#include <functional> // for std::function

// normally DEBUG is commented out
//#define DEBUG
static Print* debugPtr = NULL;  // local to this file

static  const size_t MAX_POINTS = 3000; // the maximum number of points to send to plot

// Cost calculation constants from processData.cpp


int haveSNTP() {
  if (isRTCDateTimeSet()) {
    return 1;
  }
  // else
  return 0;
}

static void handleApiDates(AsyncWebServerRequest *request);
static void handleDataRange(AsyncWebServerRequest *request);
static void handleCosts(AsyncWebServerRequest *request);
static void handleDirectory(AsyncWebServerRequest *request);
static void handleDumpData(AsyncWebServerRequest *request);
static void handleApiDumpData(AsyncWebServerRequest *request);

static String setMsg;

// ================= AsyncWebServer ==============
static AsyncWebServer server(80);
static void handleSetRTCDateTime(AsyncWebServerRequest * request);
static void handleSetTZstr(AsyncWebServerRequest * request);
static String correctedTZstr;  // empty if no problems
static String correctedTZstrDsc;
static String userInputTZstr;

const char* PARAM_MESSAGE = "message";
void notFound(AsyncWebServerRequest *request) {
  request->send(404, "text/plain", "Not found");
}

static String processor(const String& var) {
  String rtnString = "";
  if (debugPtr) {
    debugPtr->print("processing %"); debugPtr->print(var); debugPtr->print("% = '");
  }
  rtnString = RTC_webVariableProcessor(var);
  if (rtnString.length() == 0) {
    // try again
  }
  if (var == "HAVE_SNTP") {
    rtnString = String(haveSNTP());
  } else if (var == "TZ_VALUE") {
    rtnString = getTZstr();
  } else
    // local time in unix sec
    if (var == "TIME_CURRENT_HHMM") {
      rtnString =  String(getCurrentTime_hhmm());
    }
  if (debugPtr) {
    debugPtr->print(rtnString); debugPtr->println("'");
  }
  return rtnString;
}

void startAsyncWebServer() {
#ifdef DEBUG
  debugPtr = getDebugOut();
#endif
  if (!initializeFS()) {
    if (debugPtr) {
      debugPtr->println("LittleFS failed to start");
    }
    return;
  }

  server.on("/", HTTP_GET, [](AsyncWebServerRequest * request) {
    if (debugPtr) {
      debugPtr->println("/index.html");
    }
    request->send(LittleFS, "/index.html", "text/html");
  });
  server.on("/index.html", HTTP_GET, [](AsyncWebServerRequest * request) {
    if (debugPtr) {
      debugPtr->println("/index.html");
    }
    request->send(LittleFS, "/index.html", "text/html");
  });
  server.on("/setTime.html", HTTP_GET, [](AsyncWebServerRequest * request) {
    if (debugPtr) {
      debugPtr->println("/setTime.html returns setRTCtime.html");
    }
    request->send(LittleFS, "/setRTCtime.html", String(), false, processor);
  });
  server.on("/setRTCtime.html", HTTP_GET, [](AsyncWebServerRequest * request) {
    if (debugPtr) {
      debugPtr->println("/setRTCtime.html");
    }
    request->send(LittleFS, "/setRTCtime.html", String(), false, processor);
  });
  server.on("/settzRTC.html", HTTP_GET, [](AsyncWebServerRequest * request) {
    if (debugPtr) {
      debugPtr->println("/settzRTC.html");
    }
    request->send(LittleFS, "/settzRTC.html", String(), false, processor);
  });
  server.on("/chart.js", HTTP_GET, [](AsyncWebServerRequest * request) {
    if (debugPtr) {
      debugPtr->println("/chart.js");
    }
    request->send(LittleFS, "/chart.js", "application/javascript");
  });
  server.on("/canvasjs.min.js", HTTP_GET, [](AsyncWebServerRequest * request) {
    if (debugPtr) {
      debugPtr->println("/canvasjs.min.js");
    }
    AsyncWebServerResponse* response = request->beginResponse(LittleFS, "/canvasjs.min.js", "application/javascript");
    response->addHeader("Cache-Control", "max-age=31536000");
    request->send(response);
  });
  // API endpoints for chart data
  server.on("/api/dates", HTTP_GET, [](AsyncWebServerRequest * request) {
    if (debugPtr) {
      debugPtr->println("/api/dates");
    }
    handleApiDates(request);
  });
  server.on("/api/data-range", HTTP_GET, [](AsyncWebServerRequest * request) {
    if (debugPtr) {
      debugPtr->println("/api/data-range");
    }
    handleDataRange(request);
  });
  // Route to load style.css file
  server.on("/style.css", HTTP_GET, [](AsyncWebServerRequest * request) {
    request->send(LittleFS, "/style.css", "text/css");
  });
  server.on("/setRTCDateTime", HTTP_POST, [](AsyncWebServerRequest * request) {
    if (debugPtr) {
      debugPtr->println("/setRTCDateTime");
    }
    handleSetRTCDateTime(request);
  });
  server.on("/setTZstr", HTTP_GET, [](AsyncWebServerRequest * request) {
    if (debugPtr) {
      debugPtr->println("/setTZstr");
    }
    handleSetTZstr(request);
  });

  server.on("/resetTZ", HTTP_GET, [](AsyncWebServerRequest * request) {
    if (debugPtr) {
      debugPtr->println("/resetTZ");
    }
    resetRTC_TZ();
    request->redirect("/setRTCtime.html");
  });

  server.on("/costs", HTTP_GET, [](AsyncWebServerRequest * request) {
    if (debugPtr) {
      debugPtr->println("/costs");
    }
    handleCosts(request);
  });

  server.on("/dir", HTTP_GET, [](AsyncWebServerRequest * request) {
    if (debugPtr) {
      debugPtr->println("/dir");
    }
    handleDirectory(request);
  });

  server.on("/dumpdata", HTTP_GET, [](AsyncWebServerRequest * request) {
    if (debugPtr) {
      debugPtr->println("/dumpdata");
    }
    handleDumpData(request);
  });

  server.on("/api/dumpdata", HTTP_GET, [](AsyncWebServerRequest * request) {
    if (debugPtr) {
      debugPtr->println("/api/dumpdata");
    }
    handleApiDumpData(request);
  });

  server.onNotFound(notFound);
  if (debugPtr) {
    debugPtr->println("Starting webserver");
  }
  server.begin();
}


static void handleSetRTCDateTime(AsyncWebServerRequest * request) {
  const AsyncWebParameter *dateParam = request->getParam("DATE", true); // GET_POST
  const AsyncWebParameter *timeParam = request->getParam("TIME", true); // GET_POST
  if ((!dateParam) || (!timeParam)) {
    if (debugPtr) {
      debugPtr->print(" Error missing DATE or TIME arg");
      debugPtr->println();
    }
    return;
  }
  String dateArg = dateParam->value();
  String timeArg = timeParam->value();
  setRTC_DateTime(dateArg, timeArg);
  request->redirect("/index.html");
}

void handleSetTZstr(AsyncWebServerRequest * request) {
  const AsyncWebParameter *tzParam = request->getParam("TZ_INPUT_STR", false);
  if (!tzParam) {
    if (debugPtr) {
      debugPtr->print(" Error missing TZ_INPUT_STR");
    }
  }
  String tzArg = tzParam->value();
  // returns false for redirect("/settzRTC.html")
  // returns true for redirect("/setRTCtime.html");
  if (setRTC_TZstr(tzArg)) {
    request->redirect("/setRTCtime.html");
  } else {
    request->redirect("/settzRTC.html");
  }
}


// ================= END AsyncWebServer method==============



// only handles +v numbers
static void print2digits(String &result, uint num) {
  if (num < 10) {
    result += '0';
  }
  result += num;
}

static String formatYYYYMMDDwithSlashes(const String &yyyymmdd) {  // format yyyymmdd to yyyy/mm/dd
  size_t offset = 0;
  if (yyyymmdd.charAt(0) == '/') {
    offset = 1;
    if (yyyymmdd.length() != 9) {
      return yyyymmdd; // Return original if not expected format
    }
  } else {
    if (yyyymmdd.length() != 8) {
      return yyyymmdd; // Return original if not expected format
    }
  }
  String formatted = yyyymmdd.substring(0 + offset, 4 + offset);
  formatted += "/";
  formatted += yyyymmdd.substring(4 + offset, 6 + offset);
  formatted += "/";
  formatted += yyyymmdd.substring(6 + offset, 8 + offset);

  return formatted;
}


String minToHH_mm(int mins_in) {
  String rtn;
  int hrs = mins_in / 60;
  int mins = mins_in - (hrs * 60);
  print2digits(rtn, hrs);
  rtn += ':';
  print2digits(rtn, mins);
  return rtn;
}

/**
   Convert a local tm struct to Unix timestamp

   @param localTime The local time struct to convert
   @return Unix timestamp (seconds since Jan 1, 1970)
*/
static time_t local_tm_to_unix_time(const struct tm* tm_local) {
  if (!tm_local) {
    return 0;
  }

  // Create a copy of the input tm struct to avoid modifying the original
  struct tm tm_copy;
  memcpy(&tm_copy, tm_local, sizeof(struct tm));

  // Use mktime which converts local time to Unix timestamp
  time_t timestamp = mktime(&tm_copy);

  return timestamp;
}

//// Helper function to convert a date string (YYYYMMDD) and minutes to UTC seconds
static time_t dateTimeToUTC(const String& dateStr, int minutes, int seconds) {
  struct tm tm_struct;
  memset(&tm_struct, 0, sizeof(struct tm));
  int offset = 0;
  if (dateStr.charAt(0) == '/') {
    // path name not dateStr
    offset = 1;
  }

  if (!isDigit(dateStr.charAt(offset))) {
    if (debugPtr) {
      debugPtr->print("Error : dateStr arg: "); debugPtr->print(dateStr); debugPtr->println(" is not a datefile");
    }
    return 0;
  }
  // Parse the date string (format: YYYYMMDD)
  int year = dateStr.substring(0 + offset, 4 + offset).toInt();
  int month = dateStr.substring(4 + offset, 6 + offset).toInt();
  int day = dateStr.substring(6 + offset, 8 + offset).toInt();

  // Set date and time components
  tm_struct.tm_year = year - 1900;      // Years since 1900
  tm_struct.tm_mon = month - 1;         // Months are 0-based
  tm_struct.tm_mday = day;
  tm_struct.tm_hour = minutes / 60;     // Hours from minutes
  tm_struct.tm_min = minutes % 60;      // Remaining minutes
  tm_struct.tm_sec = seconds;           // Seconds

  time_t localTime = local_tm_to_unix_time(&tm_struct); // sets   tm_struct.tm_isdst = -1;
  return localTime;
}

// Create a structure to hold the state needed for chunk processing
struct stateStruct {
  String startDateStr;
  String endDateStr;
  int startMinutes;
  int endMinutes;
  std::vector<String> datesToProcess;
  size_t currentFileIndex;
  size_t recordsSent;
  size_t currentRecordIndex;
  size_t samplingInterval;
  bool headerSent;
  size_t totalRecordCount;
  int startDateInt;
  int endDateInt;
  bool countingComplete;
  size_t recordsProcessed;
  bool isStartDate;
  bool isEndDate;
  // lastIncludedTime removed - no longer needed for gap detection
  size_t bufferRecordsToInclude; // Count of buffer records that fall within the time range
  bool processedFileRecords; // Flag to indicate if we've processed all file records
  size_t bufferRecordIndex; // Current index in the buffer processing
  unsigned long startTime_ms; // when this call started
  // UTC time boundaries for filtering
  time_t startTimeUTC; // Start time in UTC seconds
  time_t endTimeUTC;   // End time in UTC seconds

  // Cost calculation variables
  uint64_t lastRecordTime;  // ms
  uint64_t t_offset;  // sec offset
  bool firstRecord; // set to true
  int32_t lastPowerW;
  int32_t lastOpW;
  int state; // 0 off , 1 solar, 2 OP
  double accumulatedCost;
};



// Helper function to initialize UTC time boundaries in the state struct
static void initializeUTCTimeBoundaries(stateStruct* statePtr) {
  // Calculate timezone offset

  // Calculate start time in UTC seconds
  statePtr->startTimeUTC = dateTimeToUTC(statePtr->startDateStr, statePtr->startMinutes, 0);

  // Calculate end time in UTC seconds (use 59 seconds)
  statePtr->endTimeUTC = dateTimeToUTC(statePtr->endDateStr, statePtr->endMinutes, 59);

  if (debugPtr) {
    debugPtr->print("Start date: "); debugPtr->print(statePtr->startDateStr);
    debugPtr->print(", startMinutes: "); debugPtr->print(statePtr->startMinutes);
    debugPtr->print(", startTimeUTC: "); debugPtr->println(statePtr->startTimeUTC);

    debugPtr->print("End date: "); debugPtr->print(statePtr->endDateStr);
    debugPtr->print(", endMinutes: "); debugPtr->print(statePtr->endMinutes);
    debugPtr->print(", endTimeUTC: "); debugPtr->println(statePtr->endTimeUTC);

  }
}


bool isFileNameDate(const char *name) {
  // Check if the filename matches the pattern YYYYMMDD (8 digits)
  cSF(sfFileName, 20);
  sfFileName = name;
  if (sfFileName.length() == 8 &&
      isDigit(sfFileName.charAt(0)) &&
      isDigit(sfFileName.charAt(1)) && isDigit(sfFileName.charAt(2)) && isDigit(sfFileName.charAt(3)) &&
      isDigit(sfFileName.charAt(4)) && isDigit(sfFileName.charAt(5)) && isDigit(sfFileName.charAt(6)) &&
      isDigit(sfFileName.charAt(7))) {
    return true;
  }
  return false;
}

// Check if the filename matches the pattern YYYYMMDD.csv (8 digits + .csv)
bool isFileNameDateCsv(const char *name) {
  cSF(sfFileName, 20);
  sfFileName = name;
  if (sfFileName.endsWith(".csv")) {
    sfFileName.substring(sfFileName, 0, sfFileName.length() - 4);
    return isFileNameDate(sfFileName.c_str());
  }
  return false;
}

// Binary search helper function using a comparison function to determine search direction
// Returns true if a matching record was found, false otherwise
// Result is placed in foundIndex
// compFunc should return:
//   -1 if record is too small (search in upper half)
//   0 if record matches what we're looking for (will stop search and return this index)
//   1 if record is too large (search in lower half)
// Binary search to find the first record that satisfies the condition
static bool binarySearchRecords(File& file, size_t totalRecords, size_t& foundIndex,
                                std::function<int(const BinaryRecordStruct&, struct tm*)> compFunc) {
  if (totalRecords == 0) {
    return false;  // No records to search
  }

  // Initialize search boundaries
  size_t low = 0;
  size_t high = totalRecords - 1;

  // Initialize result to indicate no match found yet
  bool foundMatch = false;
  foundIndex = 0;

  // Binary search
  while (low <= high) {
    size_t mid = low + (high - low) / 2;
    //    if (debugPtr) {
    //      debugPtr->print("low: "); debugPtr->print(low);
    //      debugPtr->print(" mid: "); debugPtr->print(mid);
    //      debugPtr->print("  high: "); debugPtr->print(high);
    //      debugPtr->println();
    //    }
    // Read the record at mid position
    file.seek(mid * sizeof(BinaryRecordStruct));

    BinaryRecordStruct record;
    file.read((uint8_t*)&record, sizeof(BinaryRecordStruct));

    // Convert to local time for filtering
    struct tm timeinfo;
    struct tm* timeinfoPtr = localtime_r(&record.utc_time, &timeinfo);

    // Compare this record
    int comparison = compFunc(record, timeinfoPtr);

    if (comparison < 0) {
      // Record is before our target time (e.g., time < startMinutes)
      // Need to look in the upper half
      low = mid + 1;
    } else {
      // Record is at or after our target time (e.g., time >= startMinutes)
      // This could be a match, but we need to find the first such match
      foundMatch = true;
      foundIndex = mid;

      // Now, look in the lower half to see if there's an earlier match
      if (mid > 0) {
        high = mid - 1;
      } else {
        // We're at the first record, can't go lower
        break;
      }
    }
  }

  return foundMatch;
}

// Binary search to find the last record that satisfies the condition
static bool binarySearchLastRecord(File& file, size_t totalRecords, size_t& foundIndex,
                                   std::function<int(const BinaryRecordStruct&, struct tm*)> compFunc) {
  if (totalRecords == 0) {
    return false;  // No records to search
  }

  // Initialize search boundaries
  size_t low = 0;
  size_t high = totalRecords - 1;

  // Initialize result to indicate no match found yet
  bool foundMatch = false;
  foundIndex = totalRecords - 1;  // Default to last record

  // Binary search
  while (low <= high) {
    size_t mid = low + (high - low) / 2;

    //    if (debugPtr) {
    //      debugPtr->print("low: "); debugPtr->print(low);
    //      debugPtr->print(" mid: "); debugPtr->print(mid);
    //      debugPtr->print("  high: "); debugPtr->print(high);
    //      debugPtr->println();
    //    }

    // Read the record at mid position
    file.seek(mid * sizeof(BinaryRecordStruct));
    BinaryRecordStruct record;
    file.read((uint8_t*)&record, sizeof(BinaryRecordStruct));

    // Convert to local time for filtering
    struct tm timeinfo;
    struct tm* timeinfoPtr = localtime_r(&record.utc_time, &timeinfo);

    // Compare this record
    int comparison = compFunc(record, timeinfoPtr);

    if (comparison <= 0) {
      // Record satisfies our condition (e.g., time <= endMinutes)
      // This could be a match, but we need to find the last such match
      foundMatch = true;
      foundIndex = mid;

      // Look in the upper half to see if there's a later match
      low = mid + 1;
    } else {
      // Record doesn't satisfy our condition
      // Need to look in the lower half
      if (mid == 0) {
        break; // Can't go lower than 0
      }
      high = mid - 1;
    }
  }

  return foundMatch;
}


// Handle API request for available date files in the form /YYYYMMDD
static void handleApiDates(AsyncWebServerRequest *request) {
  START_LOCK(filesLock);
  // List all files in the root directory and filter for date format
  File root = LittleFS.open("/");
  if (!root || !root.isDirectory()) {
    request->send(500, "application/json", "[]");
    return;
  }

  String dates = "[";
  bool firstItem = true;

  File file = root.openNextFile(); // does not include path in File.name()
  while (file) {
    String fileName = String(file.name());
    // Check if the filename matches the pattern YYYYMMDD (8 digits)  fileName does not have leading /
    if (isFileNameDate(fileName.c_str())) {
      if (!firstItem) {
        dates += ",";
      }
      firstItem = false;

      // Add the date to the JSON array (without the leading slash)
      dates += "\"" + fileName + "\"";
    }
    file = root.openNextFile();
  }

  dates += "]";
  request->send(200, "application/json", dates);
  END_LOCK(filesLock);
}

// Helper function to convert time string to minutes
static int timeStringToMinutes(const String& timeStr) {
  int colonPos = timeStr.indexOf(':');
  if (colonPos > 0) {
    int hours = timeStr.substring(0, colonPos).toInt();
    int minutes = timeStr.substring(colonPos + 1).toInt();
    return hours * 60 + minutes;
  }
  return 0; // Default to 00:00 if invalid
}



// Helper function to calculate the cost from power and off-peak usage
// calculates cost for just this timeDifference
static float calculateCost(float power_w, float op_w, float timeDiffHours) {
  // Convert power from watts to kilowatts and multiply by time to get kWh
  float powerKW = power_w / 1000.0;
  float opKW = op_w / 1000.0;

  // Calculate energy in kWh
  float energyKWh = powerKW * timeDiffHours;
  float opEnergyKWh = opKW * timeDiffHours;

  // Calculate import and export kWh
  float importKWh = 0;
  float exportKWh = 0;

  if (powerKW >= 0) {
    // Importing power (positive)
    importKWh = energyKWh;
  } else {
    // Exporting power (negative)
    exportKWh = -energyKWh; // Make export positive
  }

  // Calculate total cost
  float totalCost = (importKWh * IMPORT_COST_PER_kWhr) +
                    (exportKWh * EXPORT_COST_PER_kWhr) +
                    (opEnergyKWh * OFFPEAK_COST_PER_kWhr);

  return totalCost;
}

// Helper function to build a CSV line from a record
static String buildCsvLine(const BinaryRecordStruct& record, struct tm* timeinfoPtr, stateStruct* statePtr) {
  // Format date and time
  char dateStr[40]; // YYYY/MM/DD
  char timeStr[11];  // HH:MM:SS

  sprintf(dateStr, "%04d/%02d/%02d", timeinfoPtr->tm_year + 1900,
          timeinfoPtr->tm_mon + 1, timeinfoPtr->tm_mday);
  sprintf(timeStr, "%02d:%02d:%02d", timeinfoPtr->tm_hour,
          timeinfoPtr->tm_min, timeinfoPtr->tm_sec);

  // Set op_w from record.hw_w only if state is 2, otherwise zero
  int32_t op_w = (record.state == 2) ? record.hw_w : 0;
  // Set hw_w from record.hw_w only if state is 1, otherwise zero
  int32_t hw_w = (record.state == 1) ? record.hw_w : 0;

  // Build the CSV line - add accumulated cost as the last field
  String rtn = String(dateStr) + "," + String(timeStr) + "," + String(record.utc_time) + "," +
               String(record.power_w) + "," + String(hw_w) + "," + String(op_w) + "," +
               String(record.state) + "," + String(record.hw_triac) + "," +
               String(record.hw_boost) + "," + String(record.hw_kwh, 3) + "," +
               String(statePtr->accumulatedCost, 6) + "\n";
  return rtn;
}

// NULL record code for gaps removed

// Process record and add to chunk if appropriate, return true if more records could be added
static bool processRecord(stateStruct* statePtr, const BinaryRecordStruct& record, String& chunk,
                          size_t& recordsIncluded, size_t maxLen, bool incrementIndex, Print* debugPtr = NULL) {
  // Calculate cost for this record - do this for ALL records regardless of sampling
  // Set op_w from record.hw_w only if state is 2, otherwise zero
  int32_t op_w = (record.state == 2) ? record.hw_w : 0;
  // If this is not the first record, calculate cost between previous and current
  // for first record t_offset == 0
  if (record.utc_time < statePtr->t_offset) {
    if (!statePtr->processedFileRecords) {
      statePtr->currentRecordIndex++;
    } else {
      statePtr->bufferRecordIndex++;
    }
    return true;
  }
  unsigned long currentTime_ms =  ((record.utc_time - statePtr->t_offset) * 1000 ) + (record.utc_usec / 1000); // millis();
  if (!statePtr->firstRecord) {
    // Calculate time difference in hours
    if (currentTime_ms < statePtr->lastRecordTime) {
      // Increment the appropriate index based on what we're processing
      if (!statePtr->processedFileRecords) {
        statePtr->currentRecordIndex++;
      } else {
        statePtr->bufferRecordIndex++;
      }
      return true;
    }
    uint64_t deltaTime_ms = currentTime_ms - statePtr->lastRecordTime;
    if (deltaTime_ms < MAX_RECORD_GAP) {
      float timeDiffHours = ((float)deltaTime_ms) / (3600.0 * 1000);

      // Calculate average power between data points
      float avgPower = (statePtr->lastPowerW + record.power_w) / 2.0;
      float avgOpW = (statePtr->lastOpW + op_w) / 2.0;

      // Calculate cost for this interval
      float intervalCost = calculateCost(avgPower, avgOpW, timeDiffHours);

      // Accumulate the cost
      statePtr->accumulatedCost += intervalCost;
    } else {
      if (debugPtr) {
        debugPtr->print("processRecord skipping "); debugPtr->print(deltaTime_ms / 1000.0); debugPtr->print(" sec");
        debugPtr->print(" from "); debugPtr->print(((int64_t)(statePtr->lastRecordTime / 1000)) + statePtr->t_offset); debugPtr->println(" sec UTC");
      }
    }
  } else {
    // first record
    statePtr->firstRecord = false;
    statePtr->t_offset = record.utc_time;
    currentTime_ms = ((record.utc_time - statePtr->t_offset) * 1000 ) + (record.utc_usec / 1000);
  }

  // Update last record values for next calculation
  statePtr->lastRecordTime = currentTime_ms;
  statePtr->lastPowerW = record.power_w;
  if (record.state == 2) {
    // save for next record avg
    statePtr->lastOpW = op_w;
  } else {
    // not OP so start from 0 next time
    statePtr->lastOpW = 0;
  }
  statePtr->state = record.state;

  // Apply sampling for chart display
  if (statePtr->recordsProcessed % statePtr->samplingInterval == 0) {
    // Convert record time to local
    struct tm recordTime;
    struct tm* recordTimePtr = localtime_r(&record.utc_time, &recordTime);

    // Build the next CSV line
    String nextLine = buildCsvLine(record, recordTimePtr, statePtr);
    // Check if adding this line would exceed the maxLen
    if ((chunk.length() + nextLine.length()) < maxLen) {
      // It fits, so add it
      chunk += nextLine;
      recordsIncluded++;
      statePtr->recordsProcessed++;
      statePtr->recordsSent++;

      if (incrementIndex) {
        // Increment the appropriate index based on what we're processing
        if (!statePtr->processedFileRecords) {
          statePtr->currentRecordIndex++;
        } else {
          statePtr->bufferRecordIndex++;
        }
      }

      return true;  // Successfully processed, can continue
    } else {
      // Line won't fit, don't update counters
      return false; // Need to stop, won't fit
    }
  } else {
    // Not included due to sampling, but still count as processed
    statePtr->recordsProcessed++;

    if (incrementIndex) {
      // Increment the appropriate index based on what we're processing
      if (!statePtr->processedFileRecords) {
        statePtr->currentRecordIndex++;
      } else {
        statePtr->bufferRecordIndex++;
      }
    }

    return true;  // Successfully processed, can continue
  }
}

static void printState(stateStruct *statePtr, Print* debugPtr) {
  if (!debugPtr) {
    return;
  }
  debugPtr->println(" == state === ");
  debugPtr->print("statePtr->currentFileIndex:"); debugPtr->println(statePtr->currentFileIndex);
  if ((statePtr->currentFileIndex >= 0) && (statePtr->currentFileIndex < statePtr->datesToProcess.size())) {
    debugPtr->print("currentFileName:"); debugPtr->println(statePtr->datesToProcess[statePtr->currentFileIndex]);
  } else {
    debugPtr->println(" no currentFileName");
  }
  debugPtr->print("statePtr->currentRecordIndex:"); debugPtr->println(statePtr->currentRecordIndex);
  debugPtr->print("statePtr->samplingInterval:"); debugPtr->println(statePtr->samplingInterval);
  debugPtr->print("statePtr->headerSent:"); debugPtr->println(statePtr->headerSent);
  debugPtr->print("statePtr->recordsSent:"); debugPtr->println(statePtr->recordsSent);
  debugPtr->print("statePtr->totalRecordCount:"); debugPtr->println(statePtr->totalRecordCount);
  debugPtr->print("statePtr->countingComplete:"); debugPtr->println(statePtr->countingComplete);
  debugPtr->print("statePtr->processedFileRecords:"); debugPtr->println(statePtr->processedFileRecords);
  debugPtr->print("statePtr->recordsProcessed:"); debugPtr->println(statePtr->recordsProcessed);
}

// Handle requests for CSV data over multiple days with time range filtering
static void handleDataRange(AsyncWebServerRequest *request) {
  START_LOCK(filesLock);

  // Get parameters
  String startDateStr = "";
  String endDateStr = "";
  String startTimeStr = "00:00";
  String endTimeStr = "23:59";

  if (debugPtr) {
    debugPtr->println("handleDataRange");
  }

  if (request->hasParam("startDate")) {
    startDateStr = request->getParam("startDate")->value();
  }

  if (request->hasParam("endDate")) {
    endDateStr = request->getParam("endDate")->value();
  }

  if (request->hasParam("start")) {
    startTimeStr = request->getParam("start")->value();
  }

  if (request->hasParam("end")) {
    endTimeStr = request->getParam("end")->value();
  }

  // Convert time strings to minutes for easier comparison
  int startMinutes = timeStringToMinutes(startTimeStr);
  int endMinutes = timeStringToMinutes(endTimeStr);

  if (debugPtr) {
    debugPtr->print("Start Date: "); debugPtr->print(startDateStr);
    debugPtr->print(", End Date: "); debugPtr->print(endDateStr);
    debugPtr->print(", Start time: "); debugPtr->print(startTimeStr);
    debugPtr->print(" ("); debugPtr->print(startMinutes);
    debugPtr->print(" mins), End time: "); debugPtr->print(endTimeStr);
    debugPtr->print(" ("); debugPtr->print(endMinutes);
    debugPtr->println(" mins)");
  }

  // If no dates provided, use today
  if (startDateStr.isEmpty() || endDateStr.isEmpty()) {
    time_t now = time(nullptr);
    struct tm timeinfo;
    struct tm* timeinfoPtr = localtime_r(&now, &timeinfo);
    char buf[40];
    sprintf(buf, "%04d%02d%02d", timeinfoPtr->tm_year + 1900, timeinfoPtr->tm_mon + 1, timeinfoPtr->tm_mday);

    if (startDateStr.isEmpty()) startDateStr = String(buf);
    if (endDateStr.isEmpty()) endDateStr = String(buf);

    if (debugPtr) {
      debugPtr->print(" Using date: "); debugPtr->println(buf);
    }
  }

  // Ensure start date is not after end date
  if (startDateStr.toInt() > endDateStr.toInt()) {
    String temp = startDateStr;
    startDateStr = endDateStr;
    endDateStr = temp;
  }

  // Build a list of dates between startDate and endDate (inclusive)
  //std::vector<String> datesToProcess;
  int startDateInt = startDateStr.toInt();
  int endDateInt = endDateStr.toInt();

  // Get all file names in the root directory
  File root = LittleFS.open("/");
  if (!root || !root.isDirectory()) {
    request->send(500, "application/json", "[]");
    return;
  }

  stateStruct *statePtr = new stateStruct;
  // Initialize state with request parameters
  statePtr->startDateStr = startDateStr;
  statePtr->endDateStr = endDateStr;
  statePtr->startMinutes = startMinutes;
  statePtr->endMinutes = endMinutes;
  statePtr->currentFileIndex = 0;
  statePtr->currentRecordIndex = 0; // index into current file
  statePtr->samplingInterval = 1; // Will be calculated later
  statePtr->headerSent = false;
  statePtr->totalRecordCount = 0;
  statePtr->recordsSent = 0;
  statePtr->startDateInt = startDateInt;
  statePtr->endDateInt = endDateInt;
  statePtr->countingComplete = false;
  statePtr->recordsProcessed = 0;
  statePtr->isStartDate = false;
  statePtr->isEndDate = false;
  // lastIncludedTime initialization removed
  statePtr->bufferRecordsToInclude = 0;
  statePtr->processedFileRecords = false;
  statePtr->bufferRecordIndex = 0;
  statePtr->startTime_ms = millis();

  // Initialize cost calculation variables
  statePtr->lastRecordTime = 0;
  statePtr->t_offset = 0;
  statePtr->firstRecord = true;
  statePtr->lastPowerW = 0;
  statePtr->lastOpW = 0;
  statePtr->state = 0;
  statePtr->accumulatedCost = 0.0;

  printState(statePtr, debugPtr); // just to suppress not used warning

  // Initialize UTC time boundaries
  initializeUTCTimeBoundaries(statePtr);

  // Collect files to process
  File file = root.openNextFile();
  while (file) {
    String fileName = String(file.name());
    // Check if it's a date file
    if (isFileNameDate(fileName.c_str())) {
      // Convert filename to integer for comparison
      int fileDate = fileName.toInt();
      String filePath = "/" + fileName;

      // Check if this date is in our range
      if (fileDate >= startDateInt && fileDate <= endDateInt) {
        statePtr->datesToProcess.push_back(filePath); // Keep the leading slash
        if (debugPtr) {
          debugPtr->print("Adding file to process: ");
          debugPtr->println(filePath);
        }
      } else if (debugPtr) {
        debugPtr->print("Skipping file with no matching records: ");
        debugPtr->println(filePath);
      }
    }
    file = root.openNextFile();
  }

  // Sort the data files in ascending order (oldest to newest)
  std::sort(statePtr->datesToProcess.begin(), statePtr->datesToProcess.end());

  if (debugPtr) {
    debugPtr->print("Found "); debugPtr->print(statePtr->datesToProcess.size());
    debugPtr->println(" date files to process");
  }
  // First pass to calculate totalRecordCount (needed for downsampling)
  // This is now done in the first phase of the chunked response with optimized counting

  // printState(statePtr);
  // Send response in chunks
  AsyncWebServerResponse *response = request->beginChunkedResponse(
                                       "text/csv",
  [statePtr](uint8_t *buffer, size_t maxLen, size_t index) -> size_t {
    size_t chunkLen = 0;
    START_LOCK(filesLock);

    unsigned long start_ms = millis(); // process for at most 1sec OR maxLen

    //        if (debugPtr) {
    //          debugPtr->print(" Response maxLen: "); debugPtr->print(maxLen);
    //          debugPtr->println();
    //        }
    //     printState(statePtr,debugPtr);

    // Phase 1: Count total records across all files (on first call only) - optimized version
    if (!statePtr->countingComplete) {

      // Count total records across all files - optimized approach
      statePtr->currentFileIndex = 0;

      // Extra safety check - if datesToProcess is empty, skip counting phase
      if (statePtr->datesToProcess.empty()) {
        statePtr->countingComplete = true;
        statePtr->processedFileRecords = true;
        statePtr->totalRecordCount = 0;

        // Return header in first chunk
        String header = "Date,Time,Unix seconds,power_w,hw_w,op_w,state,hw_triac,hw_boost,hw_kwh,cost\n";
        size_t headerLen = header.length();
        if (headerLen > maxLen) headerLen = maxLen;
        memcpy(buffer, header.c_str(), headerLen);
        statePtr->headerSent = true;
        return headerLen;
      }
      while (statePtr->currentFileIndex < statePtr->datesToProcess.size()) {
        String filePath = statePtr->datesToProcess[statePtr->currentFileIndex];
        // Increment after retrieving the path, not in the same line
        statePtr->currentFileIndex++;
        if (debugPtr) {
          debugPtr->print(" statePtr->currentFileIndex incremented to: "); debugPtr->print(statePtr->currentFileIndex);
          debugPtr->print(" statePtr->datesToProcess.size():"); debugPtr->println(statePtr->datesToProcess.size());
        }

        if (LittleFS.exists(filePath)) {
          File file = LittleFS.open(filePath, "r");
          if (file) {
            //            if (debugPtr) {
            //              debugPtr->print("Counting records in file: "); debugPtr->println(filePath);
            //            }

            // Special time filtering for start and end dates
            statePtr->isStartDate = (filePath.substring(1) == statePtr->startDateStr);
            statePtr->isEndDate = (filePath.substring(1) == statePtr->endDateStr);

            // Get total number of records in file
            size_t fileSize = file.size();
            size_t totalRecordsInFile = fileSize / sizeof(BinaryRecordStruct);
            //            if (debugPtr) {
            //              debugPtr->print("totalRecordsInFile: "); debugPtr->println(totalRecordsInFile);
            //            }

            if (statePtr->isStartDate || statePtr->isEndDate) {
              //              if (debugPtr) {
              //                debugPtr->print("statePtr->isStartDate || statePtr->isEndDate "); debugPtr->println();
              //              }
              // For start/end dates, use binary search to find the appropriate range
              // and calculate records based on file positions
              if (statePtr->isStartDate) {
                // Binary search to find the first record after startTimeUTC
                size_t firstMatchingRecord = 0;
                bool foundMatch = binarySearchRecords(file, totalRecordsInFile, firstMatchingRecord,
                [statePtr](const BinaryRecordStruct & record, struct tm * timeinfoPtr) {
                  // Compare record's UTC time directly with startTimeUTC
                  //                  if (debugPtr) {
                  //                    debugPtr->print(" Compare record.utc_time: "); debugPtr->print(record.utc_time);
                  //                    debugPtr->print(" < statePtr->startTimeUTC: "); debugPtr->print(statePtr->startTimeUTC);
                  //                  }

                  if (record.utc_time < statePtr->startTimeUTC) {
                    //                    if (debugPtr) {
                    //                      debugPtr->println(" return -1");
                    //                    }
                    return -1; // Record is before our start time, look in upper half
                  } else {
                    //                    if (debugPtr) {
                    //                      debugPtr->println(" return 0");
                    //                    }
                    return 0;  // Record is at or after our start time, consider it a match
                  }
                });
                //                if (debugPtr) {
                //                  debugPtr->print("foundMatch: ");  debugPtr->print(foundMatch);  debugPtr->print("   firstMatchingRecord: "); debugPtr->println(firstMatchingRecord);
                //                }

                // Now we know the first matching record position
                if (foundMatch && firstMatchingRecord < totalRecordsInFile) {
                  if (statePtr->endDateStr == statePtr->startDateStr) {
                    // If this is both start and end date, we need to find the end position
                    size_t lastMatchingRecord = totalRecordsInFile - 1; // Default to last record

                    // Binary search to find the last record before or at endTimeUTC
                    bool foundEndMatch = binarySearchLastRecord(file, totalRecordsInFile, lastMatchingRecord,
                    [statePtr](const BinaryRecordStruct & record, struct tm * timeinfoPtr) {
                      //                      if (debugPtr) {
                      //                        debugPtr->print(" Compare record.utc_time: "); debugPtr->print(record.utc_time);
                      //                        debugPtr->print(" <= statePtr->endTimeUTC: "); debugPtr->print(statePtr->endTimeUTC);
                      //                      }

                      if (record.utc_time <= statePtr->endTimeUTC) {
                        //                        if (debugPtr) {
                        //                          debugPtr->println(" return -1");
                        //                        }
                        return -1; // Record is before or at our end time, could be a match
                      } else {
                        //                        if (debugPtr) {
                        //                          debugPtr->println(" return 1");
                        //                        }
                        return 1;  // Record is after our end time, not a match
                      }
                    });

                    //                    if (debugPtr) {
                    //                      debugPtr->print("foundEndMatch: ");  debugPtr->print(foundEndMatch);  debugPtr->print("   lastMatchingRecord: "); debugPtr->println(lastMatchingRecord);
                    //                    }

                    // DEBUG printout to help diagnose issues
                    if (debugPtr) {
                      debugPtr->print("First match: "); debugPtr->print(foundMatch);
                      debugPtr->print(", End match: "); debugPtr->print(foundEndMatch);
                      debugPtr->print(", firstMatchingRecord: "); debugPtr->print(firstMatchingRecord);
                      debugPtr->print(", lastMatchingRecord: "); debugPtr->println(lastMatchingRecord);
                    }
                    // Calculate records between first and last matching
                    if (foundMatch && foundEndMatch) {
                      if (lastMatchingRecord >= firstMatchingRecord) {
                        statePtr->totalRecordCount += (lastMatchingRecord - firstMatchingRecord + 1);
                        if (debugPtr) {
                          debugPtr->print("Records in range: ");
                          debugPtr->print(lastMatchingRecord - firstMatchingRecord + 1);
                          debugPtr->print(" (from ");
                          debugPtr->print(firstMatchingRecord);
                          debugPtr->print(" to ");
                          debugPtr->print(lastMatchingRecord);
                          debugPtr->print(") - startMinutes: ");
                          debugPtr->print(statePtr->startMinutes);
                          debugPtr->print(", endMinutes: ");
                          debugPtr->print(statePtr->endMinutes);
                          debugPtr->println();
                        }
                      } else {
                        // This shouldn't happen with proper binary search, but just to be safe
                        if (debugPtr) {
                          debugPtr->println("Error: lastMatchingRecord < firstMatchingRecord");
                        }
                        statePtr->totalRecordCount += 0;
                      }
                    } else if (foundMatch) {
                      // If we found a start match but no end match, count no records
                      // This should be rare - it means startMinutes was valid but no records before endMinutes
                      statePtr->totalRecordCount += 0;
                      if (debugPtr) {
                        debugPtr->println("Found start match but no end match");
                      }
                    } else {
                      // Neither found, count all records in the file for this date
                      statePtr->totalRecordCount += totalRecordsInFile;
                      if (debugPtr) {
                        debugPtr->print("No matches found, using all records: ");
                        debugPtr->println(totalRecordsInFile);
                      }
                    }
                    if (debugPtr) {
                      debugPtr->print("statePtr->totalRecordCount: ");  debugPtr->print(statePtr->totalRecordCount); debugPtr->println();
                    }
                  } else {
                    // This is only start date, count from first matching to end of file
                    if (foundMatch) {
                      statePtr->totalRecordCount += (totalRecordsInFile - firstMatchingRecord);
                      if (debugPtr) {
                        debugPtr->print("Start date - first matching record: ");
                        debugPtr->print(firstMatchingRecord);
                        debugPtr->print(", counting ");
                        debugPtr->print(totalRecordsInFile - firstMatchingRecord);
                        debugPtr->println(" records");
                      }
                    } else {
                      // No matching records found (all before start time), count none
                      if (debugPtr) {
                        debugPtr->println("Start date - no matching records found");
                      }
                    }
                  }
                }
              } else if (statePtr->isEndDate) {
                // Find the latest record before or at endTimeUTC
                size_t lastMatchingRecord = totalRecordsInFile - 1; // Default to last record
                bool foundMatch = binarySearchLastRecord(file, totalRecordsInFile, lastMatchingRecord,
                [statePtr](const BinaryRecordStruct & record, struct tm * timeinfoPtr) {
                  //                  if (debugPtr) {
                  //                    debugPtr->print(" Compare record.utc_time: "); debugPtr->print(record.utc_time);
                  //                    debugPtr->print(" <= statePtr->endTimeUTC: "); debugPtr->print(statePtr->endTimeUTC);
                  //                  }
                  if (record.utc_time <= statePtr->endTimeUTC) {
                    //                    if (debugPtr) {
                    //                      debugPtr->println(" return -1");
                    //                    }
                    return -1; // Record is before or at our end time, could be a match
                  } else {
                    //                    if (debugPtr) {
                    //                      debugPtr->println(" return 1");
                    //                    }
                    return 1;  // Record is after our end time, not a match
                  }
                });

                // Now we know the last matching record position
                if (foundMatch) {
                  // Found a valid last record, count from start to this position
                  statePtr->totalRecordCount += (lastMatchingRecord + 1); // +1 because it's zero-indexed
                  //                  if (debugPtr) {
                  //                    debugPtr->print("End date - last matching record: ");
                  //                    debugPtr->print(lastMatchingRecord);
                  //                    debugPtr->print(", counting ");
                  //                    debugPtr->print(lastMatchingRecord + 1);
                  //                    debugPtr->println(" records");
                  //                  }
                } else {
                  // No matching records found, count none
                  if (debugPtr) {
                    debugPtr->println("End date - no matching records found");
                  }
                }
              }
            } else {
              // For internal dates (not start or end), simply add the file size
              statePtr->totalRecordCount += totalRecordsInFile;
            }

            file.close();
          }
        }
      }

      // Check if the end date is today's date
      time_t now = time(nullptr);
      struct tm timeinfo;
      struct tm* timeinfoPtr = localtime_r(&now, &timeinfo);
      char todayDateStr[40];
      sprintf(todayDateStr, "%04d%02d%02d", timeinfoPtr->tm_year + 1900, timeinfoPtr->tm_mon + 1, timeinfoPtr->tm_mday);

      // Access the binary record buffer

      statePtr->bufferRecordsToInclude = 0;
      statePtr->processedFileRecords = false;

      if (statePtr->endDateStr == todayDateStr) {
        if (debugPtr) {
          debugPtr->print("End date is today, checking binaryRecordBuffer. Count: ");
          debugPtr->println(binaryRecordBufferCount);
        }

        // Count records in the buffer that fall within the time range
        statePtr->bufferRecordsToInclude = 0; // Reset the counter

        for (int i = 0; i < binaryRecordBufferCount; i++) {
          // Calculate the actual index in the circular buffer
          // Get the UTC time from the record
          time_t recordTimeUTC = binaryRecordBuffer[i].utc_time;

          // Create temporary end time UTC for today's date
          time_t todayEndTimeUTC = dateTimeToUTC(String(todayDateStr), statePtr->endMinutes, 59);

          // First check if record is before or at the end time
          if (recordTimeUTC <= todayEndTimeUTC) {
            // If this is also the start date, check if after start time
            if (statePtr->startDateStr == statePtr->endDateStr) {
              // Create temporary start time UTC for today's date
              time_t todayStartTimeUTC = dateTimeToUTC(String(todayDateStr), statePtr->startMinutes, 0);

              // Include if record time is after or at start time
              if (recordTimeUTC >= todayStartTimeUTC) {
                statePtr->bufferRecordsToInclude++;
              }
            } else {
              // Not checking start time, include all records before or at end time
              statePtr->bufferRecordsToInclude++;
            }
          }
        }

        if (debugPtr) {
          debugPtr->print("Found ");
          debugPtr->print(statePtr->bufferRecordsToInclude);
          debugPtr->println(" records in buffer to include");
        }

        // Add buffer records to total count only if there are any
        if (statePtr->bufferRecordsToInclude > 0) {
          statePtr->totalRecordCount += statePtr->bufferRecordsToInclude;
        }
      }

      // Final check - if we somehow have 0 records due to calculation problems,
      // and there are date files to process, count all records in all files
      if (statePtr->totalRecordCount == 0 && !statePtr->datesToProcess.empty()) {
        if (debugPtr) {
          debugPtr->println("WARNING: Record count was 0! Using total record count fallback.");
        }

        // Loop through all files and add their record counts
        for (int i = 0; i < statePtr->datesToProcess.size(); i++) {
          String filePath = statePtr->datesToProcess[i];
          if (LittleFS.exists(filePath)) {
            File file = LittleFS.open(filePath, "r");
            if (file) {
              size_t fileSize = file.size();
              size_t recordsInFile = fileSize / sizeof(BinaryRecordStruct);
              statePtr->totalRecordCount += recordsInFile;
              file.close();
            }
          }
        }
      }

      // Calculate sampling interval to keep points under MAX_POINTS (300)
      // Make sure we calculate the ceiling of the division
      if (statePtr->totalRecordCount > MAX_POINTS) {
        statePtr->samplingInterval = (statePtr->totalRecordCount + MAX_POINTS - 1) / MAX_POINTS;
      } else {
        statePtr->samplingInterval = 1;
      }

      // Double-check that our interval produces at most MAX_POINTS
      while ((statePtr->totalRecordCount / statePtr->samplingInterval) > MAX_POINTS) {
        statePtr->samplingInterval++;
      }

      if (debugPtr) {
        debugPtr->print("Total records in range: "); debugPtr->print(statePtr->totalRecordCount);
        debugPtr->print(" (incl. buffer: "); debugPtr->print(statePtr->bufferRecordsToInclude);
        debugPtr->print(", Sampling interval: "); debugPtr->println(statePtr->samplingInterval);
      }

      statePtr->countingComplete = true;
      if (statePtr->datesToProcess.empty()) {
        statePtr->processedFileRecords = true; // no files to process
      }

      // Initialize for the data reading phase
      // Always reset currentFileIndex regardless of whether datesToProcess is empty
      statePtr->currentFileIndex = 0;
      statePtr->currentRecordIndex = 0;
      statePtr->isStartDate = false;
      statePtr->isEndDate = false;

      //  printState(statePtr);

      // Cost calculation state is now part of stateStruct

      // Return header in first chunk
      String header = "Date,Time,Unix seconds,power_w,hw_w,op_w,state,hw_triac,hw_boost,hw_kwh,cost\n";
      size_t headerLen = header.length();
      if (headerLen > maxLen) headerLen = maxLen;
      memcpy(buffer, header.c_str(), headerLen);
      statePtr->headerSent = true;
      if (debugPtr) {
        debugPtr->print(" Counting Records time: "); debugPtr->print(millis() - start_ms); debugPtr->println(" ms");
        debugPtr->println("--------------");
        //debugPtr->print("Return send Header len: "); debugPtr->print(headerLen); debugPtr->println();
      }

      return headerLen;
    }

    // phase 2
    if (statePtr->bufferRecordsToInclude ||
        ((statePtr->currentFileIndex < statePtr->datesToProcess.size()))
       ) {
      //      if (debugPtr) {
      //        debugPtr->print("Phase 2: start");  debugPtr->println();
      //      }
    }
    File currentFile;
    // Phase 2: Process records from files one chunk at a time
    // Check if we're done with all files and buffer records
    if ((statePtr->currentFileIndex >= statePtr->datesToProcess.size()) &&
        //(statePtr->processedFileRecords &&
        ((statePtr->bufferRecordsToInclude == 0) || (statePtr->bufferRecordIndex >= binaryRecordBufferCount))) {
      // Clean up and signal end of response
      uint64_t recordsSent = statePtr->recordsSent;
      unsigned long elapsed_ms = millis() - statePtr->startTime_ms;
      delete statePtr;
      if (debugPtr) {
        debugPtr->print("Return finish processing files.  recordsSent: "); debugPtr->print(recordsSent);
        debugPtr->print("  Elapsed: "); debugPtr->print(elapsed_ms / 1000.0); debugPtr->println(" sec");
      }
      return 0;

    } else if (!statePtr->processedFileRecords) {

      // reopen current file
      if (debugPtr) {
        debugPtr->print(" statePtr->currentFileIndex: "); debugPtr->print(statePtr->currentFileIndex);
        debugPtr->print(" statePtr->datesToProcess.size(): "); debugPtr->print(statePtr->datesToProcess.size());
        debugPtr->println();
      }

      // Check if currentFileIndex is valid
      if (statePtr->currentFileIndex < statePtr->datesToProcess.size()) {
        currentFile = LittleFS.open(statePtr->datesToProcess[statePtr->currentFileIndex], "r");
      }
      if (currentFile) {
        if (statePtr->currentRecordIndex > 0) {
          currentFile.seek(statePtr->currentRecordIndex * sizeof(BinaryRecordStruct));
        }
      } // else if open fails will go onto next file below
      // Apply time filtering - ensure currentFileIndex is valid before accessing
      if (statePtr->currentFileIndex < statePtr->datesToProcess.size()) {
        statePtr->isStartDate = (statePtr->datesToProcess[statePtr->currentFileIndex].substring(1) == statePtr->startDateStr);
        statePtr->isEndDate = (statePtr->datesToProcess[statePtr->currentFileIndex].substring(1) == statePtr->endDateStr);
      } else {
        // Safety: If index is out of bounds, mark as processed files
        statePtr->processedFileRecords = true;
        statePtr->isStartDate = false;
        statePtr->isEndDate = false;
      }
    }
    // else
    // calculate this each time we are called
    static unsigned long MAX_PROCESSING_ms = 500;
    String chunk = "";
    size_t recordsIncluded = 0;
    //size_t MAX_LINE_LEN = 100; // approx
    //size_t MAX_RECORDS_PER_CHUNK = maxLen / MAX_LINE_LEN;
    // loop until recordsProcessed or no space or too long

    while (((millis() - start_ms) < MAX_PROCESSING_ms)) {
      // First process file records, then buffer records when files are done
      if (!statePtr->processedFileRecords) {
        // Process records from files

        // If current file is exhausted, move to next file
        if (!currentFile || !currentFile.available()) {
          if (debugPtr) {
            debugPtr->print(" currentFile not Open or not empty/finished "); debugPtr->println();
          }
          if (currentFile) {
            currentFile.close();
          }
          statePtr->currentFileIndex++;
          if (debugPtr) {
            debugPtr->print(" statePtr->currentFileIndex incremented to: "); debugPtr->print(statePtr->currentFileIndex);
            debugPtr->print(" statePtr->datesToProcess.size():"); debugPtr->println(statePtr->datesToProcess.size());
          }
          statePtr->currentRecordIndex = 0; // Reset record index for the new file

          // Check if we've processed all files
          if (statePtr->currentFileIndex >= statePtr->datesToProcess.size()) {
            if (debugPtr) {
              debugPtr->print(" end of files to process "); debugPtr->println();
            }

            // Mark that we've processed all file records
            statePtr->processedFileRecords = true;

            // Check if we need to process records from binaryRecordBuffer
            time_t now = time(nullptr);
            struct tm timeinfo;
            struct tm* timeinfoPtr = localtime_r(&now, &timeinfo);
            char todayDateStr[40];
            sprintf(todayDateStr, "%04d%02d%02d", timeinfoPtr->tm_year + 1900, timeinfoPtr->tm_mon + 1, timeinfoPtr->tm_mday);

            if (statePtr->endDateStr == todayDateStr) { // && statePtr->bufferRecordsToInclude > 0) {
              if (debugPtr) {
                debugPtr->println("Processing buffer records after files");
                debugPtr->print(" statePtr->bufferRecordsToInclude: "); debugPtr->println(statePtr->bufferRecordsToInclude);
              }
              statePtr->bufferRecordIndex = 0; // Reset to start of buffer for processing
              continue; // Skip to next iteration to process buffer
            }
            continue; // Skip to next iteration since we're done with files
          }
          // else still working on files - index should be valid, but double check
          if (statePtr->currentFileIndex < statePtr->datesToProcess.size()) {
            currentFile = LittleFS.open(statePtr->datesToProcess[statePtr->currentFileIndex], "r");
          } else {
            // Safety: If index is out of bounds, mark as processed files
            statePtr->processedFileRecords = true;
            continue;
          }
          //          if (debugPtr) {
          //            debugPtr->print("Opened file to process: ");
          //            debugPtr->print(statePtr->datesToProcess[statePtr->currentFileIndex]);
          //            debugPtr->println();
          //          }

          // Apply time filtering with bounds checking
          if (statePtr->currentFileIndex < statePtr->datesToProcess.size()) {
            statePtr->isStartDate = (statePtr->datesToProcess[statePtr->currentFileIndex].substring(1) == statePtr->startDateStr);
            statePtr->isEndDate = (statePtr->datesToProcess[statePtr->currentFileIndex].substring(1) == statePtr->endDateStr);
          } else {
            // Safety: If index is out of bounds, mark as processed files
            statePtr->isStartDate = false;
            statePtr->isEndDate = false;
            statePtr->processedFileRecords = true;
            continue;
          }

          // Do not reset lastIncludedTime when moving to a new file
          // We want to detect gaps between files as well

          // Skip to next iteration if file couldn't be opened
          if (!currentFile) {
            //            if (debugPtr) {
            //              debugPtr->print(" could not open file continue "); debugPtr->println();
            //            }
            continue;
          }
        }

        // Process records from current file
        BinaryRecordStruct record;

        // Check if we have enough data for a record
        if (currentFile.available() < sizeof(BinaryRecordStruct)) {
          if (debugPtr) {
            debugPtr->print(" less then one record in this file"); debugPtr->println();
          }
          currentFile.close(); // skip rest of this file
          continue;
        }

        // Read record
        currentFile.read((uint8_t*)&record, sizeof(BinaryRecordStruct));
        // Apply time filtering using UTC time comparison
        bool includeRecord = true; // for other date files

        // If this is the start date, check against the pre-calculated startTimeUTC
        if (statePtr->isStartDate) {
          if (statePtr->firstRecord) { // check start time
            if (record.utc_time < statePtr->startTimeUTC) {
              includeRecord = false;
            }
          }
        }

        if (!statePtr->firstRecord) { // have valid statePtr->lastRecordTime
          // check for invalid dates max diff 25hrs
          //lastRecordTime =  currentTime_ms = ((record.utc_time - statePtr->t_offset) * 1000 ) + (record.utc_usec / 1000); for last record
          if (record.utc_time < statePtr->t_offset) {
            // skip this on as too early
            statePtr->currentRecordIndex++;
            continue;
          }
          unsigned long currentTime_ms =  ((record.utc_time - statePtr->t_offset) * 1000 ) + (record.utc_usec / 1000); // millis();
          if (currentTime_ms < statePtr->lastRecordTime) {
            // skip this on as time went backwards
            statePtr->currentRecordIndex++;
            continue;
          } else { // >=
            unsigned long delta_ms = currentTime_ms - statePtr->lastRecordTime;
            if (delta_ms > (25ul * 60 * 60 * 1000)) {
              // more than 25hrs in advance
              statePtr->currentRecordIndex++;
              continue;
            }
          }
        }
        // If this is the end date, check against the pre-calculated endTimeUTC
        if (statePtr->isEndDate) {
          if (record.utc_time > statePtr->endTimeUTC) {
            //if (debugPtr) {
            //  debugPtr->print(" passed end date break "); debugPtr->println();
            //}
            statePtr->currentFileIndex = statePtr->datesToProcess.size(); // skip the other files also
            // Mark that we've processed all file records
            statePtr->processedFileRecords = true;
            continue;
          }
        }

        if (includeRecord) {
          // Process the record using our common method
          if (!processRecord(statePtr, record, chunk, recordsIncluded, maxLen, true, debugPtr)) {
            // If processing couldn't continue (not enough space), break the loop
            break;
          }
        } else {
          // Not included due to filtering, but still increment the current record index
          statePtr->currentRecordIndex++;
        }
      }
      // else process buffer if have some
      else if (statePtr->bufferRecordsToInclude > 0) {
        // Process records from buffer

        time_t now = time(nullptr);
        struct tm timeinfo;
        struct tm* timeinfoPtr = localtime_r(&now, &timeinfo);
        char todayDateStr[40];
        sprintf(todayDateStr, "%04d%02d%02d", timeinfoPtr->tm_year + 1900, timeinfoPtr->tm_mon + 1, timeinfoPtr->tm_mday);

        // Find the next buffer record that matches our criteria
        bool foundRecord = false;
        BinaryRecordStruct record;

        while (!foundRecord && statePtr->bufferRecordIndex < binaryRecordBufferCount) {
          // Calculate the actual index in the circular buffer
          // Get the record
          record = binaryRecordBuffer[statePtr->bufferRecordIndex];

          // Get the UTC time from the record
          time_t recordTimeUTC = record.utc_time;

          // Create temporary end time UTC for today's date
          time_t todayEndTimeUTC = dateTimeToUTC(String(todayDateStr), statePtr->endMinutes, 59);

          // First check if record is before or at the end time
          if (recordTimeUTC <= todayEndTimeUTC) {
            // If this is also the start date, check if after start time
            if (statePtr->startDateStr == todayDateStr) {
              // Create temporary start time UTC for today's date
              time_t todayStartTimeUTC = dateTimeToUTC(String(todayDateStr), statePtr->startMinutes, 0);

              // Include if record time is after or at start time
              if (recordTimeUTC >= todayStartTimeUTC) {
                foundRecord = true;
              }
            } else {
              // Not checking start time, include all records before or at end time
              foundRecord = true;
            }
          }

          if (!foundRecord) {
            statePtr->bufferRecordIndex++;
          }
        }

        // If we found a matching record, process it
        if (foundRecord) {
          // Process the record using our common method
          if (!processRecord(statePtr, record, chunk, recordsIncluded, maxLen, true, debugPtr)) {
            // If processing couldn't continue (not enough space), break the loop
            break;
          }

          // Check if we've processed all buffer records
          if (statePtr->bufferRecordIndex >= binaryRecordBufferCount) {
            if (debugPtr) {
              debugPtr->println("Finished processing buffer records");
            }
            break;
          }
        } else {
          // No more matching records in buffer
          if (debugPtr) {
            debugPtr->println("No more matching records in buffer");
          }
          break;
        }
      } else {
        // No more records to process
        break;
      }

    }

    // If no records were included in this time slice but we have dates to process,
    // return \r\n so that a non-zero length is returned to the asyncwebserver callback
    if (recordsIncluded == 0 && chunk.length() == 0 && statePtr->currentFileIndex < statePtr->datesToProcess.size()) {
      chunk = "\r\n";
      if (debugPtr) {
        debugPtr->println("No records processed in this time slice, returning empty line");
      }
    }

    // Copy chunk to buffer
    chunkLen = chunk.length();
    if (chunkLen > maxLen) chunkLen = maxLen;
    if (chunkLen > 0) {
      memcpy(buffer, chunk.c_str(), chunkLen);
    }
    //  printState(statePtr);
    if (currentFile) {
      currentFile.close();
    }

    if (debugPtr) {
      debugPtr->print("Chunk maxlen: "); debugPtr->print(maxLen); debugPtr->print("  ");
      debugPtr->print("Return chunk len: "); debugPtr->print(chunkLen); debugPtr->print("  ");
      debugPtr->print("Records Included: "); debugPtr->print(recordsIncluded); debugPtr->print("  ");
      debugPtr->print("Records Processed: "); debugPtr->print(statePtr->recordsProcessed); debugPtr->print("  ");
      debugPtr->println();
      debugPtr->print(" Processing time: "); debugPtr->print(millis() - start_ms); debugPtr->println(" ms");
      debugPtr->println("--------------");
    }
    END_LOCK(filesLock);
    return chunkLen;
  }
                                     );
  request->send(response);
  END_LOCK(filesLock);
}


const char *getCostsFilename() {
  static char costFilename[] = "costs.csv";
  return (const char*)costFilename;
}

void dumpCostsFile(Print * outPtr) {
  if (!outPtr) {
    return;
  }
  // Save results to costs.csv
  String fileName = getCostsFilename();
  fileName = "/" + fileName;
  START_LOCK(filesLock);
  File costsFile = LittleFS.open(fileName.c_str());
  if (costsFile) {
    while (costsFile.available()) {
      char c = costsFile.read();
      outPtr->print(c);
    }
    costsFile.close();
  } else {
    if (debugPtr) {
      debugPtr->print("// "); debugPtr->print(fileName); debugPtr->println("  not found.");
    }
  }
  END_LOCK(filesLock);
}
/**
   Calculate the daily cost for a given date file
   Opens the file and calculates totals for import/export kWh and Op_kWh/hw_kWh
   by integrating power consumption between consecutive records

   @param filename The date filename (YYYYMMDD format)
*/
float calculateDailyCost(const char *filename,  Accumulated_kWh_RecordStruct & accResults, bool writeCostsFile) {
  Print* oldPtr = debugPtr;
  // debugPtr = &Serial;

  if (debugPtr) {
    debugPtr->print("  Calculating daily cost for file: ");
    debugPtr->println(filename);
  }
  // Convert filename to a local date
  String dateStr = String(filename);

  // Calculate start and end times for the day in UTC
  time_t startTimeUTC = dateTimeToUTC(dateStr, 0, 0);      // 00:00:00
  time_t endTimeUTC = dateTimeToUTC(dateStr, 23 * 60 + 59, 59); // 23:59:59

  if (debugPtr) {
    debugPtr->print("  Start time UTC: "); debugPtr->print(startTimeUTC);
    debugPtr->print("  End time UTC: "); debugPtr->println(endTimeUTC);
  }

  initAccumulatedKWhRecord(accResults);

  // Open the file for reading
  START_LOCK(filesLock);
  String filePath = "/" + dateStr;
  File dataFile = LittleFS.open(filePath, "r");

  if (!dataFile) {
    if (debugPtr) {
      debugPtr->print("Failed to open file: ");
      debugPtr->println(filePath);
    }
    return 0;
  }

  BinaryRecordStruct record;

  // Read records and accumulate values
  while (dataFile.available() >= sizeof(BinaryRecordStruct)) {
    // Read the record
    dataFile.read((uint8_t*)&record, sizeof(BinaryRecordStruct));

    // Filter records by time (00:00:00 to 23:59:59 in local time)
    if (record.utc_time >= startTimeUTC && record.utc_time <= endTimeUTC) {
      // Use the common accumulate_kWh function to process each record
      accumulate_kWh(record, accResults);
    }
  }

  dataFile.close();
  END_LOCK(filesLock);

  // Calculate total cost
  accResults.cost = (accResults.import_kWhrs * IMPORT_COST_PER_kWhr) +
                    (accResults.export_kWhrs * EXPORT_COST_PER_kWhr) +
                    (accResults.offPeak_kWhrs * OFFPEAK_COST_PER_kWhr);

  if (debugPtr) {
    debugPtr->print("Daily calculations:");
    print_Accumulated_kWh(accResults, debugPtr);
    //    debugPtr->print("  Import kWh: "); debugPtr->print(accResults.import_kWhrs);
    //    debugPtr->print("  Export kWh: "); debugPtr->print(accResults.export_kWhrs);
    //    debugPtr->print("  Off-peak kWh: "); debugPtr->print(accResults.offPeak_kWhrs);
    //    debugPtr->print("  HW kWh: "); debugPtr->print(accResults.hotwater_kWhrs);
    //    debugPtr->print("  HW fullyHeated: "); debugPtr->print(accResults.fullyHeated?"true":"false");
    //    debugPtr->print("  Total cost: $"); debugPtr->println(accResults.cost, 6);
  }
  String header = "Date , Time , Import_kWh , Export_kWh , OffPeak_kWh , HW_kWh , HW_Fully_Heated, Cost , JSY_Import_kWh , JSY_Export_kWh , JSY_OffPeak_kWh , JSY_HW_kWh";

  String date = formatYYYYMMDDwithSlashes(dateStr);
  String time = getCurrentTime_hhmmss();
  int daySecs = getLocalTime_sec();
  if (daySecs < 30) { // within 30sec of new day
    // fudge the time to be just before midnight, even if a few sec after
    // usually this method is called with write try just after midnight say 2sec after
    time = "23:59:59";
  }
  cSF(sfLine, 200);
  sfLine = date.c_str();
  sfLine.print(" , ");
  sfLine += time.c_str();
  sfLine.print(" , ");
  sfLine.print(accResults.import_kWhrs ,  3);
  sfLine.print(" , ");
  sfLine.print(accResults.export_kWhrs ,  3);
  sfLine.print(" , ");
  sfLine.print(accResults.offPeak_kWhrs ,  3);
  sfLine.print(" , ");
  sfLine.print(accResults.hotwater_kWhrs ,  3);
  sfLine.print(" , ");
  sfLine.print(accResults.fullyHeated);
  sfLine.print(" , ");
  sfLine.print(accResults.cost ,  3);
  sfLine.print(" , ");
  sfLine.print(JSY_Import_kWh_v ,  3);
  sfLine.print(" , ");
  sfLine.print(JSY_Export_kWh_v ,  3);
  sfLine.print(" , ");
  sfLine.print(JSY_OffPeak_kWh_v ,  3);
  sfLine.print(" , ");
  sfLine.print(JSY_HW_kWh_v ,  3);
  sfLine.println();

  if (debugPtr) {
    debugPtr->print(" Header  : "); debugPtr->println(header);
    debugPtr->print(" CSV line: "); debugPtr->print(sfLine);
  }

  if (writeCostsFile) {
    START_LOCK(filesLock);
    // Save results to costs.csv
    File costsFile = LittleFS.open("/costs.csv", "a");
    if (costsFile) {
      // Check if file is empty - if so, add header
      if (costsFile.size() == 0) {
        costsFile.println(header);
        if (debugPtr) {
          debugPtr->print(" write header line: "); debugPtr->println(header);
        }
      }

      costsFile.print(sfLine.c_str());
      costsFile.close();

      if (debugPtr) {
        debugPtr->println("Results saved to costs.csv");
      }
    } else {
      if (debugPtr) {
        debugPtr->println("Failed to open costs.csv for writing");
      }
    }
    END_LOCK(filesLock);
  }
  debugPtr = oldPtr;
  return accResults.cost;
}

// Structure to hold state between chunk calls
struct CostsStateStruct {
  File costsFile;
  bool fileOpen;
  size_t position;
  bool finished;
};

/**
   Handle the /costs route to display the contents of costs.csv as raw text
   Uses chunked response with processing time limited to < 500ms per chunk
*/
static void handleCosts(AsyncWebServerRequest * request) {
  if (debugPtr) {
    debugPtr->println("/costs");
  }

  // Create state structure and allocate memory for it
  CostsStateStruct* statePtr = new CostsStateStruct();
  statePtr->fileOpen = false;
  statePtr->position = 0;
  statePtr->finished = false;

  // Set up a chunked response
  AsyncWebServerResponse *response = request->beginChunkedResponse(
                                       "text/plain",
  [statePtr](uint8_t *buffer, size_t maxLen, size_t index) -> size_t {
    size_t chunkLen = 0;
    unsigned long start_ms = millis(); // Track processing time
    if (statePtr->finished) {
      delete statePtr;
      return 0; // finished
    }
    // Open the file if not already open
    if (!statePtr->fileOpen) {
      String fileName = "/" + String(getCostsFilename());
      statePtr->costsFile = LittleFS.open(fileName.c_str(), "r");

      if (!statePtr->costsFile) {
        // If file doesn't exist, return error message
        const char* errMsg = "Error: costs.csv file not found\n";
        size_t msgLen = strlen(errMsg);
        // should check maxLen here likely to be OK but just in case

        // Copy error message to buffer
        memcpy(buffer, errMsg, msgLen);

        // Clean up
        statePtr->finished = true;
        return msgLen;
      }

      statePtr->fileOpen = true;
      statePtr->costsFile.seek(statePtr->position);
    }

    // Read from file to buffer, respecting time limit (500ms)
    while ((chunkLen < maxLen) &&
           (statePtr->costsFile.available()) &&
           (millis() - start_ms < 500)) {
      buffer[chunkLen++] = statePtr->costsFile.read();
      statePtr->position++;
    }

    // Check if we've reached the end of the file
    if (!statePtr->costsFile.available()) {
      // Close file and clean up
      if (statePtr->fileOpen) {
        statePtr->costsFile.close();
      }
      statePtr->finished = true;
      return chunkLen; // Return final chunk length
    }

    // Update file position for next chunk
    return chunkLen;
  }
                                     );

  // Set headers for plain text
  response->addHeader("Content-Type", "text/plain");

  // Send the response
  request->send(response);
}

/**
   Handle the /dir route to display directory listing
*/
static void handleDirectory(AsyncWebServerRequest * request) {
  if (debugPtr) {
    debugPtr->println("/dir");
  }

  // Using StringStream to capture listDir output
  class StringStream : public Print {
    public:
      String contents;

      size_t write(uint8_t c) override {
        contents += (char)c;
        return 1;
      }

      size_t write(const uint8_t *buffer, size_t size) override {
        contents.concat((const char*)buffer, size);
        return size;
      }
  };

  StringStream output;
  output.contents = ""; //"Directory listing:\n\n";
  showLittleFS_size(&output);
  output.println();
  listDir("/", &output);

  // Send the output as plain text
  request->send(200, "text/plain", output.contents);
}

/**
   Handle the /dumpdata route to display a list of date files
*/
static void handleDumpData(AsyncWebServerRequest * request) {
  unsigned long noRecs = appendBufferToFile();
  if (debugPtr) {
    debugPtr->println("/dumpdata ");
    debugPtr->print("// Saved buffered records. ");  debugPtr->print(noRecs * 2); debugPtr->println(" secs of data.");
  }

  // Get all date files from the filesystem
  String html = "<!DOCTYPE html>\n<html>\n<head>\n";
  html += "<title>Dump Data Files</title>\n";
  html += "<meta name=\"viewport\" content=\"width=device-width, initial-scale=1\">\n";
  html += "<link rel=\"icon\" href=\"data:,\">\n";
  html += "<link rel=\"stylesheet\" type=\"text/css\" href=\"style.css\">\n";
  html += "</head>\n<body>\n";
  html += "<h2>Dump Data Files</h2>\n";
  html += "<div class=\"content\">\n";
  html += "<h3>Select a date file to dump:</h3>\n";
  html += "(";
  html += noRecs * 2;
  html += " secs of pending buffered data has been written to today's file)\n";
  html += "<ul>\n"; // start of file list

  // List all files in the root directory and filter for date files
  START_LOCK(filesLock);
  File root = LittleFS.open("/");
  if (root && root.isDirectory()) {
    std::vector<String> dateFiles;

    File file = root.openNextFile();
    while (file) {
      String fileName = String(file.name());
      if (isFileNameDate(fileName.c_str())) {
        dateFiles.push_back(fileName);
      }
      file = root.openNextFile();
    }

    // Sort date files in descending order (newest first)
    std::sort(dateFiles.begin(), dateFiles.end(), std::greater<String>());

    // Add each date file to the HTML
    for (const String& dateFile : dateFiles) {
      // Format date for display (YYYYMMDD to YYYY/MM/DD)
      String formattedDate = formatYYYYMMDDwithSlashes(dateFile);
      html += "<a href=\"/api/dumpdata?filename=" + dateFile + "\">" + formattedDate + "</a><br><br>\n";
    }

    if (dateFiles.empty()) {
      html += "No date files found\n";
    }
  } else {
    html += "Error accessing filesystem\n";
  }
  END_LOCK(filesLock);

  html += "</ul>\n";
  html += "</div>\n</body>\n</html>";

  request->send(200, "text/html", html);
}

// Structure to hold state for chunked data file response
struct DumpDataStateStruct {
  String filePath;   // Path to the data file that is being dumped
  bool finished;  // delete this struct pointer and return 0 if finished == true;
  bool headerSent;  // Track if header has been sent already
  size_t position; // last position processed
  time_t last_rec_time;
  bool haveInvalidData;
};

/**
   Handle the /api/dumpdata route to fetch and display file contents
   Uses chunked response with processing time limited to less than 500ms per chunk
*/
static void handleApiDumpData(AsyncWebServerRequest * request) {
  if (debugPtr) {
    debugPtr->println("/api/dumpdata");
  }

  // Get filename parameter
  if (!request->hasParam("filename")) {
    request->send(400, "text/plain", "Missing filename parameter");
    return;
  }

  String filename = request->getParam("filename")->value();
  if (filename.length() == 0) {
    request->send(400, "text/plain", "Invalid filename");
    return;
  }

  char fullPath[32];
  sprintf(fullPath, "/%s", filename.c_str());
  // Create state structure and allocate memory for it
  DumpDataStateStruct* statePtr = new DumpDataStateStruct();
  statePtr->filePath = String(fullPath);
  statePtr->finished = false;
  statePtr->headerSent = false;
  statePtr->position = 0;
  statePtr->last_rec_time = 0;
  statePtr->haveInvalidData = false;

  unsigned long noRecs = appendBufferToFile(); // protected by lock
  if (debugPtr) {
    debugPtr->print("// Saved buffered records. ");  debugPtr->print(noRecs * 2); debugPtr->println(" secs of data.");
  }

  // Set up a chunked response
  AsyncWebServerResponse *response = request->beginChunkedResponse(
                                       "text/plain",
  [statePtr](uint8_t *buffer, size_t maxLen, size_t index) -> size_t {
    size_t processedBytes = 0;
    START_LOCK(filesLock);

    unsigned long start_ms = millis(); // Track processing time
    cSF(sfCSVline, 128);
    if (statePtr->finished) {
      delete statePtr;
      return 0; // finished
    }

    // Open the file for each chunk
    File dataFile = LittleFS.open(statePtr->filePath, FILE_READ);
    if (!dataFile) {
      String errorMsg = "Error: Could not open data file ";
      errorMsg += statePtr->filePath;
      memcpy(buffer, errorMsg.c_str(), errorMsg.length());
      statePtr->finished = true;
      return errorMsg.length();
    }

    // If header already sent, seek to the saved position
    if (statePtr->headerSent) {
      dataFile.seek(statePtr->position);
    } else {
      // First time opening, send the header
      String header = "//FileName:";
      header += statePtr->filePath;
      header += "\r\n// local date,local time,utc tv_sec,utc tv_usec,power_w,hw_w,state,hw_triac,hw_boost,hw_kwh\r\n";
      size_t len = min(maxLen, header.length());
      memcpy(buffer, header.c_str(), len); //header.length());
      statePtr->position = 0;
      statePtr->headerSent = true;  // Mark header as sent
      return len;
    }
    // else file Open
    // Calculate how much data to send in this chunk
    char dateStr[40];  // YYYY/MM/DD
    char timeStr[9];   // HH:MM:SS
    BinaryRecordStruct record;
    size_t recordSize = sizeof(BinaryRecordStruct);

    while (((millis() - start_ms) < 5000) && (processedBytes < (maxLen - (2 * sfCSVline.capacity())))) {
      // Reserve space for a full record output
      if (dataFile.read((uint8_t*)&record, recordSize) == recordSize) {
        if (statePtr->last_rec_time == 0) {
          statePtr->last_rec_time = record.utc_time;
        }
        bool invalid = false;
        if (record.utc_time < statePtr->last_rec_time) {
          invalid = true;
        } else if ((record.utc_time - statePtr->last_rec_time) > (25ul * 60 * 60)) {
          invalid = true;
        }
        if (invalid) {
          statePtr->haveInvalidData = true;
          sfCSVline.clear();
          sfCSVline = " == Invalid Time Stamp for next sample";
          sfCSVline.println();
          int len = sfCSVline.length();
          if (processedBytes + len <= maxLen) {
            memcpy(buffer + processedBytes, sfCSVline.c_str(), len);
            processedBytes += len;
          }
        } else {
          statePtr->last_rec_time = record.utc_time;
        }
        time_t localTime;
        struct tm timeinfo;

        // Convert UTC to local time
        localTime = record.utc_time;
        localtime_r(&localTime, &timeinfo);

        // Format date and time
        sprintf(dateStr, "%04d/%02d/%02d",
                timeinfo.tm_year + 1900,
                timeinfo.tm_mon + 1,
                timeinfo.tm_mday);

        sprintf(timeStr, "%02d:%02d:%02d",
                timeinfo.tm_hour,
                timeinfo.tm_min,
                timeinfo.tm_sec);

        // Print record as CSV
        sfCSVline.clear();
        sfCSVline.print(dateStr);  sfCSVline.print(',');
        sfCSVline.print(timeStr);  sfCSVline.print(',');
        sfCSVline.print((int64_t)record.utc_time); sfCSVline.print(',');
        sfCSVline.print((uint32_t)record.utc_usec); sfCSVline.print(',');

        sfCSVline.print(record.power_w);  sfCSVline.print(',');
        sfCSVline.print(record.hw_w);  sfCSVline.print(',');
        sfCSVline.print(record.state);  sfCSVline.print(',');
        sfCSVline.print(record.hw_triac);  sfCSVline.print(',');
        sfCSVline.print(record.hw_boost);  sfCSVline.print(',');
        sfCSVline.print(record.hw_kwh, 3);  sfCSVline.println();

        int len = sfCSVline.length();
        if (processedBytes + len <= maxLen) {
          memcpy(buffer + processedBytes, sfCSVline.c_str(), len);
          processedBytes += len;
        } else {
          dataFile.seek(dataFile.position() - recordSize);
          break;
        }
      } else { // read returned 0 eof
        // Check if we've processed all data
        String footer = "// EOF";
        if (statePtr->haveInvalidData) {
          footer += " == Invalid Time Stamp";
        }
        footer += "\n";
        memcpy(buffer + processedBytes, footer.c_str(), footer.length());
        processedBytes += footer.length();
        statePtr->finished = true;
        break;
      }
    }
    //    if (debugPtr) {
    //      debugPtr->print(" maxLen: "); debugPtr->print(maxLen);
    //      debugPtr->print(" processedBytes: "); debugPtr->print(processedBytes);
    //      debugPtr->println();
    //    }
    // Close the file after each chunk to allow other parts of the program to update it
    statePtr->position = dataFile.position();
    dataFile.close();
    END_LOCK(filesLock);
    return processedBytes;
  }
                                     );

  // Send the chunked response
  request->send(response);
}
