/*
  (c)2023 Forward Computing and Control Pty. Ltd.
  NSW Australia, www.forward.com.au
  This code is not warranted to be fit for any purpose. You may only use it at your own risk.
  This code may be freely used for both private and commercial use
  Provided this copyright is maintained.
*/
#include <Arduino.h>
#include "FS.h"
#include "HistoryRecord.h"
#include "HistoryFile.h"
#include "SafeStringReader.h"
#include "millisDelay.h"
#include "processData.h"
#include "locks.h"
#include "RTC_NTP_replacement.h"

extern void handleHotWaterRemoteClient(); // in SolarHotWaterMonitor_V2 call often
extern HistoryRecordStruct historyRecord;

#define LINE_LEN 265
createSafeStringReader(sfReader, LINE_LEN, "\r\n"); // allow for lines 256 long
cSF(sfToken, 80); // for reading csv values

//#define DEBUG
static Print* debugPtr =  NULL;

static const char HISTORY_FILE_NAME[] = "/history.csv";

void initHistory() {
#ifdef DEBUG
  debugPtr = getDebugOut();
#endif
  (void)(debugPtr);
}


// Binary record buffer constants
const size_t BINARY_RECORD_BUFFER_SIZE = 1200; // x 24 => 32.2 kB 1hrs of 2sec  // one / 2sec => 48mins  One record per minute for 24 hours
const size_t MAX_BINARY_RECORD_BUFFER_SIZE = BINARY_RECORD_BUFFER_SIZE - 10; // x 24 => 32.2 kB 1hrs of 2sec  // one / 2sec => 48mins  One record per minute for 24 hours

// Binary record circular buffer implementation
BinaryRecordStruct binaryRecordBuffer[BINARY_RECORD_BUFFER_SIZE];
int binaryRecordBufferCount = 0;    // Number of records in buffer
static bool currentFileNameSet = false;
static bool initCurrentDateFilename = false;
const char* getCurrentDateFilename(); // Get the current date filename

void clearCurrentDateFilename() {
  currentFileNameSet = false;
}

// called on startup
void initCurrentFileName() {
  if (initCurrentDateFilename) {
    return;
  }
  if (debugPtr) {
    debugPtr->print(" initCurrentFileName: ");
  }
  if (!isRTCDateTimeSet()) {
    if (debugPtr) {
      debugPtr->println(" time NOT set yet");
    }
    return;
  }
  initCurrentDateFilename = true;
  currentFileNameSet = false;
  const char *filename =   getCurrentDateFilename(); // sets currentFileNameSet to true
  if (debugPtr) {
    debugPtr->println(filename);
  }
}


// Get current date for filename for NOW time
const char* getCurrentDateFilename() {
  static char filename[40] = {0}; // Static to ensure it's available after function returns
  if (debugPtr) {
    debugPtr->print(" getCurrentDateFilename ");
  }
  if (currentFileNameSet) {
    if (debugPtr) {
      debugPtr->print(" already set: "); debugPtr->println(filename);
    }
    return filename;
  }
  // else set it
  if (isRTCDateTimeSet()) {
    time_t now = time(nullptr);
    struct tm timeinfo;
    struct tm* timeinfoPtr = localtime_r(&now,&timeinfo);
    sprintf(filename, "/%04d%02d%02d", timeinfoPtr->tm_year + 1900, timeinfoPtr->tm_mon + 1, timeinfoPtr->tm_mday);
  } else {
    if (debugPtr) {
      debugPtr->println(" Time not set yet!!");
    }
  }
  if (debugPtr) {
    debugPtr->print(" set to: "); debugPtr->println(filename);
  }
  currentFileNameSet = true;
  return filename;
}


// Append binary buffer to daily file when buffer is full or at midnight
unsigned long appendBufferToFile() {
  unsigned long bufferCount = binaryRecordBufferCount; // for return
  START_LOCK(filesLock);
  // Get current date for filename
  const char* filename = getCurrentDateFilename(); // will get new file name for today date if clearCurrentDateFilename() has been called
  // clearCurrentDateFilename() is called after rotate day so that next save will save to new file
  if (*filename == '\0') {
    if (debugPtr) {
      debugPtr->println(" Buffer filename not set!!");
    }
    if (debugPtr) {
      debugPtr->print(" Time is set:"); debugPtr->println(isRTCDateTimeSet());
    }
    return 0;
  }
  // Open file for append
  File file = LittleFS.open(filename, FILE_APPEND);
  if (!file) {
#ifdef DEBUG
    if (debugPtr) {
      debugPtr->printf("Failed to open %s for appending. ", filename);
      debugPtr->println(" Create new file");
    }
#endif
    file = LittleFS.open(filename, FILE_WRITE);
    if (!file) {
#ifdef DEBUG
      if (debugPtr) {
        debugPtr->printf("Failed to open %s for writing\r\n", filename);
      }
#endif
      return 0;
    }
  }
  // Write all records from buffer
  for (int i = 0; i < binaryRecordBufferCount; i++) {
    // Write the binary record
    file.write((uint8_t*)&binaryRecordBuffer[i], sizeof(BinaryRecordStruct));
  }

  file.close();
#ifdef DEBUG
  if (debugPtr) {
    debugPtr->printf("// Appended %d records to %s\r\n", binaryRecordBufferCount, filename);
  }
#endif
  // Reset buffer after writing
  binaryRecordBufferCount = 0;
  END_LOCK(filesLock);
  return bufferCount;
}

// Add a new record to the circular buffer
void addBinaryRecord(const BinaryRecordStruct & record) {
  // Critical section - disable interrupts if needed for thread safety
  START_LOCK(filesLock);

  binaryRecordBuffer[binaryRecordBufferCount++] = record;
  if (binaryRecordBufferCount > MAX_BINARY_RECORD_BUFFER_SIZE) {
    appendBufferToFile(); // sets index to 0
  }
  END_LOCK(filesLock);
  // End critical section
}

// Get the number of records in the buffer
int getBinaryRecordCount() {
  return binaryRecordBufferCount;
}

// Get a record by index (0 is oldest, count-1 is newest)
BinaryRecordStruct getBinaryRecord(int index) {
  // Validate index
  if (index < 0 || index >= binaryRecordBufferCount) {
    // Return empty record
    BinaryRecordStruct emptyRecord = {0};
    return emptyRecord;
  }
  return binaryRecordBuffer[index];
}

const char* getHistoryFileName() {
  return HISTORY_FILE_NAME;
}

void print_Accumulated_kWh(const Accumulated_kWh_RecordStruct &accResults, Print * outPtr) {
  if (!outPtr) {
    return;
  }
  outPtr->print("  Import kWh: "); outPtr->print(accResults.import_kWhrs);
  outPtr->print("  Export kWh: "); outPtr->print(accResults.export_kWhrs);
  outPtr->print("  Off-peak kWh: "); outPtr->print(accResults.offPeak_kWhrs);
  outPtr->print("  HW kWh: "); outPtr->print(accResults.hotwater_kWhrs);
  outPtr->print("  HW fullyHeated: "); outPtr->print(accResults.fullyHeated ? "true" : "false");
  outPtr->print("  Total cost: $"); outPtr->println(accResults.cost, 6);
}


//struct  HistoryRecordStruct {
//  double export_kWhrs;
//  double import_kWhrs;
//  double hotwater_kWhrs;
//  double offPeak_kWhrs;
//  bool fullyHeated;
//  double export_kWhrs_yesterday;
//  double import_kWhrs_yesterday;
//  double hotwater_kWhrs_yesterday;
//  double offPeak_kWhrs_yesterday;
//  bool fullyHeated_yesterday;
//};

void printHistoryRecord(const HistoryRecordStruct & historyRecord, Print * outPtr) {
  if (!outPtr) {
    return;
  }
  handleHotWaterRemoteClient();
  outPtr->print(" export_kWhrs:");
  outPtr->print(historyRecord.export_kWhrs, 3);
  outPtr->println();
  outPtr->print(" import_kWhrs:");
  outPtr->print(historyRecord.import_kWhrs, 3);
  outPtr->println();
  outPtr->print(" hotwater_kWhrs:");
  outPtr->print(historyRecord.hotwater_kWhrs, 3);
  outPtr->println();
  outPtr->print(" offPeak_kWhrs:");
  outPtr->print(historyRecord.offPeak_kWhrs, 3);
  outPtr->println();
  outPtr->print(" fullyHeated:");
  outPtr->print(historyRecord.fullyHeated ? "true" : "false");
  outPtr->println();
  outPtr->print(" cost:");
  outPtr->print(historyRecord.cost);
  outPtr->println();

  outPtr->print(" export_kWhrs_yesterday:");
  outPtr->print(historyRecord.export_kWhrs_yesterday, 7);
  outPtr->println();
  outPtr->print(" import_kWhrs_yesterday:");
  outPtr->print(historyRecord.import_kWhrs_yesterday, 7);
  outPtr->println();
  outPtr->print(" hotwater_kWhrs_yesterday:");
  outPtr->print(historyRecord.hotwater_kWhrs_yesterday, 7);
  outPtr->println();
  outPtr->print(" offPeak_kWhrs_yesterday:");
  outPtr->print(historyRecord.offPeak_kWhrs_yesterday, 7);
  outPtr->println();
  outPtr->print(" fullyHeated_yesterday:");
  outPtr->print(historyRecord.fullyHeated_yesterday ? "true" : "false");
  outPtr->print(" cost:");
  outPtr->print(historyRecord.cost_yesterday);
  outPtr->println();
}


bool openForRead(File & file, const char * path);



bool openForOverwrite(File & file, const char* path) {
  file = LittleFS.open(path, FILE_WRITE);
  if (!file || file.isDirectory()) {
#ifdef DEBUG
    if (debugPtr) {
      debugPtr->println("- failed to open file for writing");
    }
#endif
    file.close();
    return false;
  }
  return true;
}

bool openForRead(File & file, const char * path) {
#ifdef DEBUG
  if (debugPtr) {
    debugPtr->printf("Reading file: %s\r\n", path);
  }
#endif

  file = LittleFS.open(path);
  if (!file || file.isDirectory()) {
#ifdef DEBUG
    if (debugPtr) {
      debugPtr->println("- failed to open file for reading");
    }
#endif
    file.close();
    return false;
  }
  return true;
}

void writeFile(const char * path, const char * message) {
#ifdef DEBUG
  if (debugPtr) {
    debugPtr->printf("Writing file: %s\r\n", path);
  }
#endif

  File file = LittleFS.open(path, FILE_WRITE);
  if (!file) {
#ifdef DEBUG
    if (debugPtr) {
      debugPtr->println("- failed to open file for writing");
    }
#endif
    return;
  }
  if (file.print(message)) {
#ifdef DEBUG
    if (debugPtr) {
      debugPtr->println("- file written");
    }
#endif
  } else {
#ifdef DEBUG
    if (debugPtr) {
      debugPtr->println("- write failed");
    }
#endif
  }
  file.close();
}

void listRootDir(Print * out) {
  if (out) {
    listDir("/", out);
  }
}

bool printDataFile(const char* fileName, Print * outPtr) {
  if (!outPtr) {
    return true;
  }
  handleHotWaterRemoteClient();
  char fullPath[32];
  sprintf(fullPath, "/%s", fileName);

  cSF(sfTime, 80); // need at most 64 bytes for time_t long long

#ifdef DEBUG
  if (debugPtr) {
    debugPtr->printf("// Reading data file: %s\r\n", fullPath);
  }
#endif

  // Open file for reading
  File file = LittleFS.open(fullPath, FILE_READ);
  if (!file) {
#ifdef DEBUG
    if (debugPtr) {
      debugPtr->printf("// Failed to open %s for reading\r\n", fullPath);
    }
#endif
    outPtr->printf("// Failed to open %s\r\n", fullPath);
    return false;
  }

  // Print CSV header, first line is file name for reloading
  // parser for //FileName: and rest is the filename
  outPtr->print("//FileName:"); outPtr->println(fullPath);
  outPtr->println("// local date,local time,utc tv_sec,utc tv_usec,power_w,hw_w,state,hw_triac,hw_boost,hw_kwh");

  // Read binary records from file and convert to CSV
  BinaryRecordStruct record;
  size_t recordSize = sizeof(BinaryRecordStruct);
  char dateStr[40];  // YYYY/MM/DD + null // compiler complains if size <40 ??
  char timeStr[9];   // HH:MM:SS + null

  while (file.read((uint8_t*)&record, recordSize) == recordSize) {
    handleHotWaterRemoteClient();
    time_t localTime;
    struct tm timeinfo;

    // Convert UTC to local time
    localTime = record.utc_time;
    localtime_r(&localTime, &timeinfo);

    // Format date and time
    sprintf(dateStr, "%04d/%02d/%02d",
            timeinfo.tm_year + 1900,
            timeinfo.tm_mon + 1,
            timeinfo.tm_mday);

    sprintf(timeStr, "%02d:%02d:%02d",
            timeinfo.tm_hour,
            timeinfo.tm_min,
            timeinfo.tm_sec);

    // Print record as CSV
    while (outPtr->availableForWrite() < 60) {
      handleHotWaterRemoteClient();
    }
    outPtr->print(dateStr);  outPtr->print(',');
    outPtr->print(timeStr);  outPtr->print(',');
    sfTime.clear();
    sfTime.print((int64_t)record.utc_time);
    outPtr->print(sfTime); outPtr->print(',');
    sfTime.clear();
    sfTime.print((uint32_t)record.utc_usec);
    outPtr->print(sfTime); outPtr->print(',');

    outPtr->print(record.power_w);  outPtr->print(',');
    outPtr->print(record.hw_w);  outPtr->print(',');
    outPtr->print(record.state);  outPtr->print(',');
    outPtr->print(record.hw_triac);  outPtr->print(',');
    outPtr->print(record.hw_boost);  outPtr->print(',');
    outPtr->print(record.hw_kwh, 3);  outPtr->println();
  }
  // terminate file
  outPtr->println("// EOF");
  file.close();
  return true;
}

void printBinaryRecord(const BinaryRecordStruct & record, Print * outPtr) {
  if (!outPtr) {
    return;
  }
  outPtr->print(record.utc_time);  outPtr->print(',');
  outPtr->print(record.power_w);  outPtr->print(',');
  outPtr->print(record.hw_w);  outPtr->print(',');
  outPtr->print(record.state);  outPtr->print(',');
  outPtr->print(record.hw_triac);  outPtr->print(',');
  outPtr->print(record.hw_boost); outPtr->print(',');
  outPtr->print(record.hw_kwh, 3);  outPtr->println();
}

// Parse a CSV line into a BinaryRecordStruct record
// Format expected: local date,local time,unix seconds utc,power_w,hw_w,state,hw_triac,hw_boost,hw_kwh
// Returns true if parsing was successful, false otherwise
static bool parseCSVLineToBinaryRecord(SafeString & sfLine, BinaryRecordStruct & record, Print * outPtr) {
  cSF(sfToken, 80);
  bool validRecord = true;

  // Skip local date
  if (!sfLine.firstToken(sfToken, ',')) {
    validRecord = false;
  }

  // Skip local time
  if (validRecord && !sfLine.nextToken(sfToken, ',')) {
    validRecord = false;
  }

  // Parse UTC seconds
  if (validRecord && sfLine.nextToken(sfToken, ',')) {
    // time_t utcTime; is long long i.e. int64_t on ESP32
    int64_t result = 0; // on esp32
    if (!sfToken.toInt64_t(result)) {
      validRecord = false;
      if (outPtr) {
        outPtr->print("Error parsing UTC time: ");
        outPtr->println(sfToken);
      }
    } else {
      record.utc_time = result;
    }
  } else {
    validRecord = false;
  }

  // Parse UTC usec
  if (validRecord && sfLine.nextToken(sfToken, ',')) {
    // time_t utcTime; is long long i.e. int64_t on ESP32
    int64_t result = 0; // on esp32
    if (!sfToken.toInt64_t(result)) {
      validRecord = false;
      if (outPtr) {
        outPtr->print("Error parsing UTC usec: ");
        outPtr->println(sfToken);
      }
    } else {
      record.utc_usec = result;
    }
  } else {
    validRecord = false;
  }
  // Parse power_w
  if (validRecord && sfLine.nextToken(sfToken, ',')) {
    int32_t powerW;
    if (!sfToken.toLong(powerW)) {
      validRecord = false;
      if (outPtr) {
        outPtr->print("Error parsing power_w: ");
        outPtr->println(sfToken);
      }
    } else {
      record.power_w = powerW;
    }
  } else {
    validRecord = false;
  }

  // Parse hw_w
  if (validRecord && sfLine.nextToken(sfToken, ',')) {
    int32_t hwW;
    if (!sfToken.toLong(hwW)) {
      validRecord = false;
      if (outPtr) {
        outPtr->print("Error parsing hw_w: ");
        outPtr->println(sfToken);
      }
    } else {
      record.hw_w = hwW;
    }
  } else {
    validRecord = false;
  }

  // Parse state
  if (validRecord && sfLine.nextToken(sfToken, ',')) {
    int state;
    if (!sfToken.toInt(state)) {
      validRecord = false;
      if (outPtr) {
        outPtr->print("Error parsing state: ");
        outPtr->println(sfToken);
      }
    } else {
      record.state = (uint8_t)state;
    }
  } else {
    validRecord = false;
  }

  // Parse hw_triac
  if (validRecord && sfLine.nextToken(sfToken, ',')) {
    int hwTriac;
    if (!sfToken.toInt(hwTriac)) {
      validRecord = false;
      if (outPtr) {
        outPtr->print("Error parsing hw_triac: ");
        outPtr->println(sfToken);
      }
    } else {
      record.hw_triac = (uint8_t)hwTriac;
    }
  } else {
    validRecord = false;
  }

  // Parse hw_boost
  if (validRecord && sfLine.nextToken(sfToken, ',')) {
    int hwBoost;
    if (!sfToken.toInt(hwBoost)) {
      validRecord = false;
      if (outPtr) {
        outPtr->print("Error parsing hw_boost: ");
        outPtr->println(sfToken);
      }
    } else {
      record.hw_boost = (uint8_t)hwBoost;
    }
  } else {
    validRecord = false;
  }

  // Parse hw_kwh (may be missing in older files)
  if (validRecord && sfLine.nextToken(sfToken, ',')) {
    double hwKwh;
    if (!sfToken.toDouble(hwKwh)) {
      validRecord = false;
      if (outPtr) {
        outPtr->print("Error parsing hw_kwh: ");
        outPtr->println(sfToken);
      }
    } else {
      record.hw_kwh = hwKwh;
    }
  } else {
    // If hw_kwh is missing (older files), set to 0
    record.hw_kwh = 0.0;
  }

  // Check no more fields after hw_kwh
  if (validRecord && sfLine.nextToken(sfToken, ',')) {
    if (outPtr) {
      outPtr->print("Error parsing too many fields: ");
      outPtr->println(sfToken);
    }
    validRecord = false;
  }

  return validRecord;
}


bool readInCSVData(char* tempInFileName, int &lineCounter, Stream & input, Print * outPtr) {

  BinaryRecordStruct csvRecord = {0};

  // Create SafeStringReader for parsing CSV
  createSafeStringReader(sfCsvReader, LINE_LEN, "\r\n"); // \r will be trimmed below
  cSF(sfLine, LINE_LEN);
  cSF(sfToken, 80);
  int csvRecordCounter = 0;
  millisDelay readTimeout;
  unsigned long READ_TIMEOUT_ms = 4000;

  // sfCsvReader.returnEmptyTokens(false);
  sfCsvReader.connect(input);
  bool endOfCsv = false;

  // Open temporary file for writing new records
  File tempInFile = LittleFS.open(tempInFileName, FILE_WRITE);
  if (!tempInFile) {
    if (outPtr) {
      outPtr->print("Error creating temporary file: ");
      outPtr->println(tempInFileName);
    }
    return false;
  }

  cSF(sfProgressMsg, 120);
  readTimeout.start(READ_TIMEOUT_ms);
  // read in file data
  // parse and save
  while (!endOfCsv) {
    if (readTimeout.justFinished()) {
      endOfCsv = true;
      if (outPtr) {
        outPtr->println("End Of File (read timed out)");
      }
      tempInFile.close();
      sfCsvReader.end();
      break;
    }
    handleHotWaterRemoteClient(); // call often
    if (sfCsvReader.read()) {
      lineCounter++;
      readTimeout.restart();
      sfCsvReader.trim();
      if (outPtr) {
        // outPtr->println(sfCsvReader);
      }

      // Skip comment lines and empty lines
      if (sfCsvReader.startsWith("//")) {
        if (sfCsvReader.indexOf("EOF") >= 0) {
          endOfCsv = true;
          if (outPtr) {
            outPtr->println("End Of File (//EOF line)");
          }
          tempInFile.close();
          break;
        }
        continue;
      }

      // Process data line
      sfLine = sfCsvReader;

      // Parse the CSV line into a BinaryRecordStruct
      if (parseCSVLineToBinaryRecord(sfLine, csvRecord, outPtr)) {
        tempInFile.write((uint8_t*)&csvRecord, sizeof(BinaryRecordStruct));
        csvRecordCounter++;
      } else {
        if (outPtr) {
          outPtr->println(" Invalid record abandoning merge");
          outPtr->println(sfCsvReader);
          outPtr->print(" On line no:"); outPtr->println(lineCounter);
        }
        // Close the temporary file
        tempInFile.close();
        LittleFS.remove(tempInFileName);
        sfCsvReader.end();
        return false;
      }
      if (outPtr && (csvRecordCounter % 1000 == 0) && (csvRecordCounter > 0)) {
        if (outPtr) {
          sfProgressMsg.clear();
          sfProgressMsg.print(" CSV line:"); sfProgressMsg.print(csvRecordCounter, 0, 5); sfProgressMsg.println();
          outPtr->print(sfProgressMsg);
          //outPtr->println(sfCsvReader);
        }
      }
    }
  }
  sfCsvReader.end();
  tempInFile.close();
  if (outPtr) {
    sfProgressMsg.clear();
    sfProgressMsg.print(" Read in "); sfProgressMsg.print(csvRecordCounter, 0, 5); sfProgressMsg.println(" records");
    outPtr->print(sfProgressMsg);
  }
  return true;
}

// Reads a CSV data stream and saves the data to LittleFS
// Format:
// First line has file name after //FileName:  e.g. //FileName:filename
// Second line has field names after // e.g. // local date,local time,unix seconds utc,power_w,hw_w,state,hw_triac,hw_boost
// Following lines have data.
//
// If the file already exists, merges with existing file based on unix seconds,
// keeping any existing file data and adding missing data in utc seconds order
//
// Uses a streaming approach assuming both input and existing files are already in UTC time order
bool readDataStream(Stream & input, Print * outPtr) {
  // no need to write to file for 40mins now
  // should really use locks below but should not be needed as no writes to file need for 40mins

  cSF(sfProgressMsg, 120);

  int lineCounter = 0;

  char fileName[32] = {0};
  cSFA(sfFileName, fileName);

  // Create SafeStringReader for parsing CSV
  createSafeStringReader(sfCsvReader, LINE_LEN, "\r\n");
  cSF(sfLine, LINE_LEN);
  cSF(sfToken, 80);

  sfCsvReader.connect(input);


  // Variables to track the current record from each source
  BinaryRecordStruct csvRecord = {0};
  BinaryRecordStruct fileRecord = {0};


  bool fileNameFound = false;
  // Read file name from first line (//FileName:filename)
  while (!fileNameFound) {
    handleHotWaterRemoteClient(); // call often

    if (sfCsvReader.read()) {
      lineCounter++;
      sfCsvReader.trim();
      if (sfCsvReader.startsWith("//FileName:")) {
        sfCsvReader.substring(sfFileName, 11); // Extract after "//FileName:"
        sfFileName.trim();
        if (sfFileName.length() > 0) {
          fileNameFound = true;
          if (outPtr) {
            outPtr->print("Found filename: ");
            outPtr->println(fileName);
          }
          break;
        } else {
          if (outPtr) {
            outPtr->println(sfCsvReader);
            outPtr->println("Missing filename ");
          }
          sfCsvReader.end();
          return false;
        }
      } else if (sfCsvReader.startsWith("//") || sfCsvReader.length() == 0) {
        // Skip comment lines ends with an empty
      } else {
        if (outPtr) {
          outPtr->print("Missing filename ");
          outPtr->println(" Invalid record abandoning merge");
          outPtr->println(sfCsvReader);
          outPtr->print(" On line no:"); outPtr->println(lineCounter);
        }
        sfCsvReader.end();
        return false;
      }
    }
  }
  sfCsvReader.end();

  handleHotWaterRemoteClient(); // call often

  // Prepare file paths
  char tempInFileName[40];
  sprintf(tempInFileName, "%s.in", fileName);
  int startLineCounter = lineCounter;
  if (!readInCSVData(tempInFileName, lineCounter, input,  outPtr)) {
    return false;
  }

  // Check if the target file already exists
  bool fileExists = LittleFS.exists(fileName);

  if (!fileExists) {
    // No existing file just rename tempInFileName
    handleHotWaterRemoteClient(); // call often

    // Replace the original file with the merged file
    if (LittleFS.rename(tempInFileName, fileName)) {
      if (outPtr) {
        outPtr->print("No existing file. Successfully loaded data. Total csv records: ");
        outPtr->println(lineCounter - startLineCounter);
      }
    } else {
      if (outPtr) {
        outPtr->println("Error replacing original file with merged file");
      }
      return false;
    }

    handleHotWaterRemoteClient(); // call often
    return true;
  }

  // else file exists
  // If the original file exists, we'll need to merge
  // Open temporary file for writing new records
  // Prepare file paths
  char tempFileName[40];
  sprintf(tempFileName, "%s.new", fileName);

  File tempFile = LittleFS.open(tempFileName, FILE_WRITE);
  if (!tempFile) {
    if (outPtr) {
      outPtr->print("Error creating temporary file: ");
      outPtr->println(tempFileName);
    }
    LittleFS.remove(tempInFileName);
    return false;
  }
  if (outPtr) {
    outPtr->print(" created Temp Input file:"); outPtr->println(tempInFileName);
  }

  File origFile = LittleFS.open(fileName, FILE_READ);
  if (!origFile) {
    if (outPtr) {
      outPtr->print("Error opening existing file: ");
      outPtr->println(fileName);
    }
    tempFile.close();
    LittleFS.remove(tempFileName);
    LittleFS.remove(tempInFileName);
    return false;
  }
  if (outPtr) {
    outPtr->print(" opened existing file:"); outPtr->println(fileName);
  }
  File tempInFile = LittleFS.open(tempInFileName, FILE_READ);
  if (!tempInFile) {
    if (outPtr) {
      outPtr->print("Error opening existing file: ");
      outPtr->println(tempInFileName);
    }
    origFile.close();
    tempFile.close();
    LittleFS.remove(tempFileName);
    LittleFS.remove(tempInFileName);
    return false;
  }
  if (outPtr) {
    outPtr->print(" opened existing Input file:"); outPtr->println(tempInFileName);
  }

  handleHotWaterRemoteClient(); // call often

  bool haveCsvRecord = false;
  bool haveFileRecord = false;
  int mergedCount = 0;
  int mergedCountCsv = 0;
  int mergedCountFile = 0;
  int dupCsvCount = 0;

  // Read first record from existing file
  haveFileRecord = (origFile.read((uint8_t*)&fileRecord, sizeof(BinaryRecordStruct)) == sizeof(BinaryRecordStruct));
  haveCsvRecord = (tempInFile.read((uint8_t*)&csvRecord, sizeof(BinaryRecordStruct)) == sizeof(BinaryRecordStruct));
  //printBinaryRecord(csvRecord,outPtr);
  if (outPtr) {
    //    outPtr->print(" haveFileRecord:"); outPtr->println(haveFileRecord);
    //    outPtr->print(" haveCsvRecord:"); outPtr->println(haveCsvRecord);
  }

  // Process all records from both sources
  while (haveCsvRecord || haveFileRecord) {

    handleHotWaterRemoteClient(); // call often

    // Determine which record to write next
    if (haveCsvRecord && haveFileRecord) {
      // Both sources have records, choose the earlier one
      if (csvRecord.utc_time < fileRecord.utc_time) {
        // CSV record is earlier
        tempFile.write((uint8_t*)&csvRecord, sizeof(BinaryRecordStruct));
        mergedCount++;
        mergedCountCsv++;
        haveCsvRecord = (tempInFile.read((uint8_t*)&csvRecord, sizeof(BinaryRecordStruct)) == sizeof(BinaryRecordStruct));
        //printBinaryRecord(csvRecord,outPtr);

      } else if (fileRecord.utc_time < csvRecord.utc_time) {
        // File record is earlier
        tempFile.write((uint8_t*)&fileRecord, sizeof(BinaryRecordStruct));
        mergedCount++;
        mergedCountFile++;
        // Read next record from file
        haveFileRecord = (origFile.read((uint8_t*)&fileRecord, sizeof(BinaryRecordStruct)) == sizeof(BinaryRecordStruct));
      } else {
        // Records have same timestamp, prefer file record and skip CSV record
        tempFile.write((uint8_t*)&fileRecord, sizeof(BinaryRecordStruct));
        mergedCount++;
        mergedCountFile++;
        dupCsvCount++;
        haveCsvRecord = (tempInFile.read((uint8_t*)&csvRecord, sizeof(BinaryRecordStruct)) == sizeof(BinaryRecordStruct));
        //printBinaryRecord(csvRecord,outPtr);
        // Read next record from file
        haveFileRecord = (origFile.read((uint8_t*)&fileRecord, sizeof(BinaryRecordStruct)) == sizeof(BinaryRecordStruct));
      }
    } else if (haveCsvRecord) {
      // Only CSV has records left
      tempFile.write((uint8_t*)&csvRecord, sizeof(BinaryRecordStruct));
      mergedCount++;
      mergedCountCsv++;
      haveCsvRecord = (tempInFile.read((uint8_t*)&csvRecord, sizeof(BinaryRecordStruct)) == sizeof(BinaryRecordStruct));
      // printBinaryRecord(csvRecord,outPtr);

    } else if (haveFileRecord) {
      // Only file has records left
      tempFile.write((uint8_t*)&fileRecord, sizeof(BinaryRecordStruct));
      mergedCount++;
      mergedCountFile++;
      // Read next record from file
      haveFileRecord = (origFile.read((uint8_t*)&fileRecord, sizeof(BinaryRecordStruct)) == sizeof(BinaryRecordStruct));
    }
    // Progress reporting
    if (outPtr && (mergedCount % 1000 == 0) && (mergedCount > 0)) {
      sfProgressMsg.clear();
      sfProgressMsg.print(" Merged:");
      sfProgressMsg.print(mergedCount, 0, 5);
      sfProgressMsg.print("      File records:");
      sfProgressMsg.print(mergedCountFile, 0, 5);
      sfProgressMsg.print("      CSV records:");
      sfProgressMsg.print(mergedCountCsv, 0, 5);
      sfProgressMsg.print("      Discarded CSV dups:");
      sfProgressMsg.print(dupCsvCount, 0, 5);
      sfProgressMsg.println();
      outPtr->print(sfProgressMsg);
    }
  }

  // Progress reporting
  if (outPtr ) {
    sfProgressMsg.clear();
    sfProgressMsg.print(" Merged:");
    sfProgressMsg.print(mergedCount, 0, 5);
    sfProgressMsg.print("      File records:");
    sfProgressMsg.print(mergedCountFile, 0, 5);
    sfProgressMsg.print("      CSV records:");
    sfProgressMsg.print(mergedCountCsv, 0, 5);
    sfProgressMsg.print("      Discarded CSV dups:");
    sfProgressMsg.print(dupCsvCount, 0, 5);
    sfProgressMsg.println();
    outPtr->print(sfProgressMsg);
  }

  handleHotWaterRemoteClient(); // call often
  // Close the temporary Input file
  tempInFile.close();
  LittleFS.remove(tempInFileName);

  // Close original file
  origFile.close();

  // Close the temporary file
  tempFile.close();

  if (outPtr) {
    outPtr->println("Closed files");
  }

  // Replace the original file with the merged file
  if (LittleFS.remove(fileName) && LittleFS.rename(tempFileName, fileName)) {
    if (outPtr) {
      outPtr->print("Successfully merged data. Total records: ");
      outPtr->println(mergedCount);
    }
  } else {
    if (outPtr) {
      outPtr->println("Error replacing original file with merged file");
    }
    return false;
  }
  return true;
}
