// Chart.js for SolarHotWaterMonitor_V2
// Wait for DOM to be ready before any processing
if (document.readyState === 'loading') {
    console.log('Document still loading - waiting for ready event');
    document.addEventListener('DOMContentLoaded', function() {
        console.log('DOM content loaded via readyState check');
    });
} else {
    console.log('Document already loaded at script execution');
}

let allData = null; // Store the full dataset
let availableDates = []; // Store available dates from server
let chartInstances = []; // Store chart instances for proper disposal
let dataRequestInProgress = false; // Track if a data request is currently in progress

// Initialize on page load
// Track auto-refresh state
let autoRefreshEnabled = true;

document.addEventListener('DOMContentLoaded', function() {
    console.log('DOM content loaded - initializing chart');
    
    // Fetch available dates and populate dropdown
    fetchAvailableDates();
    
    // Initial values will be set in updateTimesForDateSelection() when dates are selected
    
    // Add event listener for chart updates
    document.getElementById('updateChart').addEventListener('click', updateChart);
    
    // Add event listener for auto-refresh toggle
    const toggleAutoRefreshBtn = document.getElementById('toggleAutoRefresh');
    if (toggleAutoRefreshBtn) {
        toggleAutoRefreshBtn.addEventListener('click', toggleAutoRefresh);
    }
    
    // Add event listeners to date selects to toggle Auto Refresh visibility
    const startDateSelect = document.getElementById('startDateSelect');
    const endDateSelect = document.getElementById('endDateSelect');
    
    function updateAutoRefreshVisibility() {
        // Get today's date for setting time values
        const today = new Date();
        
        // Auto-refresh is only available when both dates are literally "today"
        // This is intentionally strict to control auto-refresh behavior
        const bothDatesAreToday = startDateSelect.value === 'today' && endDateSelect.value === 'today';
        
        // For time setting, we need to know if the end date is today (in any form)
        // Get today's date in YYYYMMDD format for comparison with numeric dates
        const todayStr = today.getFullYear().toString() +
                         padNumber(today.getMonth() + 1) +
                         padNumber(today.getDate());
        
        // Check if end date is literally "today" (not a specific date that happens to be today)
        const endDateIsToday = endDateSelect.value === 'today';
        
        const autoRefreshCell = document.getElementById('autoRefreshLabelCell');
        const autoRefreshButtonCell = document.querySelector('td button#toggleAutoRefresh').parentNode.parentNode;
        
        if (bothDatesAreToday) {
            // Show Auto Refresh button cell always when both dates are today
            if (autoRefreshButtonCell) autoRefreshButtonCell.style.display = '';
            
            // Show Auto Refresh label only if enabled
            if (autoRefreshCell) {
                autoRefreshCell.style.display = autoRefreshEnabled ? '' : 'none';
            }
            
            // Re-enable auto-refresh if it was enabled
            if (autoRefreshEnabled) {
                // Show auto-refresh indicator
                const statusIndicator = document.getElementById('autoRefreshStatus');
                if (statusIndicator) {
                    statusIndicator.style.display = 'inline';
                }
                
                // Get the current start time - preserve it for auto-refresh
                const startTimeInput = document.getElementById('startTime');
                const currentStartTime = startTimeInput ? startTimeInput.value : "00:00";
                
                // Get end time - we no longer automatically update it when changing start date
                const endTimeInput = document.getElementById('endTime');
                const endTime = endTimeInput ? endTimeInput.value : "23:59";
                
                console.log(`Loading today's data with preserved time range: ${currentStartTime} to ${endTime}`);
                
                // Use direct fetch with current time parameters instead of loadTodayData to preserve start time
                fetchDataWithDateRange(todayStr, todayStr, currentStartTime, endTime);
                
                // Set up auto-refresh for future updates
                setAutoRefresh();
            } else {
                // When auto-refresh is disabled, we no longer update end time to current time
                // This preserves whatever time the user has set in the input field
            }
        } else {
            // Hide Auto Refresh elements
            if (autoRefreshCell) autoRefreshCell.style.display = 'none';
            if (autoRefreshButtonCell) autoRefreshButtonCell.style.display = 'none';
            
            // Stop any active auto-refresh
            if (window.refreshInterval) {
                clearInterval(window.refreshInterval);
                window.refreshInterval = null;
                
                // Hide the status indicator
                const statusIndicator = document.getElementById('autoRefreshStatus');
                if (statusIndicator) {
                    statusIndicator.style.display = 'none';
                }
            }
            
            // No longer update the end time when the date selection changes
            // This implements the requirement to not change the end time when start date changes
        }
    }
    
    // Function to update time inputs when date selections change - now only used for initial setup
    function updateTimesForDateSelection() {
        // Only set the end time on initial setup
        // This is needed for the initial chart creation
        const endDateValue = endDateSelect.value;
        const today = new Date();
        
        // Set end time based on selection
        const endTimeInput = document.getElementById('endTime');
        if (endTimeInput) {
            // Check if end date is literally "today"
            if (endDateValue === 'today') {
                if (autoRefreshEnabled) {
                    // Set end time to current time only when auto-refresh is enabled
                    endTimeInput.value = padNumber(today.getHours()) + ':' + padNumber(today.getMinutes());
                    console.log('Initial setup: Set end time to current time for today with auto-refresh enabled:', endTimeInput.value);
                } else {
                    // Use 23:59 when auto-refresh is disabled
                    endTimeInput.value = '23:59';
                    console.log('Initial setup: Set end time to 23:59 for today with auto-refresh disabled');
                }
            } else {
                // Set end time to 23:59 for non-today dates
                endTimeInput.value = '23:59';
                console.log('Initial setup: Set end time to 23:59 for non-today date');
            }
        }
    }
    
    // Add change event listeners to date selects
    if (startDateSelect) {
        startDateSelect.addEventListener('change', function() {
            // Always reset start time to 00:00 when start date changes
            const startTimeInput = document.getElementById('startTime');
            if (startTimeInput) {
                startTimeInput.value = '00:00';
                console.log('Reset start time to 00:00 when changing start date');
            }
            
            updateAutoRefreshVisibility();
            // Do not update end time when start date changes per new requirements
        });
    }
    
    if (endDateSelect) {
        endDateSelect.addEventListener('change', function() {
            updateAutoRefreshVisibility();
            updateEndTimeOnly(); // Only update end time when end date changes
        });
    }
    
    // Function to update only end time based on selection
    function updateEndTimeOnly() {
        const endDateValue = endDateSelect.value;
        
        // Get today's date in YYYYMMDD format for comparison
        const today = new Date();
        const todayStr = today.getFullYear().toString() +
                     padNumber(today.getMonth() + 1) +
                     padNumber(today.getDate());
        
        // Set end time based on selection
        const endTimeInput = document.getElementById('endTime');
        if (endTimeInput) {
            // Check if end date is literally "today" (not a specific date that happens to be today)
            if (endDateValue === 'today') {
                if (autoRefreshEnabled) {
                    // Set end time to current time only when both end date is "today" AND auto-refresh is enabled
                    endTimeInput.value = padNumber(today.getHours()) + ':' + padNumber(today.getMinutes());
                    console.log('Set end time to current time for today selection with auto-refresh enabled:', endTimeInput.value);
                } else {
                    // Use 23:59 when end date is today but auto-refresh is disabled
                    endTimeInput.value = '23:59';
                    console.log('Set end time to 23:59 for today selection with auto-refresh disabled');
                }
            } else {
                // Set end time to 23:59 for non-today dates and specific dates
                endTimeInput.value = '23:59';
                console.log('Reset end time to 23:59 for non-today date');
            }
        }
    }
    
    // Initial chart load with today's data
    loadTodayData();
    
    // Set up auto-refresh for today's data
    if (autoRefreshEnabled) {
        // Show auto-refresh indicator
        const statusIndicator = document.getElementById('autoRefreshStatus');
        if (statusIndicator) {
            statusIndicator.style.display = 'inline';
        }
        
        // Set up the refresh interval
        setAutoRefresh();
    }
    
    // Set flag to indicate DOM is loaded
    initialLoadComplete = true;
    
    // Make sure times are set correctly on initial load
    updateTimesForDateSelection();
});

// Toggle auto-refresh on/off
function toggleAutoRefresh() {
    const toggleBtn = document.getElementById('toggleAutoRefresh');
    const statusIndicator = document.getElementById('autoRefreshStatus');
    const autoRefreshCell = document.getElementById('autoRefreshLabelCell');
    
    if (autoRefreshEnabled) {
        // Disable auto-refresh
        autoRefreshEnabled = false;
        
        // Update button text
        if (toggleBtn) {
            toggleBtn.textContent = 'Enable Auto-refresh';
        }
        
        // Clear any existing interval
        if (window.refreshInterval) {
            clearInterval(window.refreshInterval);
            window.refreshInterval = null;
        }
        
        // Hide status indicator
        if (statusIndicator) {
            statusIndicator.style.display = 'none';
        }
        
        // Hide Auto Refresh label
        if (autoRefreshCell) {
            autoRefreshCell.style.display = 'none';
        }
        
        console.log('Auto-refresh disabled');
    } else {
        // Enable auto-refresh
        autoRefreshEnabled = true;
        
        // Update button text
        if (toggleBtn) {
            toggleBtn.textContent = 'Disable Auto-refresh';
        }
        
        // Get the current start time before changing anything
        const startTimeInput = document.getElementById('startTime');
        const currentStartTime = startTimeInput ? startTimeInput.value : "00:00";
        
        // Reset date selection to today
        const startDateSelect = document.getElementById('startDateSelect');
        const endDateSelect = document.getElementById('endDateSelect');
        if (startDateSelect) startDateSelect.value = 'today';
        if (endDateSelect) endDateSelect.value = 'today';
        
        // Update only the end time
        const today = new Date();
        const endTimeInput = document.getElementById('endTime');
        if (endTimeInput) {
            endTimeInput.value = padNumber(today.getHours()) + ':' + padNumber(today.getMinutes());
        }
        
        // Show Auto Refresh label
        if (autoRefreshCell) {
            autoRefreshCell.style.display = '';
        }
        
        // Format today's date as YYYYMMDD
        const todayStr = today.getFullYear().toString() +
                       padNumber(today.getMonth() + 1) +
                       padNumber(today.getDate());
        
        // Use direct fetch instead of loadTodayData to preserve start time
        fetchDataWithDateRange(todayStr, todayStr, currentStartTime, endTimeInput ? endTimeInput.value : "23:59");
        
        // Set up auto-refresh
        setAutoRefresh();
        
        console.log('Auto-refresh enabled, reset to today\'s data');
    }
}

// Set a flag to track if initial load is complete
let initialLoadComplete = false;

// Force loadTodayData after a delay
function forceLoadData() {
    console.log('Force loading data');
    
    // Skip if a data request is already in progress
    if (dataRequestInProgress) {
        console.log('Data request already in progress, skipping forceLoadData');
        return;
    }
    
    // Skip if charts are already loaded
    if (chartInstances.length > 0) {
        console.log('Charts already loaded, skipping forceLoadData');
        return;
    }
    
    // Wait for DOM to be fully ready
    if (document.readyState === 'complete') {
        console.log('Document ready, loading today\'s data');
        // Clear any existing charts first
        if (chartInstances.length > 0) {
            console.log('Clearing existing chart instances');
            for (let i = 0; i < chartInstances.length; i++) {
                chartInstances[i].destroy();
            }
            chartInstances = [];
        }
        
        // Get end date select to check if it's set to today
        const endDateSelect = document.getElementById('endDateSelect');
        const startDateSelect = document.getElementById('startDateSelect');
        
        // Get today's date in YYYYMMDD format for comparison
        const now = new Date();
        const todayStr = now.getFullYear().toString() +
                         padNumber(now.getMonth() + 1) +
                         padNumber(now.getDate());
        
        // Check if start date is "today"
        const startIsToday = startDateSelect && startDateSelect.value === 'today';
        
        // Check if end date is literally "today" (not a specific date that happens to be today)
        const endIsToday = endDateSelect && endDateSelect.value === 'today';
        
        // Don't reset start time - preserve user's choice
        const startTimeInput = document.getElementById('startTime');
        
        // Only update end time if end date is today AND auto-refresh is enabled
        // Per requirements, we only update end time when the end date changes, not when we force load data
        // We leave this code here for compatibility with auto-refresh functionality
        
        // Get current times for data fetching
        const startTime = startTimeInput ? startTimeInput.value : "00:00";
        const endTime = endTimeInput ? endTimeInput.value : "23:59";
        
        // Format today's date as YYYYMMDD
        const todayFormatted = now.getFullYear().toString() +
                     padNumber(now.getMonth() + 1) +
                     padNumber(now.getDate());
        
        console.log(`Loading data with time range: ${startTime} to ${endTime}`);
        
        // Use fetchDataWithDateRange directly to preserve start time
        fetchDataWithDateRange(todayFormatted, todayFormatted, startTime, endTime);
        
        // Set up auto-refresh only if it's enabled
        if (autoRefreshEnabled) {
            console.log('Setting up auto-refresh from forceLoadData (auto-refresh enabled)');
            
            // Show auto-refresh indicator right away
            const statusIndicator = document.getElementById('autoRefreshStatus');
            if (statusIndicator) {
                statusIndicator.style.display = 'inline';
            }
            
            // Set up the auto-refresh interval
            setAutoRefresh();
        } else {
            console.log('Auto-refresh is disabled, not setting up refresh interval from forceLoadData');
        }
    } else {
        // If document not ready yet, wait a bit more
        console.log('Document not ready, waiting...');
        setTimeout(forceLoadData, 500);
    }
}

// Ensure chart loads on refresh - more robust approach
window.addEventListener('load', function() {
    console.log('Window load event fired');
    
    // Give a small delay to ensure DOM is ready
    setTimeout(function() {
        // Skip if data request is already in progress
        if (dataRequestInProgress) {
            console.log('Data request already in progress, skipping window load handler');
            return;
        }
        
        // If normal load didn't work, force it now
        if (!chartInstances || chartInstances.length === 0) {
            console.log('No chart instances found after window load event, forcing load');
            forceLoadData();
        } else {
            console.log('Charts already loaded properly');
        }
        
        // Get the date select elements
        const startDateSelect = document.getElementById('startDateSelect');
        const endDateSelect = document.getElementById('endDateSelect');
        
        // Check if they exist
        if (!startDateSelect || !endDateSelect) {
            return;
        }
        
        // Auto-refresh is only available when both dates are literally "today"
        // This is intentionally strict to control auto-refresh behavior
        const bothDatesAreToday = startDateSelect.value === 'today' && endDateSelect.value === 'today';
                     
        // Set up auto-refresh only if we're showing today's data AND auto-refresh is enabled
        if (bothDatesAreToday && autoRefreshEnabled) {
            
            // Ensure the auto-refresh indicator is displayed
            const statusIndicator = document.getElementById('autoRefreshStatus');
            if (statusIndicator) {
                statusIndicator.style.display = 'inline';
            }
            
            console.log('Setting up auto-refresh from window load event (auto-refresh enabled)');
            setAutoRefresh();
        } else if (bothDatesAreToday) {
            console.log('Today view detected but auto-refresh is disabled, not setting up auto-refresh');
        }
    }, 300);
});

// Set up automatic refresh of today's data
function setAutoRefresh() {
    // Check if auto-refresh is disabled
    if (!autoRefreshEnabled) {
        console.log('Auto-refresh is disabled, not setting up interval');
        return;
    }
    
    // Always clear any existing interval to avoid duplicates
    if (window.refreshInterval) {
        console.log('Clearing existing refresh interval');
        clearInterval(window.refreshInterval);
        window.refreshInterval = null;
    }
    
    console.log('Setting up auto-refresh interval');
    
    // Show auto-refresh indicator
    const statusIndicator = document.getElementById('autoRefreshStatus');
    if (statusIndicator) {
        statusIndicator.style.display = 'inline';
    }
    
    // Show Auto Refresh label if viewing today's data (only exact "today" strings)
    // This is intentionally strict to control auto-refresh behavior
    const autoRefreshCell = document.getElementById('autoRefreshLabelCell');
    if (autoRefreshCell && startDateSelect.value === 'today' && endDateSelect.value === 'today') {
        autoRefreshCell.style.display = '';
    }
    
    // Update the toggle button text to match the current state
    const toggleBtn = document.getElementById('toggleAutoRefresh');
    if (toggleBtn) {
        toggleBtn.textContent = 'Disable Auto-refresh';
    }
    
    // Set new refresh interval (60 seconds)
    window.refreshInterval = setInterval(function() {
        // Check if auto-refresh has been disabled via the toggle button
        if (!autoRefreshEnabled) {
            console.log('Auto-refresh was disabled, stopping interval');
            clearInterval(window.refreshInterval);
            window.refreshInterval = null;
            
            // Hide the status indicator
            if (statusIndicator) {
                statusIndicator.style.display = 'none';
            }
            
            // Hide Auto Refresh label
            const autoRefreshCell = document.getElementById('autoRefreshLabelCell');
            if (autoRefreshCell) {
                autoRefreshCell.style.display = 'none';
            }
            
            return;
        }
        
        // Only refresh automatically if we're still viewing today's data with literal "today" selection
        // This is intentionally strict to control auto-refresh behavior
        if (document.getElementById('startDateSelect').value === 'today' &&
            document.getElementById('endDateSelect').value === 'today') {
            console.log('Auto-refreshing today\'s data');
            
            // Skip if a data request is already in progress
            if (dataRequestInProgress) {
                console.log('Data request already in progress, skipping auto-refresh');
                return;
            }
            
            // Update the end time to current time before refreshing (auto-refresh is always enabled here)
            const now = new Date();
            const endTimeInput = document.getElementById('endTime');
            if (endTimeInput) {
                endTimeInput.value = padNumber(now.getHours()) + ':' + padNumber(now.getMinutes());
                console.log('Updated end time to current time in auto-refresh interval:', endTimeInput.value);
            }
            
            // Disable the update button during auto-refresh
            const updateButton = document.getElementById('updateChart');
            if (updateButton) {
                updateButton.disabled = true;
            }
            
            // Get the current start time - preserve it for auto-refresh
            const startTimeInput = document.getElementById('startTime');
            const currentStartTime = startTimeInput ? startTimeInput.value : "00:00";
            
            // Get end time (which we just updated to current time)
            const endTime = endTimeInput ? endTimeInput.value : "23:59";
            
            // Format today's date as YYYYMMDD
            const today = new Date();
            const todayStr = today.getFullYear().toString() +
                           padNumber(today.getMonth() + 1) +
                           padNumber(today.getDate());
            
            console.log(`Auto-refreshing today's data with time range: ${currentStartTime} to ${endTime}`);
            
            // Use fetchDataWithDateRange instead of loadTodayData to preserve start time
            fetchDataWithDateRange(todayStr, todayStr, currentStartTime, endTime);
        } else {
            // Stop the auto-refresh if we're not viewing today's data anymore
            clearInterval(window.refreshInterval);
            window.refreshInterval = null;
            
            // Hide the status indicator
            if (statusIndicator) {
                statusIndicator.style.display = 'none';
            }
            
            // Hide Auto Refresh label
            const autoRefreshCell = document.getElementById('autoRefreshLabelCell');
            if (autoRefreshCell) {
                autoRefreshCell.style.display = 'none';
            }
        }
    }, 60000); // 60 seconds
}

// Fetch available date files from server using XMLHttpRequest
function fetchAvailableDates() {
    const xhr = new XMLHttpRequest();
    xhr.open('GET', '/api/dates', true);
    
    xhr.onload = function() {
        if (xhr.status === 200) {
            const dates = JSON.parse(xhr.responseText);
            availableDates = dates;
            
            // Update both date select dropdowns
            updateDateSelect('startDateSelect', dates);
            updateDateSelect('endDateSelect', dates);
        } else {
            console.error('Error fetching available dates, status:', xhr.status);
        }
    };
    
    xhr.onerror = function() {
        console.error('Network error when fetching available dates');
    };
    
    xhr.send();
}

// Helper function to update a date select dropdown
function updateDateSelect(selectId, dates) {
    const select = document.getElementById(selectId);
    
    // Keep the "today" option and add the rest
    while (select.options.length > 1) {
        select.remove(1);
    }
    
    // Add dates in reverse order (newest first)
    dates.sort().reverse().forEach(date => {
        const option = document.createElement('option');
        // Format from YYYYMMDD to YYYY-MM-DD for display
        const formattedDate = date.slice(0, 4) + '-' + date.slice(4, 6) + '-' + date.slice(6, 8);
        option.value = date;
        option.text = formattedDate;
        select.add(option);
    });
}

// Load today's data
function loadTodayData() {
    // Check if data request is already in progress
    if (dataRequestInProgress) {
        console.log('Data request already in progress, ignoring loadTodayData request');
        return;
    }
    
    // Disable the update button during data loading
    const updateButton = document.getElementById('updateChart');
    if (updateButton) {
        updateButton.disabled = true;
    }
    
    // Get the current date selectors
    const startDateSelect = document.getElementById('startDateSelect');
    const endDateSelect = document.getElementById('endDateSelect');
    
    // Get today's date in YYYYMMDD format for comparison
    const now = new Date();
    const todayStr = now.getFullYear().toString() +
                     padNumber(now.getMonth() + 1) +
                     padNumber(now.getDate());
    
    // Check if start date is "today"
    const startIsToday = startDateSelect && startDateSelect.value === 'today';
    
    // Check if end date is literally "today" (not a specific date that happens to be today)
    const endIsToday = endDateSelect && endDateSelect.value === 'today';
    
    // Don't reset start time - preserve user's choice
    const startTimeInput = document.getElementById('startTime');
    
    // Per new requirements, we no longer update the end time automatically when loading data
    // We only update end time when the end date is explicitly changed
    const endTimeInput = document.getElementById('endTime');
    
    // Get start and end times for data fetching
    const startTime = startTimeInput ? startTimeInput.value : "00:00";
    
    // Get end time
    const endTime = endTimeInput ? endTimeInput.value : "23:59";
    
    // Format today's date as YYYYMMDD
    const todayFormatted = now.getFullYear().toString() +
                     padNumber(now.getMonth() + 1) +
                     padNumber(now.getDate());
    
    console.log(`Loading today's data with time range: ${startTime} to ${endTime}`);
    
    // Use the date range function instead of single date to properly handle time ranges
    fetchDataWithDateRange(todayFormatted, todayFormatted, startTime, endTime);
}

// Update chart based on user selection
function updateChart() {
    // Check if data request is already in progress
    if (dataRequestInProgress) {
        console.log('Data request already in progress, ignoring update request');
        return;
    }
    
    // Disable the update button to prevent multiple clicks
    const updateButton = document.getElementById('updateChart');
    if (updateButton.disabled) {
        return; // Exit if button is already disabled (update in progress)
    }
    
    // Get start date
    const startDateSelect = document.getElementById('startDateSelect');
    let startDateValue = startDateSelect.value;
    
    if (startDateValue === 'today') {
        // Get today's date in YYYYMMDD format
        const today = new Date();
        startDateValue = today.getFullYear().toString() +
                         padNumber(today.getMonth() + 1) +
                         padNumber(today.getDate());
    }
    
    // Get end date
    const endDateSelect = document.getElementById('endDateSelect');
    let endDateValue = endDateSelect.value;
    
    if (endDateValue === 'today') {
        // Get today's date in YYYYMMDD format
        const today = new Date();
        endDateValue = today.getFullYear().toString() +
                       padNumber(today.getMonth() + 1) +
                       padNumber(today.getDate());
    }
    
    // Get start and end times
    const startTime = document.getElementById('startTime').value;
    const endTime = document.getElementById('endTime').value;
    
    // Send date range and time to server for filtering and downsampling
    fetchDataWithDateRange(startDateValue, endDateValue, startTime, endTime);
    
    // Get today's date in YYYYMMDD format for comparison
    const today = new Date();
    const todayStr = today.getFullYear().toString() +
                     padNumber(today.getMonth() + 1) +
                     padNumber(today.getDate());
    
    // Check if end date is literally "today" (not a specific date that happens to be today)
    // This is used for time setting
    const endDateIsToday = endDateSelect.value === 'today';
    
    // Auto-refresh is only available when both dates are literally "today"
    // This is intentionally strict to control auto-refresh behavior
    const bothDatesAreToday = startDateSelect.value === 'today' && endDateSelect.value === 'today';
    
    // Show/hide Auto Refresh elements based on date selection
    const autoRefreshLabel = document.querySelector('td[style*="color: green"] label');
    const autoRefreshCell = document.getElementById('autoRefreshLabelCell');
    const autoRefreshButtonCell = document.querySelector('td button#toggleAutoRefresh').parentNode.parentNode;
    
    if (bothDatesAreToday) {
        // Always show the Auto Refresh button when both dates are today
        if (autoRefreshButtonCell) autoRefreshButtonCell.style.display = '';
        
        // Only show the Auto Refresh label if enabled
        if (autoRefreshCell) {
            autoRefreshCell.style.display = autoRefreshEnabled ? '' : 'none';
        }
        
        // Get reference to the end time input for use with auto-refresh later
        const endTimeInput = document.getElementById('endTime');
        
        if (autoRefreshEnabled) {
            // Show auto-refresh indicator only if auto-refresh is enabled
            const statusIndicator = document.getElementById('autoRefreshStatus');
            if (statusIndicator) {
                statusIndicator.style.display = 'inline';
            }
            
            // Set up auto-refresh for future updates
            setAutoRefresh();
        }
        
        // Always load today's data immediately
        loadTodayData();
    } else {
        // Hide Auto Refresh elements
        if (autoRefreshCell) autoRefreshCell.style.display = 'none';
        if (autoRefreshButtonCell) autoRefreshButtonCell.style.display = 'none';
        
        // Stop any active auto-refresh
        if (window.refreshInterval) {
            clearInterval(window.refreshInterval);
            window.refreshInterval = null;
            
            // Hide the status indicator
            const statusIndicator = document.getElementById('autoRefreshStatus');
            if (statusIndicator) {
                statusIndicator.style.display = 'none';
            }
        }
        
        // When end date is today but not both dates are today, we no longer update time automatically
        // This ensures the Update button simply uses whatever time is already in the input field
    }
}

// Note: This function was removed as it's no longer called.
// All data fetching now goes through fetchDataWithDateRange

// Fetch data for a date range spanning potentially multiple days using XMLHttpRequest
function fetchDataWithDateRange(startDate, endDate, startTime, endTime) {
    // Prevent duplicate requests
    if (dataRequestInProgress) {
        console.log('Data request already in progress, ignoring duplicate request');
        return;
    }
    
    console.log('Fetching data range:', startDate, 'to', endDate, 'time:', startTime, 'to', endTime);
    dataRequestInProgress = true;
    
    // Disable the update button during data loading
    const updateButton = document.getElementById('updateChart');
    if (updateButton) {
        updateButton.disabled = true;
    }
    
    // Initialize temporary data storage for incremental processing
    let tempData = {
        powerData: [],
        hwData: [],
        opData: [],
        hwtriacData: [],
        hwKwhData: [],
        costData: []
    };
    
    // Track if charts are already created
    let chartsCreated = false;
    
    // Keep track of accumulated CSV text
    let accumulatedCSV = '';
    
    // Use a special API endpoint that handles multiple days and time-based filtering
    const xhr = new XMLHttpRequest();
    xhr.open('GET', `/api/data-range?startDate=${startDate}&endDate=${endDate}&start=${startTime}&end=${endTime}`, true);
    
    xhr.onload = function() {
        // Re-enable the update button
        if (updateButton) {
            updateButton.disabled = false;
        }
        dataRequestInProgress = false;
        
        // Clear any active timeout since the request completed
        if (timeoutTimer) {
            clearTimeout(timeoutTimer);
            timeoutTimer = null;
        }
        
        if (xhr.status === 200) {
            console.log('Data range fetched successfully, length:', xhr.responseText.length);
            try {
                if (xhr.responseText.trim().length === 0) {
                    console.error('Empty response received for date range');
                    // Create empty chart with message
                    document.getElementById("chartContainer").innerHTML = 
                        "<div style='text-align:center; padding-top:50px;'><h3>No data available for selected time range</h3></div>";
                    document.getElementById("hwtriacChartContainer").innerHTML = 
                        "<div style='text-align:center; padding-top:50px;'><h3>No data available for selected time range</h3></div>";
                    return;
                }
                
                // Get the date select elements
                const startDateSelect = document.getElementById('startDateSelect');
                const endDateSelect = document.getElementById('endDateSelect');
                
                // Check if either start or end is NOT 'today'
                const eitherDateNotToday = startDateSelect.value !== 'today' || endDateSelect.value !== 'today';
                
                // Final processing with complete data
                allData = parseCSV(xhr.responseText);
                console.log('Data range parsed successfully:', 
                    'power points:', allData.powerData.length,
                    'hw points:', allData.hwData.length);
                    
                // Only do the final render if auto-refresh is enabled and both dates are 'today'
                // Otherwise the chart has been incrementally updated as data came in
                if (autoRefreshEnabled && !eitherDateNotToday) {
                    // Final render with complete dataset
                    filterAndDisplayData();
                }
            } catch (error) {
                console.error('Error processing CSV data from range:', error);
                // Show error in chart container
                document.getElementById("chartContainer").innerHTML = 
                    "<div style='text-align:center; padding-top:50px;'><h3>Error loading data: " + error.message + "</h3></div>";
                document.getElementById("hwtriacChartContainer").innerHTML = 
                    "<div style='text-align:center; padding-top:50px;'><h3>Error loading data: " + error.message + "</h3></div>";
            }
        } else {
            console.error('Failed to fetch data range, status:', xhr.status);
            // Show error in chart container
            document.getElementById("chartContainer").innerHTML = 
                "<div style='text-align:center; padding-top:50px;'><h3>Error: Could not load data range (Status " + xhr.status + ")</h3></div>";
            document.getElementById("hwtriacChartContainer").innerHTML = 
                "<div style='text-align:center; padding-top:50px;'><h3>Error: Could not load data range (Status " + xhr.status + ")</h3></div>";
        }
    };
    
    xhr.onerror = function() {
        // Re-enable the update button on error
        if (updateButton) {
            updateButton.disabled = false;
        }
        dataRequestInProgress = false;
        
        // Clear any active timeout on error
        if (timeoutTimer) {
            clearTimeout(timeoutTimer);
            timeoutTimer = null;
        }
        
        console.error('Network error when fetching data range');
        // Show error in chart container
        document.getElementById("chartContainer").innerHTML = 
            "<div style='text-align:center; padding-top:50px;'><h3>Network error: Could not load data range</h3></div>";
        document.getElementById("hwtriacChartContainer").innerHTML = 
            "<div style='text-align:center; padding-top:50px;'><h3>Network error: Could not load data range</h3></div>";
            
        // Show a small network error message to the user
        const statusMessage = document.createElement('div');
        statusMessage.innerHTML = '<div style="position:fixed; top:10px; right:10px; background-color:#ff6b6b; color:white; padding:8px 12px; border-radius:4px; box-shadow:0 2px 4px rgba(0,0,0,0.2);">Network connection lost</div>';
        document.body.appendChild(statusMessage);
        
        // Remove the message after 5 seconds
        setTimeout(function() {
            if (statusMessage.parentNode) {
                statusMessage.parentNode.removeChild(statusMessage);
            }
        }, 5000);
    };
    
    // Add timeout handling with reset capability
    let timeoutTimer = null;
    const timeoutDuration = 5000; // 5 seconds timeout
    
    // Add a loadend event handler to ensure we always clean up properly
    xhr.addEventListener('loadend', function() {
        // This event fires when the request has completed, regardless of success or failure
        // Clean up the timeout if it's still active
        if (timeoutTimer) {
            clearTimeout(timeoutTimer);
            timeoutTimer = null;
        }
        
        // Ensure button is re-enabled and flag is reset
        if (updateButton) {
            updateButton.disabled = false;
        }
        dataRequestInProgress = false;
        
        // If we have partial data but onload wasn't fired (like in case of a network issue),
        // still display what we have
        if (!allData && tempData && (tempData.powerData?.length > 0 || tempData.hwData?.length > 0 || 
            tempData.opData?.length > 0 || tempData.hwtriacData?.length > 0)) {
            console.log('Request completed but data wasn\'t processed - using partial data');
            
            // Get the date select elements
            const startDateSelect = document.getElementById('startDateSelect');
            const endDateSelect = document.getElementById('endDateSelect');
            
            // Check if either start or end is NOT 'today'
            const eitherDateNotToday = startDateSelect.value !== 'today' || endDateSelect.value !== 'today';
            
            allData = tempData;
            
            // If auto-refresh is enabled and both dates are 'today', or if this is the first time creating charts,
            // call filterAndDisplayData to create and render the charts
            if (autoRefreshEnabled && !eitherDateNotToday || !chartsCreated) {
                filterAndDisplayData();
            } else {
                // If auto-refresh is disabled OR either date is NOT 'today', just update the existing charts
                updateChartData(tempData);
            }
        }
    });
    
    function setupTimeout() {
        // Clear any existing timeout
        if (timeoutTimer) {
            clearTimeout(timeoutTimer);
        }
        
        // Set new timeout
        timeoutTimer = setTimeout(function() {
            // When timeout occurs, handle it
            console.log('Request timeout when fetching data range - no chunks received for 5 seconds');
            
            // Plot the data that has been received, if any
            if (tempData && (tempData.powerData?.length > 0 || tempData.hwData?.length > 0 || 
                tempData.opData?.length > 0 || tempData.hwtriacData?.length > 0)) {
                console.log('Timeout occurred but plotting partial data that was received');
                
                // Get the date select elements
                const startDateSelect = document.getElementById('startDateSelect');
                const endDateSelect = document.getElementById('endDateSelect');
                
                // Check if either start or end is NOT 'today'
                const eitherDateNotToday = startDateSelect.value !== 'today' || endDateSelect.value !== 'today';
                
                // Save accumulated data to allData for final display
                allData = tempData;
                
                // If auto-refresh is enabled and both dates are 'today', or if this is the first time creating charts,
                // call filterAndDisplayData to create and render the charts
                if (autoRefreshEnabled && !eitherDateNotToday || !chartsCreated) {
                    filterAndDisplayData();
                } else {
                    // If auto-refresh is disabled OR either date is NOT 'today', just update the existing charts
                    updateChartData(tempData);
                }
            } else {
                // If no data was received, show error message
                document.getElementById("chartContainer").innerHTML = 
                    "<div style='text-align:center; padding-top:50px;'><h3>Request timeout: Server took too long to respond</h3></div>";
                document.getElementById("hwtriacChartContainer").innerHTML = 
                    "<div style='text-align:center; padding-top:50px;'><h3>Request timeout: Server took too long to respond</h3></div>";
            }
            
            // Reset flags and re-enable button
            if (updateButton) {
                updateButton.disabled = false;
            }
            dataRequestInProgress = false;
            
            // Abort the request
            xhr.abort();
        }, timeoutDuration);
    }
    
    // Initialize timeout
    setupTimeout();
    
    // Set up progress handler to process data chunks as they arrive
    xhr.onprogress = function(event) {
        // Reset the timeout whenever we receive data
        console.log('Received data chunk, resetting timeout timer');
        setupTimeout();
        
        // Check if this is the final chunk by comparing loaded vs total
        const isComplete = event.loaded === event.total && event.total > 0;
        
        // Process the new chunk of data
        if (xhr.responseText.length > 0) {
            try {
                // Get the new data that's arrived since the last chunk
                const newCSV = xhr.responseText.substring(accumulatedCSV.length);
                accumulatedCSV = xhr.responseText;
                
                // Split the new data into lines
                const newLines = newCSV.split('\n');
                
                // Process the complete lines in the new chunk
                if (newLines.length > 0) {
                    console.log(`Processing ${newLines.length} new lines from data chunk`);
                    
                    // Parse only the new complete lines
                    const newData = parseCSVChunk(newLines);
                    
                    // Add new data points to temporary storage
                    tempData.powerData = tempData.powerData.concat(newData.powerData);
                    tempData.hwData = tempData.hwData.concat(newData.hwData);
                    tempData.opData = tempData.opData.concat(newData.opData);
                    tempData.hwtriacData = tempData.hwtriacData.concat(newData.hwtriacData);
                    tempData.hwKwhData = tempData.hwKwhData.concat(newData.hwKwhData);
                    tempData.costData = tempData.costData.concat(newData.costData);
                    
                    // Update the chart with the accumulated data
                    if (tempData.powerData.length > 0 || tempData.hwData.length > 0) {
                        console.log('Updating chart with incremental data:', 
                            'power points:', tempData.powerData.length,
                            'hw points:', tempData.hwData.length);
                        
                        // Make a copy to prevent issues with concurrent modification
                        const dataToPlot = {
                            powerData: [...tempData.powerData],
                            hwData: [...tempData.hwData],
                            opData: [...tempData.opData],
                            hwtriacData: [...tempData.hwtriacData],
                            hwKwhData: [...tempData.hwKwhData],
                            costData: [...tempData.costData]
                        };
                        
                        // Create or update the chart with the current accumulated data
                        allData = dataToPlot;
                        
                        // Get the date select elements
                        const startDateSelect = document.getElementById('startDateSelect');
                        const endDateSelect = document.getElementById('endDateSelect');
                        
                        // Check if either start or end is NOT 'today'
                        const eitherDateNotToday = startDateSelect.value !== 'today' || endDateSelect.value !== 'today';
                        
                        if (autoRefreshEnabled && !eitherDateNotToday) {
                            // When auto-refresh is enabled and BOTH dates are 'today',
                            // only store the data but don't update the chart until the end
                            chartsCreated = true;
                        } else {
                            // When auto-refresh is disabled OR either date is NOT 'today',
                            // update incrementally with each chunk
                            if (!chartsCreated) {
                                filterAndDisplayData();
                                chartsCreated = true;
                            } else {
                                updateChartData(dataToPlot);
                            }
                        }
                    }
                    
                    // If we detect this is the final chunk, process all data now
                    if (isComplete) {
                        console.log('Final data chunk received (detected by progress event)');
                        
                        // Clear the timeout since we're processing the final data now
                        if (timeoutTimer) {
                            clearTimeout(timeoutTimer);
                            timeoutTimer = null;
                        }
                        
                        // Get the date select elements
                        const startDateSelect = document.getElementById('startDateSelect');
                        const endDateSelect = document.getElementById('endDateSelect');
                        
                        // Check if either start or end is NOT 'today'
                        const eitherDateNotToday = startDateSelect.value !== 'today' || endDateSelect.value !== 'today';
                        
                        // Process the full accumulated data
                        allData = parseCSV(accumulatedCSV);
                        
                        // Only update the chart at the end if autoRefresh is enabled and both dates are 'today'
                        // Otherwise, the chart has already been incrementally updated
                        if (autoRefreshEnabled && !eitherDateNotToday) {
                            filterAndDisplayData();
                        }
                        
                        // Re-enable the update button and reset flag
                        if (updateButton) {
                            updateButton.disabled = false;
                        }
                        dataRequestInProgress = false;
                    }
                }
            } catch (error) {
                console.error('Error processing incremental data:', error);
                // Continue receiving data even if there's an error in processing a chunk
            }
        }
    };
    
    xhr.send();
}

// Parse a chunk of CSV lines
function parseCSVChunk(lines) {
    const powerData = [];
    const hwData = [];
    const opData = [];
    const hwtriacData = [];
    const hwKwhData = [];
    const costData = [];
    
    // Process each complete line
    for (let i = 0; i < lines.length; i++) {
        const line = lines[i].trim();
        
        // Skip empty lines and comment lines
        if (line === '' || line.startsWith('//')) continue;
        
        // Skip header lines
        if (/date,\s*time,\s*unix/i.test(line)) continue;
        
        const fields = line.split(',').map(field => field.trim());
        
        // We need at least 8 fields for a valid line
        if (fields.length < 8) continue;
        
        try {
            // CSV format: Date,Time,Unix seconds,power_w,hw_w,op_w,state,hw_triac,hw_boost,hw_kwh,cost
            const dateStr = fields[0]; // YYYY/MM/DD
            const timeStr = fields[1]; // HH:MM:SS
            const power_w = parseFloat(fields[3]); // Power in watts
            const hw_w = parseFloat(fields[4]); // Hot water in watts
            const op_w = parseFloat(fields[5]); // Off-peak in watts 
            const hw_triac = parseInt(fields[7]); // HW triac value
            const hw_kwh = fields.length >= 10 ? parseFloat(fields[9]) : null; // Hot water kWh
            const cost = fields.length >= 11 ? parseFloat(fields[10]) : 0; // Cost in $
            
            // Parse the local date and time from the CSV
            const dateParts = dateStr.split('/');
            const timeParts = timeStr.split(':');
            
            // Skip if date format is invalid
            if (dateParts.length !== 3) continue;
            
            // Create date object using the local date/time strings
            const dateTime = new Date(
                parseInt(dateParts[0]),                                 // Year
                parseInt(dateParts[1]) - 1,                             // Month (0-based)
                parseInt(dateParts[2]),                                 // Day
                parseInt(timeParts[0]),                                 // Hours
                parseInt(timeParts[1]),                                 // Minutes
                timeParts.length > 2 ? parseInt(timeParts[2]) : 0       // Seconds
            );
            
            // Skip invalid dates or NaN values
            if (isNaN(dateTime.getTime()) || isNaN(power_w) || isNaN(hw_w) || isNaN(op_w)) continue;
            
            powerData.push({
                x: dateTime,
                y: power_w / 1000 // Convert to kW
            });
            
            hwData.push({
                x: dateTime,
                y: hw_w / 1000 // Convert to kW
            });
            
            opData.push({
                x: dateTime,
                y: op_w / 1000 // Convert to kW
            });
            
            hwtriacData.push({
                x: dateTime,
                y: hw_triac // Binary 0 or 1
            });
            
            // Only add hw_kwh data if the value is valid
            if (hw_kwh !== null && !isNaN(hw_kwh)) {
                hwKwhData.push({
                    x: dateTime,
                    y: hw_kwh // Already in kWh
                });
            }
            
            // Add cost data
            if (!isNaN(cost)) {
                costData.push({
                    x: dateTime,
                    y: cost
                });
            }
        } catch (error) {
            // Skip lines with parsing errors
            continue;
        }
    }
    
    return { powerData, hwData, opData, hwtriacData, hwKwhData, costData };
}

// Update existing chart with new data points
function updateChartData(data) {
    // Check if charts exist
    if (!chartInstances || chartInstances.length === 0) {
        console.log('No charts exist yet, creating new charts');
        allData = data;
        filterAndDisplayData();
        return;
    }
    
    console.log('Updating existing charts with new data points');
    
    // Update main chart
    if (chartInstances[0]) {
        // Update main chart data
        chartInstances[0].options.data[0].dataPoints = data.powerData;
        chartInstances[0].options.data[1].dataPoints = data.hwData;
        chartInstances[0].options.data[2].dataPoints = data.opData;
        chartInstances[0].options.data[3].dataPoints = data.costData;
        
        // Render the updated chart
        chartInstances[0].render();
    }
    
    // Update secondary chart
    if (chartInstances[1]) {
        // Adjust the hwKwhData to subtract the first value (make it relative from start point)
        let adjustedHwKwhData = [];
        if (data.hwKwhData.length > 0) {
            const startValue = data.hwKwhData[0].y;
            adjustedHwKwhData = data.hwKwhData.map(point => ({
                x: point.x,
                y: point.y - startValue
            }));
        }
        
        // Update secondary chart data
        chartInstances[1].options.data[0].dataPoints = data.hwtriacData;
        chartInstances[1].options.data[1].dataPoints = adjustedHwKwhData;
        
        // Render the updated chart
        chartInstances[1].render();
    }
}

// Filter data by time range and create chart
function filterAndDisplayData() {
    if (!allData) {
        console.error('Data object is null or undefined');
        
        // Show error in chart container
        document.getElementById("chartContainer").innerHTML = 
            "<div style='text-align:center; padding-top:50px;'><h3>Error: No data is available</h3></div>";
        document.getElementById("hwtriacChartContainer").innerHTML = 
            "<div style='text-align:center; padding-top:50px;'><h3>Error: No data is available</h3></div>";
        
        return;
    }
    
    // Check if we have any data at all
    const hasData = allData.powerData && allData.powerData.length > 0 || 
                    allData.hwData && allData.hwData.length > 0 || 
                    allData.opData && allData.opData.length > 0 ||
                    allData.hwtriacData && allData.hwtriacData.length > 0 ||
                    allData.costData && allData.costData.length > 0;
                    
    if (!hasData) {
        console.error('No valid data points available');
        
        // Show error in chart container
        document.getElementById("chartContainer").innerHTML = 
            "<div style='text-align:center; padding-top:50px;'><h3>No data points available for the selected time range</h3></div>";
        document.getElementById("hwtriacChartContainer").innerHTML = 
            "<div style='text-align:center; padding-top:50px;'><h3>No data points available for the selected time range</h3></div>";
        
        return;
    }
    
    // Get time range from inputs
    const startTimeStr = document.getElementById('startTime').value;
    const endTimeStr = document.getElementById('endTime').value;
    
    // Check if we're coming from a data-range API call
    // In this case, allData is already filtered by the server, so we just use it directly
    if (document.getElementById('startDateSelect').value !== document.getElementById('endDateSelect').value) {
        // Make sure we have costData even if it wasn't in the original data structure
        if (!allData.costData) {
            allData.costData = [];
        }
        // Data is from a date range query (different start/end dates), so it's already filtered by the server
        createChart(allData);
        return;
    }
    
    // Convert to minutes for easier comparison
    const startTimeMinutes = timeToMinutes(startTimeStr);
    const endTimeMinutes = timeToMinutes(endTimeStr);
    
    if (startTimeMinutes > endTimeMinutes) {
        alert('Start time must be before end time');
        return;
    }
    
    // Filter data by time range
    const filteredData = {
        powerData: allData.powerData.filter(point => {
            const pointTime = new Date(point.x);
            const pointMinutes = pointTime.getHours() * 60 + pointTime.getMinutes();
            return pointMinutes >= startTimeMinutes && pointMinutes <= endTimeMinutes;
        }),
        hwData: allData.hwData.filter(point => {
            const pointTime = new Date(point.x);
            const pointMinutes = pointTime.getHours() * 60 + pointTime.getMinutes();
            return pointMinutes >= startTimeMinutes && pointMinutes <= endTimeMinutes;
        }),
        opData: allData.opData.filter(point => {
            const pointTime = new Date(point.x);
            const pointMinutes = pointTime.getHours() * 60 + pointTime.getMinutes();
            return pointMinutes >= startTimeMinutes && pointMinutes <= endTimeMinutes;
        }),
        hwtriacData: allData.hwtriacData.filter(point => {
            const pointTime = new Date(point.x);
            const pointMinutes = pointTime.getHours() * 60 + pointTime.getMinutes();
            return pointMinutes >= startTimeMinutes && pointMinutes <= endTimeMinutes;
        }),
        hwKwhData: allData.hwKwhData.filter(point => {
            const pointTime = new Date(point.x);
            const pointMinutes = pointTime.getHours() * 60 + pointTime.getMinutes();
            return pointMinutes >= startTimeMinutes && pointMinutes <= endTimeMinutes;
        }),
        costData: allData.costData ? allData.costData.filter(point => {
            const pointTime = new Date(point.x);
            const pointMinutes = pointTime.getHours() * 60 + pointTime.getMinutes();
            return pointMinutes >= startTimeMinutes && pointMinutes <= endTimeMinutes;
        }) : []
    };
    
    // Create the chart - server now handles downsampling
    createChart(filteredData);
}

// Parse CSV format of SolarHotWaterMonitor data files
function parseCSV(csvText) {
    console.log('Starting CSV parsing');
    
    // Show a sample of the data to help debug format issues
    const sampleData = csvText.slice(0, 200) + '...';
    console.log('Sample of received data:', sampleData);
    
    // Split into lines and filter out empty lines and comment lines (starting with //)
    const lines = csvText.split('\n')
                   .map(line => line.trim())
                   .filter(line => line && !line.startsWith('//'));
                   
    console.log('Found', lines.length, 'non-empty, non-comment lines in CSV');
    
    if (lines.length === 0) {
        // Return empty data sets instead of throwing an error
        console.warn('CSV file is empty or improperly formatted');
        
        // Show error in chart container
        document.getElementById("chartContainer").innerHTML = 
            "<div style='text-align:center; padding-top:50px;'><h3>No valid data found in CSV file</h3></div>";
        document.getElementById("hwtriacChartContainer").innerHTML = 
            "<div style='text-align:center; padding-top:50px;'><h3>No valid data found in CSV file</h3></div>";
            
        return { powerData: [], hwData: [], opData: [], hwtriacData: [], hwKwhData: [] };
    }
    
    // Check if first line is a header row 
    // (contains field names like Date,Time,Unix seconds)
    const headerPattern = /date,\s*time,\s*unix/i;
    if (lines.length > 0 && headerPattern.test(lines[0])) {
        console.log('Detected header line, skipping:', lines[0]);
        // Remove header line
        lines.shift();
    }
    
    // Make sure we still have data after removing the header
    if (lines.length === 0) {
        console.warn('CSV file contains only header, no data rows');
        // Show error in chart container
        document.getElementById("chartContainer").innerHTML = 
            "<div style='text-align:center; padding-top:50px;'><h3>No data rows found in CSV file</h3></div>";
        document.getElementById("hwtriacChartContainer").innerHTML = 
            "<div style='text-align:center; padding-top:50px;'><h3>No data rows found in CSV file</h3></div>";
            
        return { powerData: [], hwData: [], opData: [], hwtriacData: [], hwKwhData: [] };
    }
    
    // Log the first data line (after potential header removal)
    console.log('First data line:', lines[0]);
    
    const powerData = [];
    const hwData = [];
    const opData = [];
    const hwtriacData = [];
    const hwKwhData = [];
    const costData = [];
    
    let validLines = 0;
    let invalidLines = 0;
    
    // Assume format: Date,Time,Unix seconds,power_w,hw_w,op_w,state,hw_triac,hw_boost,hw_kwh
    for (let i = 0; i < lines.length; i++) {
        const fields = lines[i].split(',').map(field => field.trim());
        
        // Debug first few lines to see what we're getting
        if (i < 2) {
            console.log(`Line ${i} fields:`, fields);
        }
        
        // We need at least 8 fields for a valid line based on CSV format
        // Date,Time,Unix seconds,power_w,hw_w,op_w,state,hw_triac
        if (fields.length < 8) {
            invalidLines++;
            if (i < 5) {
                console.warn(`Skipping incomplete line ${i+1}: expected >=8 fields, got ${fields.length}. Line: ${lines[i]}`);
            }
            continue; // Skip incomplete lines
        }
        
        try {
            // CSV format: Date,Time,Unix seconds,power_w,hw_w,op_w,state,hw_triac,hw_boost,hw_kwh,cost
            const dateStr = fields[0]; // YYYY/MM/DD
            const timeStr = fields[1]; // HH:MM:SS
            const unixTime = parseInt(fields[2]); // Unix timestamp
            const power_w = parseFloat(fields[3]); // Power in watts
            const hw_w = parseFloat(fields[4]); // Hot water in watts
            const op_w = parseFloat(fields[5]); // Off-peak in watts 
            const state = parseInt(fields[6]); // State
            const hw_triac = parseInt(fields[7]); // HW triac value
            const hw_boost = fields.length >= 9 ? parseInt(fields[8]) : 0; // HW boost
            const hw_kwh = fields.length >= 10 ? parseFloat(fields[9]) : null; // Hot water kWh
            const cost = fields.length >= 11 ? parseFloat(fields[10]) : 0; // Cost in $
            
            // Parse the local date and time from the CSV directly
            // The unix timestamp is UTC, but the dateStr and timeStr are in local time
            const dateParts = dateStr.split('/');
            const timeParts = timeStr.split(':');
            
            // Better error handling for date format
            if (dateParts.length !== 3) {
                if (i < 5) {
                    console.warn(`Invalid date format at line ${i+1}: ${dateStr} - expected YYYY/MM/DD, got ${dateParts.length} parts`);
                }
                invalidLines++;
                continue;
            }
            
            // Create date object using the local date/time strings, not the UTC timestamp
            const dateTime = new Date(
                parseInt(dateParts[0]),                                 // Year
                parseInt(dateParts[1]) - 1,                             // Month (0-based)
                parseInt(dateParts[2]),                                 // Day
                parseInt(timeParts[0]),                                 // Hours
                parseInt(timeParts[1]),                                 // Minutes
                timeParts.length > 2 ? parseInt(timeParts[2]) : 0       // Seconds
            );
            
            if (isNaN(dateTime.getTime())) {
                console.warn(`Invalid date/time at line ${i+1}: ${dateStr} ${timeStr}`);
                invalidLines++;
                continue;
            }
            
            // Skip NaN values
            if (isNaN(power_w) || isNaN(hw_w) || isNaN(op_w)) {
                if (i < 5) {
                    console.warn(`Skipping line ${i+1} with NaN values: power=${power_w}, hw=${hw_w}, op=${op_w}`);
                }
                invalidLines++;
                continue;
            }
            
            powerData.push({
                x: dateTime,
                y: power_w / 1000 // Convert to kW
            });
            
            hwData.push({
                x: dateTime,
                y: hw_w / 1000 // Convert to kW
            });
            
            opData.push({
                x: dateTime,
                y: op_w / 1000 // Convert to kW
            });
            
            hwtriacData.push({
                x: dateTime,
                y: hw_triac // Binary 0 or 1
            });
            
            // Only add hw_kwh data if the value is valid
            if (hw_kwh !== null && !isNaN(hw_kwh)) {
                hwKwhData.push({
                    x: dateTime,
                    y: hw_kwh // Already in kWh
                });
            }
            
            // Add cost data
            if (!isNaN(cost)) {
                costData.push({
                    x: dateTime,
                    y: cost
                });
            }
            
            validLines++;
        } catch (error) {
            console.warn(`Error parsing line ${i+1}: ${error.message}`);
            invalidLines++;
        }
    }
    
    console.log(`CSV parsing complete. Valid lines: ${validLines}, Invalid lines: ${invalidLines}`);
    console.log(`Data points created: power=${powerData.length}, hw=${hwData.length}, op=${opData.length}, triac=${hwtriacData.length}, kwh=${hwKwhData.length}, cost=${costData.length}`);
    
    // Check if we have at least some data
    if (powerData.length === 0 && hwData.length === 0 && opData.length === 0) {
        console.error('No valid data points could be extracted from CSV');
        
        // Show error in chart container
        document.getElementById("chartContainer").innerHTML = 
            "<div style='text-align:center; padding-top:50px;'><h3>Could not extract any valid data from file</h3><p>Check console for details</p></div>";
        document.getElementById("hwtriacChartContainer").innerHTML = 
            "<div style='text-align:center; padding-top:50px;'><h3>Could not extract any valid data from file</h3><p>Check console for details</p></div>";
    }
    
    return { powerData, hwData, opData, hwtriacData, hwKwhData, costData };
}

// Downsampling is now handled on the server side

// Create the chart with the given data
function createChart(data) {
    console.log('Creating chart with data:', {
        powerDataLength: data.powerData ? data.powerData.length : 0,
        hwDataLength: data.hwData ? data.hwData.length : 0,
        opDataLength: data.opData ? data.opData.length : 0,
        hwtriacDataLength: data.hwtriacData ? data.hwtriacData.length : 0,
        hwKwhDataLength: data.hwKwhData ? data.hwKwhData.length : 0,
        costDataLength: data.costData ? data.costData.length : 0
    });
    
    // Make sure we have valid data arrays
    if (!data.powerData) data.powerData = [];
    if (!data.hwData) data.hwData = [];
    if (!data.opData) data.opData = [];
    if (!data.hwtriacData) data.hwtriacData = [];
    if (!data.hwKwhData) data.hwKwhData = [];
    if (!data.costData) data.costData = [];
    
    // Check if we have any data to display
    const hasMainChartData = data.powerData.length > 0 || data.hwData.length > 0 || 
                            data.opData.length > 0 || data.costData.length > 0;
    const hasSecondChartData = data.hwtriacData.length > 0 || data.hwKwhData.length > 0;
    
    if (!hasMainChartData && !hasSecondChartData) {
        // Display a message in the chart container
        document.getElementById("chartContainer").innerHTML = 
            "<div style='text-align:center; padding-top:50px;'><h3>No data available for the selected time range</h3></div>";
        document.getElementById("hwtriacChartContainer").innerHTML = 
            "<div style='text-align:center; padding-top:50px;'><h3>No data available for the selected time range</h3></div>";
        return;
    }

    // Clean up any existing chart instances to prevent memory leaks and event handler issues
    if (chartInstances.length > 0) {
        for (let i = 0; i < chartInstances.length; i++) {
            chartInstances[i].destroy();
        }
        chartInstances = [];
    }
    
    // Adjust the hwKwhData to subtract the first value (make it relative from start point)
    if (data.hwKwhData.length > 0) {
        const startValue = data.hwKwhData[0].y;
        data.hwKwhData = data.hwKwhData.map(point => ({
            x: point.x,
            y: point.y - startValue
        }));
    }

    // Create an array to hold both charts for synchronization
    let charts = [];

    // Use the costData from the server
    // Note: We're now using data.costData directly instead of redefining it
    
    // First chart configuration
    const chart = new CanvasJS.Chart("chartContainer", {
        title: {
            text: ""
        },
        zoomEnabled: true,
        panEnabled: true,
        zoomType: "xy",
        exportEnabled: true,
        axisX: {
            title: "Date/Time",
            valueFormatString: "MMM DD HH:mm",
            labelFormatter: function(e) {
                // Get the date directly from the value
                const date = e.value;
                
                // Create a formatted string showing the date and time in a consistent way
                // Using direct access to date components to avoid any timezone issues
                const months = ['Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec'];
                return `${String(date.getHours()).padStart(2, '0')}:${String(date.getMinutes()).padStart(2, '0')}\n${date.getDate()} ${months[date.getMonth()]}`;
            },
            scaleBreaks: {
                autoCalculate: true,
                collapsibleThreshold: "50%"
            }
        },
        axisY: {
            title: "Power (kW)",
            includeZero: true
        },
        axisY2: {
            title: "Cost ($)",
            includeZero: true,
            lineColor: "#006400", // Dark green color for cost
            tickColor: "#006400",
            labelFontColor: "#006400",
            titleFontColor: "#006400"
        },
        legend: {
            cursor: "pointer",
            verticalAlign: "top",
            horizontalAlign: "center",
            dockInsidePlotArea: false
        },
        rangeSelector: {
            enabled: true
        },
        toolTip: {
            shared: true,
            contentFormatter: function(e) {
                // Get the data point date directly - this will be the local date we parsed
                const date = e.entries[0].dataPoint.x;
                
                // Format using individual date components to ensure we get the correct date
                const formattedDate = `${date.getFullYear()}-${String(date.getMonth() + 1).padStart(2, '0')}-${String(date.getDate()).padStart(2, '0')}`;
                const formattedTime = `${String(date.getHours()).padStart(2, '0')}:${String(date.getMinutes()).padStart(2, '0')}`;
                
                // For display, format in a friendly way
                const options = { 
                    month: 'short', 
                    day: 'numeric', 
                    year: 'numeric',
                    hour: '2-digit',
                    minute: '2-digit',
                    hour12: false
                };
                
                // Create a formatted date/time string showing the exact day and month directly
                const formattedDateTime = `${['Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec'][date.getMonth()]} ${date.getDate()}, ${date.getFullYear()} ${String(date.getHours()).padStart(2, '0')}:${String(date.getMinutes()).padStart(2, '0')}`;
                
                let content = "<div style='text-align:center;'><strong>" + formattedDateTime + "</strong></div><hr/>";
                
                for (let i = 0; i < e.entries.length; i++) {
                    const entry = e.entries[i];
                    content += "<div style='color: " + entry.dataSeries.color + ";'>";
                    if (entry.dataSeries.name === "Cost ($)") {
                        content += entry.dataSeries.name + ": <strong>$" + entry.dataPoint.y.toFixed(2) + "</strong>";
                    } else {
                        content += entry.dataSeries.name + ": <strong>" + entry.dataPoint.y.toFixed(3) + " kW</strong>";
                    }
                    content += "</div>";
                }
                
                return content;
            }
        },
        data: [
            {
                type: "line",
                name: "Total Power (kW)",
                showInLegend: true,
                nullDataLineDashType: "dot",
                connectNullData: false,
                dataPoints: data.powerData
            },
            {
                type: "line",
                name: "Hot Water (kW)",
                showInLegend: true,
                nullDataLineDashType: "dot",
                connectNullData: false,
                dataPoints: data.hwData
            },
            {
                type: "line",
                name: "Off Peak (kW)",
                showInLegend: true,
                nullDataLineDashType: "dot",
                connectNullData: false,
                dataPoints: data.opData
            },
            {
                type: "line",
                name: "Cost ($)",
                showInLegend: true,
                axisYType: "secondary",
                color: "#006400", // Dark green color for cost
                nullDataLineDashType: "dot",
                connectNullData: false,
                dataPoints: data.costData
            }
        ]
    });
    
    // Create the second chart for HW triac and accumulated kWh
    const hwtriacChart = new CanvasJS.Chart("hwtriacChartContainer", {
        title: {
            text: ""
        },
        zoomEnabled: true,
        panEnabled: true,
        zoomType: "xy",
        exportEnabled: true,
        axisX: {
            title: "Date/Time",
            valueFormatString: "MMM DD HH:mm",
            labelFormatter: function(e) {
                const date = e.value;
                const months = ['Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec'];
                return `${String(date.getHours()).padStart(2, '0')}:${String(date.getMinutes()).padStart(2, '0')}\n${date.getDate()} ${months[date.getMonth()]}`;
            },
            scaleBreaks: {
                autoCalculate: true,
                collapsibleThreshold: "50%"
            }
        },
        axisY: {
            title: "Triac State",
            includeZero: true,
            maximum: 1.5, // For binary 0/1 values
            interval: 1,
            lineColor: "#0000FF", // Blue to match the triac line
            tickColor: "#0000FF",
            labelFontColor: "#0000FF",
            titleFontColor: "#0000FF",
            labelFormatter: function(e) {
                return e.value === 0 ? "OFF" : e.value === 1 ? "ON" : e.value;
            }
        },
        axisY2: {
            title: "Hot Water kWh",
            includeZero: true,
            labelFontSize: 20,
            lineColor: "#C24642",
            tickColor: "#C24642",
            labelFontColor: "#C24642",
            titleFontColor: "#C24642",
            labelFormatter: function(e) {
               return CanvasJS.formatNumber(e.value, "0.0");
            }
        },
        legend: {
            cursor: "pointer",
            verticalAlign: "top",
            horizontalAlign: "center",
            dockInsidePlotArea: false
        },
        rangeSelector: {
            enabled: true
        },
        toolTip: {
            shared: true,
            contentFormatter: function(e) {
                const date = e.entries[0].dataPoint.x;
                const formattedDateTime = `${['Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec'][date.getMonth()]} ${date.getDate()}, ${date.getFullYear()} ${String(date.getHours()).padStart(2, '0')}:${String(date.getMinutes()).padStart(2, '0')}`;
                
                let content = "<div style='text-align:center;'><strong>" + formattedDateTime + "</strong></div><hr/>";
                
                for (let i = 0; i < e.entries.length; i++) {
                    const entry = e.entries[i];
                    let valueDisplay;
                    
                    if (entry.dataSeries.name === "HW Triac") {
                        valueDisplay = entry.dataPoint.y === 0 ? "OFF" : "ON";
                    } else {
                        valueDisplay = entry.dataPoint.y.toFixed(3) + " kWh";
                    }
                    
                    content += "<div style='color: " + entry.dataSeries.color + ";'>";
                    content += entry.dataSeries.name + ": <strong>" + valueDisplay + "</strong>";
                    content += "</div>";
                }
                
                return content;
            }
        },
        data: [
            {
                type: "stepLine",
                name: "HW Triac",
                showInLegend: true,
                markerSize: 0,
                color: "#0000FF", // Blue color
                dataPoints: data.hwtriacData
            },
            {
                type: "line",
                name: "Hot Water (kWh)",
                showInLegend: true,
                axisYType: "secondary",
                color: "#C24642",
                dataPoints: data.hwKwhData
            }
        ]
    });
    
    // Add charts to the array for synchronization and also store them in the global chartInstances
    charts.push(chart);
    charts.push(hwtriacChart);
    
    // Store chart instances in global variable for future cleanup
    chartInstances = charts;
    
    // Configure crosshair for all charts
    for (let i = 0; i < charts.length; i++) {
        charts[i].options.axisX.crosshair = {
            enabled: true,
            snapToDataPoint: true,
            valueFormatString: "HH:mm:ss"
        };
    }
    
    // Synchronize charts (tooltips, crosshair, and zoom/pan)
    // Reset the event handlers by passing resetHandlers=true
    syncCharts(charts, true, true, true);
    
    // Render all charts
    for (let i = 0; i < charts.length; i++) {
        charts[i].render();
    }
}

// Helper: Convert time string "HH:MM" to minutes
function timeToMinutes(timeStr) {
    const [hours, minutes] = timeStr.split(':').map(Number);
    return hours * 60 + minutes;
}

// Helper: Pad numbers with leading zeros
function padNumber(num) {
    return num.toString().padStart(2, '0');
}

// Synchronize charts (tooltips, crosshair, and zoom/pan)
function syncCharts(charts, syncToolTip, syncCrosshair, syncAxisXRange) {
    // Always redefine event handlers to ensure they're working with the current charts array
    // This is critical for chart synchronization to work on subsequent updates
    
    // Define the tooltip updated handler
    this.onToolTipUpdated = function(e) {
        for (var j = 0; j < charts.length; j++) {
            if (charts[j] != e.chart)
                charts[j].toolTip.showAtX(e.entries[0].xValue);
        }
    };

    // Define the tooltip hidden handler
    this.onToolTipHidden = function(e) {
        for( var j = 0; j < charts.length; j++){
            if(charts[j] != e.chart)
                charts[j].toolTip.hide();
        }
    };

    // Define the crosshair updated handler
    this.onCrosshairUpdated = function(e) {
        for(var j = 0; j < charts.length; j++){
            if(charts[j] != e.chart)
                charts[j].axisX[0].crosshair.showAt(e.value);
        }
    };

    // Define the crosshair hidden handler
    this.onCrosshairHidden = function(e) {
        for( var j = 0; j < charts.length; j++){
            if(charts[j] != e.chart)
                charts[j].axisX[0].crosshair.hide();
        }
    };

    // Define the range changed handler
    this.onRangeChanged = function(e) {
        for (var j = 0; j < charts.length; j++) {
            if (e.trigger === "reset") {
                charts[j].options.axisX.viewportMinimum = charts[j].options.axisX.viewportMaximum = null;
                charts[j].options.axisY.viewportMinimum = charts[j].options.axisY.viewportMaximum = null;
                charts[j].render();
            } else if (charts[j] !== e.chart) {
                charts[j].options.axisX.viewportMinimum = e.axisX[0].viewportMinimum;
                charts[j].options.axisX.viewportMaximum = e.axisX[0].viewportMaximum;
                charts[j].render();
            }
        }
    };

    for(var i = 0; i < charts.length; i++) { 
        // Sync ToolTip
        if(syncToolTip) {
            if(!charts[i].options.toolTip)
                charts[i].options.toolTip = {};

            charts[i].options.toolTip.updated = this.onToolTipUpdated;
            charts[i].options.toolTip.hidden = this.onToolTipHidden;
        }

        // Sync Crosshair
        if(syncCrosshair) {
            if(!charts[i].options.axisX)
                charts[i].options.axisX = { crosshair: { enabled: true }};
            
            charts[i].options.axisX.crosshair.updated = this.onCrosshairUpdated; 
            charts[i].options.axisX.crosshair.hidden = this.onCrosshairHidden; 
        }

        // Sync Zoom / Pan
        if(syncAxisXRange) {
            charts[i].options.zoomEnabled = true;
            charts[i].options.rangeChanged = this.onRangeChanged;
        }
    }
}