#include <Arduino.h>
#include "processData.h"
#include "millisDelay.h"
#include "HistoryRecord.h"
#include "HistoryFile.h"
#include "volatileVars.h"
#include "FS.h"
#include "SafeStringReader.h"
#include "HW_webpages.h"
#include <vector>
#include <algorithm> // for std::sort
#include "locks.h"
#include "systemTimeUtils.h"
#include "RTC_NTP_replacement.h"

const float IMPORT_COST_PER_kWhr = 0.3674;
const float EXPORT_COST_PER_kWhr =  -0.050000;  // Note Export_kWhr is a +ve number always
const float OFFPEAK_COST_PER_kWhr =  0.2189;
const float DAILY_CHARGE = 1.617;

// Constant for fully heated timeout (20 seconds)
static const unsigned long FULLY_HEATED_TIMEOUT_MS = 20000;
const unsigned long MAX_RECORD_GAP = 6000; // if time between records > 6 sec, do not accumulate

//#define DEBUG
static Print* debugPtr = NULL;

#define MAX_RESPONSE_SIZE 265

HistoryRecordStruct historyRecord;

#define LINE_LEN 265
createSafeStringReader(sfProcessReader, LINE_LEN, "\r\n"); // allow for lines 256 long

static Accumulated_kWh_RecordStruct processResults; // initialize on startup from existing date file

void updateYesterdaysValues();

/**
   Initialize an Accumulated_kWh_RecordStruct with default values

   @param results The structure to initialize
*/
void initAccumulatedKWhRecord(Accumulated_kWh_RecordStruct &results) {
  results.export_kWhrs = 0.0;
  results.import_kWhrs = 0.0;
  results.offPeak_kWhrs = 0.0;
  results.hotwater_kWhrs = 0.0;
  results.cost = 0.0;
  results.lastPower_kW = 0.0;
  results.lastHW_kW = 0.0;
  results.t_offset = 0; // Will be set in accumulate_kWh
  results.lasttime_ms = 0;
  results.fullyHeated = false;
  results.lowPowerTracking = false;
  results.lowPowerStartTime = 0;
}

/**
   Accumulate the kilowatt-hours from power readings

   @param record The binary record containing power measurements
   @param results The structure to accumulate kWh into
*/
void accumulate_kWh(const BinaryRecordStruct &record, Accumulated_kWh_RecordStruct &results) {
  // Calculate current time in milliseconds
  unsigned long currentTime_ms;


  // First call: initialize and store initial values
  if (results.t_offset == 0) {
    results.t_offset = record.utc_time;
    // Calculate current time with new offset
    currentTime_ms = ((record.utc_time - results.t_offset) * 1000) + (record.utc_usec / 1000);
    results.lasttime_ms = currentTime_ms;
    return;
  }
  if (record.utc_time < results.t_offset) {
    return; // to early time when backwards
  }
  if (record.utc_usec > 1000000ul) {
    return; // invalid
  }
  // Calculate current time with existing offset
  currentTime_ms = ((record.utc_time - results.t_offset) * 1000) + (record.utc_usec / 1000);

  if (currentTime_ms > (results.lasttime_ms + (25ul*60*60*1000))) {
    return; // to far in the future
  }

  // Calculate time difference in milliseconds
  if (currentTime_ms < results.lasttime_ms) {
    return; // time when backwards
  }
  uint64_t deltaTime_ms = currentTime_ms - results.lasttime_ms;

  if (deltaTime_ms < 0) {
    deltaTime_ms = 0;
  }

  // Reset low power tracking if gap is too large
  if (deltaTime_ms >= MAX_RECORD_GAP) {
    results.lowPowerTracking = false;
  }

  // Check for fully heated condition
  if (!results.fullyHeated) {
    // Check conditions for starting or continuing low power tracking
    if (record.hw_triac == 1 && record.state == 1 && record.hw_w < 100) {
      // If tracking hasn't started yet, start it
      if (!results.lowPowerTracking) {
        results.lowPowerTracking = true;
        results.lowPowerStartTime = currentTime_ms;
      }
      // Check if we've been in low power state for 20 seconds
      else if ((currentTime_ms - results.lowPowerStartTime) >= FULLY_HEATED_TIMEOUT_MS) {
        // Mark as fully heated
        results.fullyHeated = true;
      }
    } else {
      // Reset tracking if conditions aren't met
      results.lowPowerTracking = false;
    }
  }
  double power_kW = 0;
  double HW_kW = 0;
  if (deltaTime_ms < MAX_RECORD_GAP) { // Only process if gap is less than 5 seconds
    double deltaTime_hrs = (float)deltaTime_ms / (3600.0 * 1000.0);

    // Calculate average power for this period in kW
    power_kW = (double)(record.power_w);
    HW_kW = (double)(record.hw_w);
    // this average may underestimate kWh after gap, but already have the gap!!
    double avgPower_kW = ((power_kW + results.lastPower_kW) / 2.0) / 1000.0;
    double avgHW_kW = ((HW_kW + results.lastHW_kW) / 2.0) / 1000.0;

    // Calculate energy for this period
    double energy_kWh = avgPower_kW * deltaTime_hrs;

    // Update accumulators
    if (avgPower_kW >= 0) { // Importing
      results.import_kWhrs += energy_kWh;
    } else { // Exporting
      results.export_kWhrs -= energy_kWh; // Store as positive value
    }

    double hw_energy_kWh = avgHW_kW * deltaTime_hrs;
    // For off-peak, check if in off-peak mode
    if (record.state == 2) { // Off-peak mode
      results.offPeak_kWhrs += hw_energy_kWh;
    }
    results.hotwater_kWhrs += hw_energy_kWh;
  }

  // Update lasttime_ms for next calculation
  results.lasttime_ms = currentTime_ms;
  results.lastPower_kW = power_kW; // zero if there is a gap
  results.lastHW_kW = HW_kW;
}


static bool processDataInitialized = false;


void calculateTotalCost(Accumulated_kWh_RecordStruct& accResults) {
  // Calculate total cost
  accResults.cost = (accResults.import_kWhrs * IMPORT_COST_PER_kWhr) +
                    (accResults.export_kWhrs * EXPORT_COST_PER_kWhr) +
                    (accResults.offPeak_kWhrs * OFFPEAK_COST_PER_kWhr);
}



void printDailyHistory(Print *outPtr) {
  if (!outPtr) {
    return;
  }
  outPtr->print(" Today:     ");
  outPtr->print(" Import: "); outPtr->print(historyRecord.import_kWhrs, 3); outPtr->print(" kWh,");
  outPtr->print("  Export: "); outPtr->print(historyRecord.export_kWhrs, 3);   outPtr->print(" kWh,");
  outPtr->print("  Offpeak: "); outPtr->print(historyRecord.offPeak_kWhrs, 3);  outPtr->print(" kWh,");
  outPtr->print("  HW fullyHeated: "); outPtr->print(historyRecord.fullyHeated ? "true" : "false");
  outPtr->print(" Cost: $");  outPtr->println(historyRecord.cost, 2);
  outPtr->print(" Yesterday: ");
  outPtr->print(" Import: "); outPtr->print(historyRecord.import_kWhrs_yesterday, 3); outPtr->print(" kWh,");
  outPtr->print("  Export: "); outPtr->print(historyRecord.export_kWhrs_yesterday, 3);   outPtr->print(" kWh,");
  outPtr->print("  Offpeak: "); outPtr->print(historyRecord.offPeak_kWhrs_yesterday, 3);  outPtr->print(" kWh,");
  outPtr->print("  HW fullyHeated: "); outPtr->print(historyRecord.fullyHeated_yesterday ? "true" : "false");
  outPtr->print(" Cost: $");  outPtr->println(historyRecord.cost_yesterday, 2);
}

// Function to initialize daily running cost values from the current day's file
void initializeDailyRunningValues() {
  const char* todaysDateStr = getCurrentDateFilename(); // Get the current date filename

  // Check if today's file exists
  String filePath = "/" + String(todaysDateStr);
  if (!LittleFS.exists(filePath)) {
    if (debugPtr) {
      debugPtr->print("Today's data file not found: ");
      debugPtr->println(filePath);
    }
    return; // No file for today, leave values at zero
  }
  if (debugPtr) {
    debugPtr->print("Initializing daily values from file: ");
    debugPtr->println(filePath);
  }

  initAccumulatedKWhRecord(processResults);

  calculateDailyCost(todaysDateStr, processResults, false);

  // Copy accumulated values to history record
  historyRecord.import_kWhrs = processResults.import_kWhrs;
  historyRecord.export_kWhrs = processResults.export_kWhrs;
  historyRecord.offPeak_kWhrs = processResults.offPeak_kWhrs;
  historyRecord.hotwater_kWhrs = processResults.hotwater_kWhrs;
  historyRecord.fullyHeated = processResults.fullyHeated;
  historyRecord.cost = processResults.cost;

  //Member objects
  //int tm_sec  seconds after the minute – [0, 61] (until C99)[0, 60] (since C99)[for leap second]
  //int tm_min minutes after the hour – [0, 59]
  //int tm_hour hours since midnight – [0, 23]
  //int tm_mday day of the month – [1, 31]
  //int tm_mon months since January – [0, 11]
  //int tm_year years since 1900
  //int tm_wday days since Sunday – [0, 6]
  //int tm_yday days since January 1 – [0, 365]
  //int tm_isdst Daylight Saving Time flag. The value is positive if DST is in effect, zero if not and negative if no information is available

  // calculate yesterdays
  struct tm yesterday;
  time_t now = time(nullptr);
  struct tm timeinfo;
  struct tm* timeinfoPtr = localtime_r(&now, &timeinfo);
  yesterday.tm_year = timeinfoPtr->tm_year;
  yesterday.tm_mon = timeinfoPtr->tm_mon;
  yesterday.tm_mday = timeinfoPtr->tm_mday;
  yesterday.tm_hour = timeinfoPtr->tm_hour;
  yesterday.tm_hour = 0;
  yesterday.tm_min = 0;
  yesterday.tm_sec = 0;
  yesterday.tm_isdst = -1;

  yesterday.tm_mday -= 1; // one day ago
  mktime(&yesterday); // clean up
  char yesterdaysFilename[40] = {0}; // Static to ensure it's available after function returns
  sprintf(yesterdaysFilename, "/%04d%02d%02d", yesterday.tm_year + 1900, yesterday.tm_mon + 1, yesterday.tm_mday);

  String yesterdaysFilePath = "/" + String(yesterdaysFilename);
  if (!LittleFS.exists(yesterdaysFilePath)) {
    if (debugPtr) {
      debugPtr->print("Yesterdays's data file not found: ");
      debugPtr->println(yesterdaysFilePath);
    }
    return; // No file for today, leave values at zero
  }
  if (debugPtr) {
    debugPtr->print("Initializing yesterdays values from file: ");
    debugPtr->println(yesterdaysFilePath);
  }

  Accumulated_kWh_RecordStruct accResults;
  initAccumulatedKWhRecord(accResults);
  calculateDailyCost(yesterdaysFilename, accResults, false);

  // Copy accumulated values to history record
  historyRecord.import_kWhrs_yesterday = accResults.import_kWhrs;
  historyRecord.export_kWhrs_yesterday = accResults.export_kWhrs;
  historyRecord.offPeak_kWhrs_yesterday = accResults.offPeak_kWhrs;
  historyRecord.hotwater_kWhrs_yesterday = accResults.hotwater_kWhrs;
  historyRecord.fullyHeated_yesterday = accResults.fullyHeated;
  historyRecord.cost_yesterday = accResults.cost;


  Print* oldPtr = debugPtr;
  //  debugPtr = &Serial;
  if (debugPtr) {
    debugPtr->println("Initialized daily running totals - ");
    printDailyHistory(debugPtr);
  }
  debugPtr = oldPtr;
}


void processDataInit() {
  if (processDataInitialized) {
    return;
  }
#ifdef DEBUG
  debugPtr = getDebugOut();
#endif
  if (!isRTCDateTimeSet()) {
    return;
  }
  // Initialize daily running values from today's file
  initializeDailyRunningValues();
  // set yesterday's total cost
  updateYesterdaysValues();

  // Load saved import/export data
  //loadImportExport();
}

bool parseResponse(SafeString & sfLine);

void updateYesterdaysValues() {
  Print *old_debugPtr = debugPtr;
  debugPtr = &Serial;
  if (debugPtr) {
    debugPtr->print(" updateYesterdaysValues ");
  }
//  Accumulated_kWh_RecordStruct accResults;
//  accResults.export_kWhrs = historyRecord.export_kWhrs_yesterday;
//  accResults.import_kWhrs = historyRecord.import_kWhrs_yesterday;
//  accResults.offPeak_kWhrs =  historyRecord.offPeak_kWhrs_yesterday;
//  calculateTotalCost(accResults);
  totalCost_yesterday_v = historyRecord.cost_yesterday;
  totalHotWaterKWhrs_yesterday_v = historyRecord.hotwater_kWhrs_yesterday;
  fullyHeated_yesterday_v = historyRecord.fullyHeated_yesterday;
  updated_v = true;
  if (debugPtr) {
    debugPtr->print(" totalCost_yesterday_v: ");debugPtr->print(totalCost_yesterday_v);
    debugPtr->print(" totalHotWaterKWhrs_yesterday_v: ");debugPtr->print(totalHotWaterKWhrs_yesterday_v);
    debugPtr->print(" fullyHeated_yesterday_v: ");debugPtr->print(fullyHeated_yesterday_v);
    debugPtr->println();
  }
  debugPtr = old_debugPtr;
}


// data format in CSV
// field 1 -- HWC constant string, "HWC"
// field 2 -- readCounter, 1 to max uint32_t
// field 3 -- state, 0 Off, 1 Solar Hot Water, 2 Off Peak
// field 4 -- HWtriac, 0 off, 1 on
// field 5 -- HWboost, 0 off, 1 on  automatically turns off after 1hr.
// field 6 -- modbusError, 0 for no errors else response code
// field 7 -- Power_W, current household Watts, +ve for import, -ve for export (int32_t)
// field 8 -- HW_W, current Hot Water Watts, always +ve (int32_t)
// field 9 -- Import_kWh, accumulated house imported kWh (double) // Note this a +ve number always
// field 10 -- Export_kWh, accumulated house exported kWh (double) // Note this a +ve number always
// field 11 -- HW_kWh, accumulated hot water kWh, includes both solar,boost and OffPeak. (double)
// field 12 -- OffPeak_kWh, accumulated off peam kWh. (double)
// field 13 -- EOD const string "EOD" to catch truncated lines

// welcome message
/*
  Hot Water Controller
  use <newline>Boost ON<newline> to turn on Hot Water for 1hr
  use <newline>Boost OFF<newline> to turn of boost
  Data format
  HWC, readCounter, state, HWtriac, HWboost, modbusError, Power_W, HW_W, Import_kWh, Export_kWh, HW_kWh, OffPeak_kWh, EOD
  for Three Phase JSY-MK-345 ID:2 Single Phase JSY-MK-193 ID:1 system
*/
//const char WelcomeMsg[] = "Hot Water Controller\nuse <newline>Boost ON<newline> to turn on Hot Water for 1hr\nuse <newline>Boost OFF<newline> to turn of boost\nData format\n"
//                          "HWC, readCounter, state, HWtriac, HWboost, modbusError, Power_W, HW_W, Import_kWh, Export_kWh, HW_kWh, OffPeak_kWh, EOD\n";

// globals for parser to populate if get error part way through then only some will be updated, but since kWh are the total to day, this will be corrected on next successful parse.
uint32_t readCounter = 0;
int state = 0; // off,  0 Off, 1 Solar Hot Water, 2 Off Peak
int HWtriac = 0;
int HWboost = 0;
int modbusError = 0;
int32_t Power_W = 0;
int32_t HW_W = 0;
double HW_kWh = 0;
// these are local to the parser
//double Import_kWh = 0; // Note this a +ve number always
//double Export_kWh = 0; // Note this a +ve number always
//double OffPeak_kWh = 0;



void rotateDayHistory(HistoryRecordStruct & historyRecord) {
  historyRecord.export_kWhrs_yesterday = historyRecord.export_kWhrs;
  historyRecord.export_kWhrs = 0.0;
  historyRecord.import_kWhrs_yesterday = historyRecord.import_kWhrs;
  historyRecord.import_kWhrs = 0.0;
  historyRecord.hotwater_kWhrs_yesterday = historyRecord.hotwater_kWhrs;
  historyRecord.hotwater_kWhrs = 0.0;
  historyRecord.offPeak_kWhrs_yesterday = historyRecord.hotwater_kWhrs;
  historyRecord.offPeak_kWhrs = 0.0;
  historyRecord.fullyHeated_yesterday = historyRecord.fullyHeated;
  historyRecord.cost_yesterday = historyRecord.cost;

  // Reset daily running cost and daily accumulated values at midnight
  historyRecord.export_kWhrs = 0;
  historyRecord.import_kWhrs = 0;
  historyRecord.hotwater_kWhrs = 0;
  historyRecord.hotwater_kWhrs = 0;
  historyRecord.fullyHeated = false;
  historyRecord.cost = 0;
  
  initAccumulatedKWhRecord(processResults); // clear todays accumulated values

  // update yesterdays values
  updateYesterdaysValues();
}

/**
   delete excess days
   only allow last 6 full days + today for 7 days
*/
static void eraseOldDays() {
  static const int maxNoDateFiles = 6; 
  // includes today's file so when rotating history trim to 6 files
  // as todays has not been created until the first buffer save
  
  Print *old_debugPtr = debugPtr;
 // debugPtr = &Serial;
  if (debugPtr) {
    debugPtr->println(" eraseOldDays");
  }

  START_LOCK(filesLock);
  // Check if we have more than 6 date files and delete the oldest one
  std::vector<String> dateFiles;

  // List all files in the root directory and filter for date format
  File root = LittleFS.open("/");
  if (!root || !root.isDirectory()) {
    if (debugPtr) {
      debugPtr->println("Failed to open root dir for date file cleanup");
    }
    return;
  }

  File file = root.openNextFile();
  while (file) {
    String fileName = String(file.name());
    if (isFileNameDate(fileName.c_str())) {
      // Add / to front to create proper path
      String filePath = '/' + fileName;
      dateFiles.push_back(filePath);
    }
    file = root.openNextFile();
  }
  root.close();

  // If we have 6 or more files, sort them and delete the oldest one
  // note after rolling the new day file has not been written yet
  // so
  if (debugPtr) {
    debugPtr->print(" Found "); debugPtr->print(dateFiles.size()); debugPtr->println(" date files");
  }

  if (dateFiles.size() > maxNoDateFiles) {
    // Sort filenames (which are in YYYYMMDD format, so string comparison works)
    std::sort(dateFiles.begin(), dateFiles.end());
    while (dateFiles.size() > maxNoDateFiles) {
      // Delete the oldest file (first in sorted list)
      String oldestFile = dateFiles[0];
        if (debugPtr) {
          debugPtr->print("Deleting oldest date file: "); debugPtr->println(oldestFile);
        }
      if (LittleFS.remove(oldestFile.c_str())) {
        dateFiles.erase(dateFiles.begin()); // remove first element
      } else {
        if (debugPtr) {
          debugPtr->print(" Delete of oldest date file: ");  debugPtr->println(oldestFile);  debugPtr->println(" failed"); 
        }
        break; // delete failed
      }
    }
  }
  END_LOCK(filesLock);
  debugPtr = old_debugPtr;
}

/**
   rotateDayHistory
   called at midnight.
   May be just before or just after midnight.
   getCurrentDateFilename() will return the previous day (the one just before midnight

*/
void rotateDayHistory() {
  Print *old_debugPtr = debugPtr;
  //  debugPtr = &Serial;
  if (debugPtr) {
    debugPtr->println(" rotateDayHistory");
  }
  const char* dateFileName = getCurrentDateFilename(); // yestedays date even if have moved on to today (just after midnight)
  Accumulated_kWh_RecordStruct accResults;
  initAccumulatedKWhRecord(accResults);
  // this writes the values from the previous day (the one just before midnight)
  calculateDailyCost(dateFileName, accResults, true);
  rotateDayHistory(historyRecord);
  eraseOldDays();
  debugPtr = old_debugPtr;
}

/**
  uint32_t readCounter = 0;
  int state = 0; // off
  int HWtriac = 0;
  int HWboost = 0;
  int modbusError = 0;
  int32_t Power_W = 0;
  int32_t HW_W = 0;
  double Import_kWh = 0;  // Note this a +ve number always
  double Export_kWh = 0; // Note this a +ve number always
  double HW_kWh = 0;
  double OffPeak_kWh = 0;
**/

void updateHistory() {
  static millisDelay fullyHeatedDelay;
  static unsigned long fullHeatedDelay_ms = 1ul * 1000 * 60; // 1min
  bool fullyHeated = false;
  if (HWtriac && (HW_W < 100) && (state == 1)) { // in Solar mode and triac turned and HotWater not drawing current
    if (!fullyHeatedDelay.isRunning()) {
      fullyHeatedDelay.start(fullHeatedDelay_ms);
    }
  } else {
    fullyHeatedDelay.stop();
  }
  if (fullyHeatedDelay.justFinished()) {
    fullyHeated = true;
  }

  historyRecord.fullyHeated = historyRecord.fullyHeated || fullyHeated;

  if (debugPtr) {
    printHistoryRecord(historyRecord, debugPtr);
  }

  watts3PhKW_v = ((float)Power_W) / 1000.0;
  solarHotWaterKW_v = ((float)HW_W) / 1000.0;
  totalHotWaterKWhrs_v = ((float)historyRecord.hotwater_kWhrs);
  totalCost_v = historyRecord.cost;
  fullyHeated_v = historyRecord.fullyHeated;
  state_v = state;
  updated_v = true;
}


static bool parseDoubleToken(SafeString & sfToken, double & result) {
  bool rtn = true;
  // CHECK FOR 0
  sfToken.trim();
  if (sfToken.toDouble(result)) {
    // OK
  } else {
    if (debugPtr) {
      debugPtr->print(F("   ,")); debugPtr->print(sfToken); debugPtr->println(F(", is not a number"));
    }
    rtn = false;
  }
  return rtn;
}

static bool parseLongToken(SafeString & sfToken, long & result) {
  bool rtn = true;
  // CHECK FOR 0
  sfToken.trim();
  if (sfToken.toLong(result)) {
    // OK
  } else {
    if (debugPtr) {
      debugPtr->print(F("   ,")); debugPtr->print(sfToken); debugPtr->println(F(", is not a number"));
    }
    rtn = false;
  }
  return rtn;
}

static bool parseUnsignedLongToken(SafeString & sfToken, unsigned long & result) {
  bool rtn = true;
  // CHECK FOR 0
  sfToken.trim();
  if (sfToken.toUnsignedLong(result)) {
    // OK
  } else {
    if (debugPtr) {
      debugPtr->print(F("   ,")); debugPtr->print(sfToken); debugPtr->println(F(", is not a number"));
    }
    rtn = false;
  }
  return rtn;
}

// data format in CSV
// field 1 -- HWC constant string, "HWC"
// field 2 -- readCounter, 1 to max uint32_t
// field 3 -- state, 0 Off, 1 Solar Hot Water, 2 Off Peak
// field 4 -- HWtriac, 0 off, 1 on
// field 5 -- HWboost, 0 off, 1 on  automatically turns off after 1hr.
// field 6 -- modbusError, 0 for no errors else response code
// field 7 -- Power_W, current household Watts, +ve for import, -ve for export (int32_t)
// field 8 -- HW_W, current Hot Water Watts, always +ve (int32_t)
// field 9 -- Import_kWh, accumulated house imported kWh (double) // Note this a +ve number always AND un-reliable
// field 10 -- Export_kWh, accumulated house exported kWh (double) // Note this a +ve number always AND un-reliable
// field 11 -- HW_kWh, accumulated hot water kWh, includes both solar,boost and OffPeak. (double)
// field 12 -- OffPeak_kWh, accumulated off peam kWh. (double)
// field 13 -- EOD const string "EOD" to catch truncated lines


static int DATA_UPDATE_INTERVAL = 1; // = 2 sec; // 0,1,2 == 6 sec
void processData(SafeString & sfLine) {
  //  static bool firstCall = true;
  static int lineCount = 0;

  if (parseResponse(sfLine)) {
    // Create a binary record with current UTC time
    BinaryRecordStruct record;
    // Serial.println(sizeof(record)); == 24
    // 24hrs of 2sec = 43200 x 24 = 860K
    // 2hrs of 6sec = 1200 entries
    // 2hrs of 2sec = 3600 entries
    struct timeval tv;
    gettimeofday(&tv, NULL);
    record.utc_time = tv.tv_sec;
    record.utc_usec = tv.tv_usec;
    record.power_w = Power_W;
    record.hw_w = HW_W;
    record.hw_kwh = HW_kWh;  // Add HW_kWh to the record
    record.state = state;
    record.hw_triac = HWtriac;
    record.hw_boost = HWboost;

    // Use accumulate_kWh to update running totals
    accumulate_kWh(record, processResults); // use global setup in initialize
    calculateTotalCost(processResults);
    // Copy accumulated values to history record
    historyRecord.import_kWhrs = processResults.import_kWhrs;
    historyRecord.export_kWhrs = processResults.export_kWhrs;
    historyRecord.offPeak_kWhrs = processResults.offPeak_kWhrs;
    historyRecord.hotwater_kWhrs = processResults.hotwater_kWhrs;

    // Calculate running daily cost
    historyRecord.cost = processResults.cost;

    if (debugPtr) {
      debugPtr->print("Daily running totals - ");
      printDailyHistory(debugPtr);
    }

    // Add to binary buffer
    if (lineCount == 0) {
      START_LOCK(filesLock);
      addBinaryRecord(record); // will write to current file when full.
      END_LOCK(filesLock);
    }
    lineCount++;
    if (lineCount >= DATA_UPDATE_INTERVAL) {
      lineCount = 0;
    }

    // Continue with normal history update
    updateHistory(); // updates state_v
  }
}

// returns false on error
// otherwise updates globals
/**
  uint32_t readCounter = 0;
  int state = 0; // off
  int HWtriac = 0;
  int HWboost = 0;
  int modbusError = 0;
  int32_t Power_W = 0;
  int32_t HW_W = 0;
  double Import_kWh = 0; // Note this a +ve number always AND un-reliable
  double Export_kWh = 0; // Note this a +ve number always AND un-reliable
  double HW_kWh = 0;
  double OffPeak_kWh = 0;
**/

bool parseResponse(SafeString & sfLine) {

  // local to the parser  the volatiles make them available
  double Import_kWh = 0; // Note this a +ve number always
  double Export_kWh = 0; // Note this a +ve number always
  double OffPeak_kWh = 0;

  if (debugPtr) {
    debugPtr->print(" response: "); debugPtr->println(sfLine);
  }
  cSF(sfResp, sfLine.length());
  cSF(sfToken, sfLine.length());
  cSF(sfTime, 20);
  sfResp = sfLine.c_str();
  bool haveToken = sfResp.firstToken(sfToken, ','); // HWC
  if (haveToken) {
    sfToken.trim();
    if (sfToken != "HWC") {
      return false;
    }
  } else {
    return false;
  }

  haveToken = sfResp.nextToken(sfToken, ','); // readCounter
  if (haveToken) {
    if (!parseUnsignedLongToken(sfToken, readCounter)) {
      return false;
    }
  } else {
    if (debugPtr) {
      //    debugPtr->print("Bad response: "); debugPtr->println(sfLine);
    }
    return false;
  }

  haveToken = sfResp.nextToken(sfToken, ','); // state
  long stateLong = 0;
  if (haveToken) {
    if (!parseLongToken(sfToken, stateLong)) {
      return false;
    }
    if ((stateLong != 0) && (stateLong != 1) && (stateLong != 2)) {
      return false;
    }
    state = stateLong;

  } else {
    if (debugPtr) {
      //    debugPtr->print("Bad response: "); debugPtr->println(sfLine);
    }
    return false;
  }

  haveToken = sfResp.nextToken(sfToken, ','); // HWtriac
  long triacLong = 0;
  if (haveToken) {
    if (!parseLongToken(sfToken, triacLong)) {
      return false;
    }
    if ((triacLong != 0) && (triacLong != 1)) {
      return false;
    }
    HWtriac = triacLong;

  } else {
    if (debugPtr) {
      //    debugPtr->print("Bad response: "); debugPtr->println(sfLine);
    }
    return false;
  }

  haveToken = sfResp.nextToken(sfToken, ','); // HWboost
  long boostLong = 0;
  if (haveToken) {
    if (!parseLongToken(sfToken, boostLong)) {
      return false;
    }
    if ((boostLong != 0) && (boostLong != 1)) {
      return false;
    }
    HWboost = boostLong;

  } else {
    if (debugPtr) {
      //    debugPtr->print("Bad response: "); debugPtr->println(sfLine);
    }
    return false;
  }

  haveToken = sfResp.nextToken(sfToken, ','); // modbusError
  long modbusErrorLong = 0;
  if (haveToken) {
    if (!parseLongToken(sfToken, modbusErrorLong)) {
      return false;
    }
    if ((modbusErrorLong < 0)) {
      return false;
    }
    modbusError = modbusErrorLong;

  } else {
    if (debugPtr) {
      //    debugPtr->print("Bad response: "); debugPtr->println(sfLine);
    }
    return false;
  }

  haveToken = sfResp.nextToken(sfToken, ','); // Power_W
  long Power_WLong = 0;
  if (haveToken) {
    if (!parseLongToken(sfToken, Power_WLong)) {
      return false;
    }
    Power_W = Power_WLong;

  } else {
    if (debugPtr) {
      //    debugPtr->print("Bad response: "); debugPtr->println(sfLine);
    }
    return false;
  }

  haveToken = sfResp.nextToken(sfToken, ','); // HW_W
  long HW_WLong = 0;
  if (haveToken) {
    if (!parseLongToken(sfToken, HW_WLong)) {
      return false;
    }
    HW_W = HW_WLong;

  } else {
    if (debugPtr) {
      //    debugPtr->print("Bad response: "); debugPtr->println(sfLine);
    }
    return false;
  }


  haveToken = sfResp.nextToken(sfToken, ','); // Import_kWh // Note this a +ve number always AND un-reliable
  if (haveToken) {
    Import_kWh = 0;
    if (!parseDoubleToken(sfToken, Import_kWh)) {
      return false;
    }
  } else {
    if (debugPtr) {
      //    debugPtr->print("Bad response: "); debugPtr->println(sfLine);
    }
    return false;
  }

  haveToken = sfResp.nextToken(sfToken, ','); // Export_kWh // Note this a +ve number always AND un-reliable
  if (haveToken) {
    Export_kWh = 0;
    if (!parseDoubleToken(sfToken, Export_kWh)) {
      return false;
    }
  } else {
    if (debugPtr) {
      //    debugPtr->print("Bad response: "); debugPtr->println(sfLine);
    }
    return false;
  }

  haveToken = sfResp.nextToken(sfToken, ','); // HW_kWh
  if (haveToken) {
    if (!parseDoubleToken(sfToken, HW_kWh)) {
      return false;
    }
  } else {
    if (debugPtr) {
      //    debugPtr->print("Bad response: "); debugPtr->println(sfLine);
    }
    return false;
  }

  haveToken = sfResp.nextToken(sfToken, ','); // OffPeak_kWh
  if (haveToken) {
    if (!parseDoubleToken(sfToken, OffPeak_kWh)) {
      return false;
    }
  } else {
    if (debugPtr) {
      //    debugPtr->print("Bad response: "); debugPtr->println(sfLine);
    }
    return false;
  }

  haveToken = sfResp.nextToken(sfToken, ','); // unknown
  if (haveToken) {
    sfToken.trim();
    if (sfToken != "EOD") {
      return false;
    }
  }

  haveToken = sfResp.nextToken(sfToken, ','); // unknown
  if (haveToken) {
    if (debugPtr) {
      debugPtr->print("Bad response (too many fields): "); debugPtr->println(sfLine);
    }
    return false;
  }

  JSY_Import_kWh_v = Import_kWh; // Note this a +ve number always
  JSY_Export_kWh_v = Export_kWh; // Note this a +ve number always
  JSY_HW_kWh_v = HW_kWh;
  JSY_OffPeak_kWh_v = OffPeak_kWh;

  // else
  return true;
}
