/*
   systemTimeUtils.cpp
   by Matthew Ford,  2021/12/06
   (c)2021 Forward Computing and Control Pty. Ltd.
   NSW, Australia  www.forward.com.au
   This code may be freely used for both private and commerical use.
   Provide this copyright is maintained.
*/
#include <Arduino.h>
#include <time.h>                       // time() ctime()
#include <sys/time.h>                   // struct timeval
#include "systemTimeUtils.h"

// normally DEBUG is commented out
//#define DEBUG
static Print* debugPtr = NULL;  // local to this file



//struct tm;
//Defined in header <time.h>
//Structure holding a calendar date and time broken down into its components.
//Member objects
//int tm_sec  seconds after the minute – [0, 61] (until C99)[0, 60] (since C99)[for leap second]
//int tm_min minutes after the hour – [0, 59]
//int tm_hour hours since midnight – [0, 23]
//int tm_mday day of the month – [1, 31]
//int tm_mon months since January – [0, 11]
//int tm_year years since 1900
//int tm_wday days since Sunday – [0, 6]
//int tm_yday days since January 1 – [0, 365]
//int tm_isdst Daylight Saving Time flag. The value is positive if DST is in effect, zero if not and negative if no information is available
//
//The Standard mandates only the presence of the aforementioned members in either order.
//The implementations usually add more data-members to this structure.

//static timeval tv;
//static timespec tp;
//static time_t now;
//static uint32_t now_ms, now_us;



/**
 * Convert a tm struct to a SafeString
 *
 * @param tmPtr The tm struct to convert
 * @param sf The SafeString to store the result
 * @return Reference to the SafeString
 */
SafeString& tm_to_sf(struct tm* tmPtr, SafeString &sf) {
  sf.print(tmPtr->tm_year + 1900);
  sf.print("-");
  if(tmPtr->tm_mon + 1 < 10) sf.print("0");
  sf.print(tmPtr->tm_mon + 1);
  sf.print("-");
  if(tmPtr->tm_mday < 10) sf.print("0");
  sf.print(tmPtr->tm_mday);
  sf.print(" ");
  if(tmPtr->tm_hour < 10) sf.print("0");
  sf.print(tmPtr->tm_hour);
  sf.print(":");
  if(tmPtr->tm_min < 10) sf.print("0");
  sf.print(tmPtr->tm_min);
  sf.print(":");
  if(tmPtr->tm_sec < 10) sf.print("0");
  sf.print(tmPtr->tm_sec);
  
  return sf;
}

/**
 * Convert a tm struct to a character array, thread safe
 *
 * @param timeStruct The tm struct to convert
 * @param buffer The character buffer to store the result (must be at least 26 bytes)
 * @param bufferSize Size of the buffer
 * @return Pointer to the buffer or empty string if buffer is too small
 */
const char* tmToChar26(struct tm* timeStruct, char* buffer, size_t bufferSize) {
  static const char empty_str[] = "";
  
  if (buffer == NULL || bufferSize < 26) {
    return empty_str;  // Return empty string if buffer too small or NULL
  }
  snprintf(buffer, bufferSize,
           "%04d-%02d-%02d %02d:%02d:%02d",
           timeStruct->tm_year + 1900,
           timeStruct->tm_mon + 1,
           timeStruct->tm_mday,
           timeStruct->tm_hour,
           timeStruct->tm_min,
           timeStruct->tm_sec);
           
  return buffer;
}


void printLocalSystemTime(Print *outPtr) {
  if (!outPtr) {
    return;
  }
  struct tm timeinfo;
  if (!getLocalTime(&timeinfo)) {
    outPtr->println("Failed to obtain time 1");
    return;
  }
  outPtr->println(&timeinfo, "%A, %B %d %Y %H:%M:%S zone %Z %z ");
}


// only handles +v numbers
static void print2digits(String & result, uint num) {
  if (num < 10) {
    result += '0';
  }
  result += num;
}

static String getYYYYMMDD(struct tm * tmPtr) {
  String result; // yyyy/mm/dd
  print2digits(result, tmPtr->tm_year + 1900);
  result += '/';
  print2digits(result, tmPtr->tm_mon + 1);
  result += '/';
  print2digits(result, tmPtr->tm_mday );
  return result;
}

static String getDDMMYYYY(struct tm * tmPtr) {
  String result; // dd/mm/yyyy
  print2digits(result, tmPtr->tm_mday );
  result += '/';
  print2digits(result, tmPtr->tm_mon + 1);
  result += '/';
  print2digits(result, tmPtr->tm_year + 1900);
  return result;
}

static String getHHMMss(struct tm * tmPtr) {
  String result; // hh:mm:ss
  print2digits(result, tmPtr->tm_hour);
  result += ':';
  print2digits(result, tmPtr->tm_min);
  result += ':';
  print2digits(result, tmPtr->tm_sec);
  return result;
}

static String getHHMM(struct tm * tmPtr) {
  String result; // hh:mm:ss
  print2digits(result, tmPtr->tm_hour);
  result += ':';
  print2digits(result, tmPtr->tm_min);
  return result;
}

// local time of day in mins
unsigned int getLocalTime_mins() {
  time_t now = time(nullptr);
  struct tm tm;
  struct tm* tmPtr = localtime_r(&now,&tm);
  unsigned int rtn = (tmPtr->tm_hour);
  rtn = rtn * 60 + tmPtr->tm_min;
  return rtn;
}

// local time of day in sec
unsigned int getLocalTime_sec() {
  time_t now = time(nullptr);
  struct tm tm;
  struct tm* tmPtr = localtime_r(&now,&tm);
  unsigned int rtn = (tmPtr->tm_hour);
  rtn = rtn * 60 + tmPtr->tm_min;
  rtn = rtn * 60 + tmPtr->tm_sec;
  return rtn;
}

// local time the number of sec until midnight
unsigned int getLocalSecLeftInDay() {
  time_t now = time(nullptr);
  struct tm tm;
  struct tm* tmPtr = localtime_r(&now,&tm);
  int rtn = (tmPtr->tm_hour);
  rtn = rtn * 60 + tmPtr->tm_min;
  rtn = rtn * 60 + tmPtr->tm_sec;
  rtn = (24*60*60) - rtn;
  if (rtn < 0) {
    rtn = 0;
  } 
  return rtn;
}
// local time the number of sec until midnight
unsigned int getLocalSecLeftInHr() {
  time_t now = time(nullptr);
  struct tm tm;
  struct tm* tmPtr = localtime_r(&now,&tm);
  int rtn = tmPtr->tm_min;
  rtn = rtn * 60 + tmPtr->tm_sec;
  rtn = (60*60) - rtn;
  if (rtn < 0) {
    rtn = 0;
  } 
  return rtn;
}


//// local time HH:MM:ss in sec
String getLocalTime_s() {
  time_t now = time(nullptr);
  struct tm tm;
  struct tm* tmPtr = localtime_r(&now,&tm);
  uint32_t rtn = (tmPtr->tm_hour);
  rtn = rtn * 60 + tmPtr->tm_min;
  rtn = rtn * 60 + tmPtr->tm_sec;
  return String(rtn);
}

// small String <=10char) in ESP8266/ESP32 use built in char[]
String getCurrentTime_hhmm() {
  //  gettimeofday(&tv, nullptr);
  //  clock_gettime(0, &tp);
  time_t now = time(nullptr);
  struct tm tm;
  struct tm* tmPtr = localtime_r(&now,&tm);
  return getHHMM(tmPtr);
}
// small String <=10char) in ESP8266/ESP32 use built in char[]
String getCurrentDate_ddmmyyyy() {
  //  gettimeofday(&tv, nullptr);
  //  clock_gettime(0, &tp);
  time_t now = time(nullptr);
  struct tm tm;
  struct tm* tmPtr = localtime_r(&now,&tm);
  return getDDMMYYYY(tmPtr);
}

String getCurrentDate_yyyymmdd() {
  //  gettimeofday(&tv, nullptr);
  //  clock_gettime(0, &tp);
  time_t now = time(nullptr);
  struct tm tm;
  struct tm* tmPtr = localtime_r(&now,&tm);
  return getYYYYMMDD(tmPtr);
}

// small String <=10char) in ESP8266/ESP32 use built in char[]
// local time in hhmmss
String getCurrentTime_hhmmss() {
  //  gettimeofday(&tv, nullptr);
  //  clock_gettime(0, &tp);
  time_t now = time(nullptr);
  struct tm tm;
  struct tm* tmPtr = localtime_r(&now,&tm);
  return getHHMMss(tmPtr);
}

String getUTCTime() {
  //  gettimeofday(&tv, nullptr);
  //  clock_gettime(0, &tp);
  time_t now = time(nullptr);
  struct tm* tmPtr = gmtime(&now);
  return getHHMMss(tmPtr);
}

#define PTM(w) \
  debugPtr->print(" " #w "="); \
  debugPtr->print(tm->tm_##w);

void printTm(const char* what, const tm * tm) {
  if (!debugPtr) {
    return;
  }
  debugPtr->print(what);
  PTM(isdst); PTM(yday); PTM(wday);
  PTM(year);  PTM(mon);  PTM(mday);
  PTM(hour);  PTM(min);  PTM(sec);
}

void showTimeTZ(Print *debugPtr) {
  if (!debugPtr) {
    return;
  }
  time_t now = time(nullptr);
  // human readable
  debugPtr->print("ctime:     ");
  debugPtr->print(ctime(&now));
    // timezone and demo in the future
  debugPtr->printf("timezone:  %s\r\n", getenv("TZ") ? : "(none)");
}

void showTimeDebug(Print *debugPtr) {
  if (!debugPtr) {
    return;
  }
  debugPtr->println("ShowTimeDebug");
  timeval tv;
  timespec tp;
  time_t now;
  uint32_t now_ms, now_us;

  gettimeofday(&tv, nullptr);
  clock_gettime(0, &tp);
  now = time(nullptr);
  now_ms = millis();
  now_us = micros();

  debugPtr->println();
  struct tm tm;
  printTm("localtime:", localtime_r(&now, &tm));
  debugPtr->println();
  printTm("gmtime:   ", gmtime(&now));
  debugPtr->println();

  // time from boot
  debugPtr->print("clock:     ");
  debugPtr->print((uint32_t)tp.tv_sec);
  debugPtr->print("s + ");
  debugPtr->print((uint32_t)tp.tv_nsec);
  debugPtr->println("ns");

  // time from boot
  debugPtr->print("millis:    ");
  debugPtr->println(now_ms);
  debugPtr->print("micros:    ");
  debugPtr->println(now_us);

  // EPOCH+tz+dst
  debugPtr->print("gtod:      ");
  debugPtr->print((uint32_t)tv.tv_sec);
  debugPtr->print("s + ");
  debugPtr->print((uint32_t)tv.tv_usec);
  debugPtr->println("us");

  // EPOCH+tz+dst
  debugPtr->print("time:      ");
  debugPtr->println((uint32_t)now);

  // timezone and demo in the future
  debugPtr->printf("timezone:  %s\r\n", getenv("TZ") ? : "(none)");

  // human readable
  debugPtr->print("ctime:     ");
  debugPtr->print(ctime(&now));


  debugPtr->println();

}
